/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Main.java,v 1.25 1999/10/16 23:25:35 vgp Exp $
 */

package at.dms.classfile;

import java.io.*;
import at.dms.util.Utils;

/**
 * This class allow to test the classfile reader / writer system by reading / writing
 * every class given as parameters.
 */
public class Main {

  // --------------------------------------------------------------------
  // ENTRY POINT
  // --------------------------------------------------------------------

  /**
   * Entry point to the disassembler
   */
  public static void main(String[] args) {
    String[]		inputFiles = new String[args.length];
    int			fileCount = 0;

    for (int i = 0; i < args.length; i++) {
      if (args[i].equals("-d")) {
	i += 1;
	if (i < args.length) {
	  destination = args[i];
	} else {
	  displayUsage();
	}
      } else if (args[i].equals("-version")) {
	System.out.println("KOPI Classfile Version: " + VERSION);
      } else if (args[i].equals("-R")) {
	i += 1;
	if (i < args.length) {
	  try {
	    optionRepeat = Integer.valueOf(args[i]).intValue();
	  } catch (Exception e) {
	    displayUsage();
	  }
	} else {
	  displayUsage();
	}
      } else if (args[i].equals("-inter")) {
	optionInterfaceOnly = true;
      } else {
	inputFiles[fileCount++] = args[i];
      }
    }

    if (fileCount == 0) {
      displayUsage();
    }

    for (int i = 0; i < fileCount; i++) {
      doFile(new File(inputFiles[i]));
    }

    if (errorsFound) {
      System.exit(1);
    }
  }

  /**
   * Read all classfiles
   */
  public static void doFile(File f) {
    // If this is a directory, walk each file/dir in that directory
    if (f.isDirectory()) {
      String[] files = f.list();

      for (int i = 0; files != null && i < files.length; i++) {
	doFile(new File(f, files[i]));
      }
    } else if ((f.getName().length() > 6) &&
	     f.getName().substring(f.getName().length() - 6).equals(".class")) {
      // otherwise, if this is a class file, parse it!
      try {
	readFile(f.getName(), f);
      } catch (Exception exc) {
	exc.printStackTrace();
	System.err.println("Can't load file " + f);
	return;
      }
    }
  }

  /**
   * Read a classfile
   */
  public static void readFile(String f, File file) throws Exception {
    ClassInfo		classInfo;
    System.err.println("Read " + f + (optionInterfaceOnly ? " interface" : ""));
    // choose your reader:
    FileInputStream s = new FileInputStream(file);
    int		size = s.available();
    if (size > data.length) {
      data = new byte[size];
    } else if (size == 0) {
      s.close();
      return;
    }
    s.read(data);
    s.close();

    for (int i = 0; i < optionRepeat; i++) {
      if (i > 0) {
	System.err.println(".");
      }
      InputStream is = new ByteArrayInputStream(data, 0, size);
      classInfo = new ClassInfo(new DataInputStream(is), optionInterfaceOnly);
      is.close();
      if (!optionInterfaceOnly && destination != null) {
	writeClassfile(classInfo, destination);
      }
    }
  }

  /**
   * Creates a class file from class info
   */
  private static void writeClassfile(ClassInfo classInfo, String destination) {
    String[]	classPath = Utils.splitQualifiedName(classInfo.getName());

    if (classPath[0] != null) {		// the class is part of a package
      String	classDir =
	classPath[0].replace('.', File.separatorChar).replace('/', File.separatorChar);

      if (destination != null) {
	destination += File.separator + classDir;
      }
    }

    if (destination != null) {		// check that destination exists or else create it
      File	destDir = new File(destination);

      if (! destDir.exists()) {
	destDir.mkdirs();
      }

      if (! destDir.isDirectory()) {
	throw new RuntimeException("File " + destination + " is not a directory.");
      }
    }

    File	outputFile = new File(destination, classPath[1] + ".class");

    try {
      DataOutputStream		out =
	new DataOutputStream(new BufferedOutputStream(new FileOutputStream(outputFile)));

      classInfo.write(out);
      System.err.println("Write " + classPath[1]);
      out.flush();
      out.close();

      System.out.println("Generated: " + outputFile.getPath());
    } catch (ClassFileFormatException e) {
      throw new RuntimeException("Semantic error on " + outputFile.getPath() + ": " + e.getMessage());
    } catch (java.io.IOException e) {
      throw new RuntimeException("I/O Exception on " + outputFile.getPath() + ": " + e.getMessage());
    }
  }

  // --------------------------------------------------------------------
  // PRIVATE METHODS
  // --------------------------------------------------------------------

  /**
   * Display an usage information
   */
  private static void displayUsage() {
    System.err.println("usage: at.dms.classfile.Main [-d <directory>] " +
		       "[-inter] [-R x] (classfile | zipfile | directory)*");
    System.err.println("This tools read recursively every class given as parameters");
    System.err.println("-inter:\tOnly read interface [force nowrite]");
    System.err.println("-d: d  \tSelect a directory to generate classfile");
    System.err.println("-R: r  \tRepeate r times the read/write process");
    System.exit(1);
  }

  // --------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // --------------------------------------------------------------------

  private static final String	VERSION			= "1.3D";
  private static int		optionRepeat		= 1;
  private static boolean	optionInterfaceOnly;
  private static boolean	errorsFound;
  private static String		destination;
  private static byte[]		data = new byte[100000];
}
