/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Compiler.java,v 1.6 1999/10/08 11:18:08 vgp Exp $
 */

package at.dms.compiler;

import java.io.File;
import java.io.PrintWriter;
import java.util.Vector;

/**
 * This class defines the common behaviour of all KOPI compilers.
 */
public abstract class Compiler {

  /**
   * Creates a new compiler instance.
   */
  protected Compiler() {
    timestamp = System.currentTimeMillis();
  }

  /**
   * Sets the directory where to search for source files.
   * @param	dir		the name of the directory
   */
  protected void setWorkingDirectory(String dir) {
    this.workingDirectory = dir;
  }

  /**
   * Returns the timestamp of the compilation session,
   * identifying it uniquely.
   */
  public final long getTimestamp() {
    return timestamp;
  }

  /**
   * Takes a vector of file names an checks that each exists.
   * @param	files			a vector of names
   * @return	a vector of files known to exist
   * @exception	UnpositionedError	at least one file does not exist
   */
  public Vector verifyFiles(Vector names) throws UnpositionedError {
    Vector	files = new Vector(names.size());

    for (int i = 0; i < names.size(); i++) {
      String	name;
      File	file;

      if (workingDirectory == null) {
	name = (String)names.elementAt(i);
      } else {
	name = workingDirectory + File.separatorChar + (String)names.elementAt(i);
      }
      file = new File(name);

      if (!file.exists()) {
	file = new File((String)names.elementAt(i));
	if (!file.exists() || !file.isAbsolute()) {
	  throw new UnpositionedError("usage-unknown-file", names.elementAt(i), null);
	}
      }
      files.addElement(file);
    }

    return files;
  }

  /**
   * Checks if destination is absolute or relative to working directory.
   * !!! CHECK WAS: if (!file.exists() || !dest.getAbsolutePath().equals(destination))
   */
  protected String checkDestination(String destination) {
    if (workingDirectory == null) {
      return destination;
    } else if (destination.equals("")) {
      return workingDirectory;
    } else if (new File(destination).isAbsolute()) {
      return destination;
    } else {
      return workingDirectory + File.separatorChar + destination;
    }
  }

  /**
   * Runs a compilation session
   *
   * @param	dir		the working directory
   * @param	err		the error output stream
   * @param	console		the console to show the progess of the compilation (!!! not used)
   * @param	args		the arguments to the compiler
   * @return	true iff the compilation succeeded
   */
  public abstract boolean run(String dir, PrintWriter err, Console console, String[] args);

  /**
   * Reports a trouble (error or warning).
   *
   * @param	trouble		a description of the trouble to report.
   */
  public abstract void reportTrouble(PositionedError trouble);

  /**
   * Returns true iff comments should be parsed (false if to be skipped).
   */
  public abstract boolean parseComments();

  /**
   * Returns true iff compilation runs in verbose mode.
   */
  public abstract boolean verboseMode();

  // --------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // --------------------------------------------------------------------

  // each compilation session is identified by a timestamp
  private final long		timestamp;

  // the directory where to search source files
  private String		workingDirectory;
}
