/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: HandLexer.java,v 1.6 1999/10/28 14:03:11 vgp Exp $
 */

package at.dms.compiler;

import java.util.Vector;

import antlr.Token;
import antlr.TokenStream;

/**
 * This class describes the capabilities of lexers (token streams)
 * to share a common input buffer and to handle line numbers.
 */
public abstract class HandLexer implements TokenStream {

  // --------------------------------------------------------------------
  // CONSTRUCTORS
  // --------------------------------------------------------------------

  /**
   * Constructs a new hand written lexer
   */
  protected HandLexer(Compiler compiler, InputBuffer buffer) {
    this.compiler = compiler;
    this.buffer = buffer;
  }

  // --------------------------------------------------------------------
  // ACCESSORS
  // --------------------------------------------------------------------

  /**
   * Returns the current line number in the source code.
   */
  public final int getLine() {
    return buffer.line;
  }

  /**
   * Sets the current line number in the source code.
   */
  public final void setLine(int line) {
    buffer.line = line;
  }

  /**
   * >>> COMMENT
   */
  public final int getPosition() {
    return buffer.pos;
  }

  /**
   * >>> COMMENT
   */
  public final void setPosition(int pos) {
    buffer.pos = pos;
  }

  /**
   * >>> COMMENT
   */
  public final TokenReference getTokenReference() {
    return TokenReference.build(buffer.file, buffer.line);
  }

  /**
   * Add an error into the list and eat it
   * This method should be called after a try catch block after catching exception
   * or directly without exception thrown
   * @param	error		the error
   */
  protected final void reportTrouble(PositionedError trouble) {
    compiler.reportTrouble(trouble);
  }

  /**
   * Add an error into the list and eat it
   * This method should be called after a try catch block after catching exception
   * or directly without exception thrown
   * @param	mess		the error message
   * @param	param1		the first message parameter
   * @param	param2		the second message parameter
   * @exception	PositionedError	an error with reference to the source file
   */
  protected final void reportTrouble(String mess, Object param1, Object param2) {
    reportTrouble(new PositionedError(getTokenReference(), mess, param1, param2));
  }

  /**
   * Returns true iff comments should be parsed (false if to be skipped)
   */
  protected final boolean parseComments() {
    return compiler.parseComments();
  }

  /**
   * >>> COMMENT
   */
  protected abstract Token getToken() throws PositionedError;

  /**
   *  Identifiers/Keywords/true/false/null (start with java letter)
   *  numeric literal (start with number)
   *  character literal (start with single quote)
   *  string (start with double quote)
   *  separator (parens, braces, brackets, semicolon, comma, period)
   *  operator (equals, plus, minus, etc)
   *  whitespace
   *  comment (start with slash)
   */
  public final Token nextToken() {
    while (true) {
      try {
	if (buffer.pos >= buffer.size) {
	  return EOF;
	}

	switch (buffer.data[buffer.pos]) {

	  // White space:
	case '\n':	// LineTerminator
	  buffer.line++;
	case '\r':	// LineTerminator
	case ' ':		// ASCII SP
	case '\t':	// ASCII HT
	case '\f':	// ASCII FF
	  buffer.pos++;
	  continue;

	  // EOF character:
	case '\020': // ASCII SUB
	  buffer.pos++;
	  return EOF;

	  // else, a Token
	default:
	  {
	    Token	tok = getToken();

	    if (tok != null) {
	      return tok;
	    }
	    continue;
	  }
	}
      } catch (PositionedError e) {
	reportTrouble(e);
	buffer.pos++;
      } catch (Exception e) {
	e.printStackTrace();
	if (buffer.pos < buffer.size) {
	  reportTrouble("no-viable-alt-for-char", "" + buffer.data[buffer.pos], null);
	} else {
	  reportTrouble("unexpected-eof", null, null);
	  return EOF;
	}
	buffer.pos++;
      }
    }
  }

  /**
   *
   */
  public JavaStyleComment[] getStatementComment() {
    JavaStyleComment[]	comm = (JavaStyleComment[])Utils.vectorToArray(comments, JavaStyleComment.class);

    comments.setSize(0);

    return comm;
  }

  /**
   * Whats's that
   *
  public Vector getComment() {
    Vector	result = comments;

    comments = new Vector();
    return result;
  }

  /**
   *
   */
  protected boolean getJavaComment(boolean store) {
    // line.charAt(line_buffer.pos+0) is '/'
    if (buffer.data[buffer.pos + 1] == '/') {
      // EndOfLineComment
      buffer.pos += 2;

      int	start = buffer.pos;

      while (buffer.data[buffer.pos] != '\n') {
	buffer.pos++;
      }
      
      if (store) {
	addComment(new JavaStyleComment(new String(buffer.data, start, buffer.pos - start),
					true,
					spaceBefore(start),
					spaceAfter(buffer.pos)));
      }
      return true;
    } else if (buffer.data[buffer.pos + 1] == '*') {
      // TraditionalComment
      snarfJavaComment(store);
      return true;
    } else {
      // it's a token, not a comment.
      return false;
    }
  }

  /**
   * >>> COMMENT
   */
  protected final void addComment(JavaStyleComment comment) {
    comments.addElement(comment);
  }

  protected void snarfJavaComment(boolean store) {
    int		start = buffer.pos;

    buffer.pos += 2;
    while (true) { // Grab CommentTail
    loop:
      while (true) {
	switch(buffer.data[buffer.pos]) {
	case '*':
	  break loop;
	case '\n':
	  buffer.line++;
	default:
	  buffer.pos++;
	}
      }

      // At this point, line.charAt(line_buffer.pos)=='*'
      // Grab CommentTailStar starting at line_buffer.pos+1.

      if (buffer.data[buffer.pos + 1]=='/') {
	buffer.pos += 2;
	if (store) {
	  addComment(new JavaStyleComment(new String(buffer.data, start + 2, buffer.pos - start - 4), 
					  false,
					  spaceBefore(start),
					  spaceAfter(buffer.pos)));
	}
	return;
      }
      buffer.pos++;
    }
  }

  // --------------------------------------------------------------------
  // PROTECTED UTILITIES
  // --------------------------------------------------------------------

  private int previousReturn(int i) {
    for (; i >= 0; i--) {
      if (buffer.data[i] == '\n') {
	return i;
      }
    }
    return 0;
  }

  private int previousNonBlank(int i) {
    for (; i >= 0; i--) {
      if (buffer.data[i] != ' ' && buffer.data[i] != '\t') {
	return i;
      }
    }
    return 0;
  }

  private int nextReturn(int i) {
    for (; i < buffer.data.length; i++) {
      if (buffer.data[i] == '\n' || buffer.data[i] == 0) {
	return i;
      }
    }

    return buffer.data.length;
  }

  private int nextNonBlank(int i) {
    for (; i < buffer.data.length; i++) {
      if ((buffer.data[i] != ' ' && buffer.data[i] != '\t') || buffer.data[i] == 0) {
	return i;
      }
    }

    return buffer.data.length;
  }

  protected boolean spaceBefore(int pos) {
    int		previousReturn = previousReturn(pos);

    return previousReturn == 0 ? 
      false :
      previousNonBlank(previousReturn - 1) <= previousReturn(previousReturn - 1);
  }

  protected boolean spaceAfter(int pos) {
    int	  	nextReturn = nextReturn(pos);

    return nextNonBlank(nextReturn + 1) >= nextReturn(nextReturn + 1);
  }

  // --------------------------------------------------------------------
  // PROTECTED AND PRIVATE DATA MEMBERS
  // --------------------------------------------------------------------

  private final Compiler		compiler;
  protected InputBuffer			buffer;
  protected Vector			comments = new Vector();

  private static final CToken		EOF = new CToken(1, "End of file");
}
