/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CClass.java,v 1.63 1999/10/28 12:45:47 vgp Exp $
 */

package at.dms.kjc;

import java.io.File;
import java.io.DataOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileOutputStream;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;

import at.dms.classfile.*;
import at.dms.compiler.Utils;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.util.VectorCache;

/**
 * This class represents the exported members of a class (inner classes, methods and fields)
 */
public abstract class CClass extends CMember {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Constructs a class export from file
   */
  public CClass(CClass owner,
		int version,
		String sourceFile,
		int modifiers,
		String ident,
		String qualifiedName,
		CClassType superClass,
		boolean deprecated) {
    super(owner, modifiers, ident, deprecated);
    this.version = version;
    this.sourceFile = sourceFile;
    this.qualifiedName = qualifiedName;
    this.superClass = superClass;

    int		cursor = qualifiedName.lastIndexOf('/');

    this.packageName = cursor > 0 ? qualifiedName.substring(0, cursor).intern() : "";
    timestamp = Main.getTimeStamp();
  }

  /**
   * Ends the definition of this class
   */
  public void close(CClassType[] interfaces,
		    CField[] fields,
		    Hashtable fields_H,
		    CMethod[] methods,
		    Hashtable methods_H) {
    this.interfaces = interfaces;
    this.fields = fields;
    this.fields_H = fields_H;
    this.methods = methods;
    this.methods_H = methods_H;
    assert((fields_H != null) || (fields != null));
    assert(interfaces != null);
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * @return	the interface
   */
  public CClass getCClass() {
    return this;
  }

  /**
   * Return the compilation process that have generated this class
   */
  public long getTimeStamp() {
    return timestamp;
  }

  /**
   * @return true if this class is an interface
   */
  public boolean isInnerClass() {
    return getOwner() != null;
  }

  /**
   * @return	true if this member is static
   */
  public boolean isAbstract() {
    return hasFlag(getModifiers(), ACC_ABSTRACT);
  }

  /**
   * @return true if this class is an interface
   */
  public boolean isAnnonymousClass() {
    return getIdent().length() == 0;
  }

  /**
   * @return true if this class is an interface
   */
  public boolean hasOuterThis() {
    return hasOuterThis;
  }

  /**
   * @return true if this class is an interface
   */
  public boolean isInterface() {
    return hasFlag(getModifiers(), Constants.ACC_INTERFACE);
  }

  /**
   * @return the full name of this class
   */
  public String getQualifiedName() {
    return qualifiedName;
  }

  /**
   * @return the name of the package of this class
   */
 public String getPackage() {
    return packageName;
  }

  /**
   * Returns the super class of this class
   */
  public CClass getSuperClass() {
    return superClass == null ? null : superClass.getCClass();
  }

  /**
   * Sets the super class of this class
   */
  public void setSuperClass(CClassType superClass) {
    this.superClass = superClass;
  }

  /**
   * Returns the type of this class
   */
  public CClassType getType() {
    return type == null ? (type = CClassType.lookup(getQualifiedName())) : type;
  }

  /**
   * @return the source file of this class
   */
  public String getSourceFile() {
    return sourceFile;
  }

  /**
   * Sets hasOuterThis
   */
  public void setHasOuterThis(boolean hot) {
    hasOuterThis = hot;
  }

  /**
   * @param	ident		the name of the field
   * @return	the field
   */
  public CField getField(String ident) {
    if (fields_H != null) {
      return (CField)fields_H.get(ident);
    } else {
      for (int i = 0; i < fields.length; i++) {
	if (fields[i].getIdent() == ident) {
	  return fields[i];
	}
      }
    }
    
    return null;
  }

  public int getFieldCount() {
    return fields_H == null ? fields.length : fields_H.size();
  }

  /**
   * @return the InnerClasses
   */
  public CClassType[] getInnerClasses() {
    return innerClasses;
  }

  /**
   * End of first pass, we need innerclassesinterfaces
   */
  public void setInnerClasses(CClassType[] inners) {
    innerClasses = inners;
  }

  /**
   * descendsFrom
   * @param	from	an other CClass
   * @return	true if this class inherit from "from" or equals "from"
   */
  public boolean descendsFrom(CClass from) {
    if (from == this) {
      return true;
    }
    if (superClass == null) {
      // java/lang/object
      return false;
    }

    if (superClass != null && getSuperClass().descendsFrom(from)) {
      return true;
    }

    for (int i = 0; i < interfaces.length; i++) {
      if (interfaces[i].getCClass().descendsFrom(from)) {
	return true;
      }
    }

    return false;
  }

  /**
   * descendsFrom
   * @param	outer	an other CClass
   * @return	true if this class is define inside outer
   */
  public boolean isDefineInside(CClass outer) {
    if (this == outer) {
      return true;
    }

    if (getOwner() == null) {
      return false;
    }

    return getOwner().isDefineInside(outer);
  }

  // ----------------------------------------------------------------------
  // LOOKUP
  // ----------------------------------------------------------------------

  /**
   * This can be used to see if a given class name is visible
   *    inside of this file.  This includes globally-qualified class names that
   *    include their package and simple names that are visible thanks to some
   *    previous import statement or visible because they are in this file.
   *    If one is found, that entry is returned, otherwise null is returned.
   * @param	name		a TypeName (6.5.2)
   */
  public CClass lookupClass(String name) {
    if ((name.indexOf('/') < 0)) {
      for (int i = 0; i < innerClasses.length; i++) {
	if (innerClasses[i].getCClass().getIdent().equals(name)) {
	  return innerClasses[i].getCClass();
	}
      }
    }
    if (superClass != null) {
      return superClass.getCClass().lookupClass(name);
    }

    return null;
  }

  /**
   * lookupMethod
   * search for a matching method with the provided type parameters
   * look in parent hierarchy as needed
   * @param	name		method name
   * @param	params		method parameters
   * @exception UnpositionedError	this error will be positioned soon
   */
  public CMethod lookupMethod(String name, CType[] params) throws UnpositionedError {
    CMethod	method = getMatchingMethod(name, params);
    if (method != null) {
      return method;
    }

    return lookupSuperMethod(name, params);
  }

  /**
   * lookupSuperMethod
   * search for a matching method with the provided type parameters
   * look in parent hierarchy as needed
   * @param	name		method name
   * @param	params		method parameters
   * @exception UnpositionedError	this error will be positioned soon
   */
  public CMethod lookupSuperMethod(String name, CType[] params) throws UnpositionedError {
    CMethod	method;

    if (superClass == null) {
      method = null;
    } else {
      method = getSuperClass().lookupMethod(name, params);
    }

    for (int i = 0; i < interfaces.length && method == null; i++) {
      method = interfaces[i].getCClass().lookupMethod(name, params);
    }

    return method;
  }

  /**
   * lookupField
   * search for a field
   * look in parent hierarchy as needed
   * @param	name		method name
   * @param	params		method parameters
   * @exception UnpositionedError	this error will be positioned soon
   */
  public CField lookupField(String name) throws UnpositionedError {
    CField	field = getField(name);

    if (field != null) {
      return field;
    }

    if (superClass != null) {
      // java.lang.object / interfaces
      field = getSuperClass().lookupField(name);
    }

    for (int i = 0; i < interfaces.length; i++) {
      CField	newField = interfaces[i].getCClass().lookupField(name);

      if ((field != null) && (newField != null) && (field.getOwner() != newField.getOwner())) {
	throw new UnpositionedError("field-ambiguous", name);
      }

      if (newField != null) {
	field = newField;
      }
    }

    return field;
  }

  /**
   * getMatchingMethod
   * search for a matching method with the provided type parameters
   * Linear search (if method exists)
   * @param	ident		method name
   * @param	params		method parameters
   * @exception UnpositionedError	this error will be positioned soon
   */
  private CMethod getMatchingMethod(String ident, CType[] params) throws UnpositionedError {
    CMethod	method = null;
    boolean	lastMethodAmbiguous = false;

    if (methods_H != null &&
	ident != JAV_CONSTRUCTOR &&
	ident != JAV_INIT &&
	ident != JAV_STATIC_INIT &&
	!methods_H.containsKey(ident)) {
      return null;
    }

    for (int i = 0; i < methods.length; i++) {
      CMethod	current = methods[i];

      if (current.isMatching(ident, params)) {
	if (method == null || current.isMoreSpecificThan(method)) {
	  method = current;
	  lastMethodAmbiguous = false;
	} else if (method.isMoreSpecificThan(current)) {
	  lastMethodAmbiguous = false;
	} else {
	  lastMethodAmbiguous = true;
	}
      }
    }

    if (lastMethodAmbiguous) {
      throw new UnpositionedError("method-ambigous", ident);
    }

    return method;
  }

  /**
   * Returns a list of abstract methods
   */
  public CMethod[] getAbstractMethods() {
    Vector v = VectorCache.request(VectorCache.SMALL);
    collectAbstractMethods(v);
    CMethod[] methods = new CMethod[v.size()];
    v.copyInto(methods);
    VectorCache.release(v);
    return methods;
  }

  /**
   * Returns a list of interface methods
   */
  public CMethod[] getInterfaceMethods() {
    Vector v = VectorCache.request(VectorCache.SMALL);
    collectInterfaceMethods(v);
    CMethod[] methods = new CMethod[v.size()];
    v.copyInto(methods);
    VectorCache.release(v);
    return methods;
  }

  /**
   * collectAbstractMethods
   * search for a matching method with the provided type parameters
   * look in parent hierarchy as needed
   * @param	name		method name
   * @param	params		method parameters
   */
  public void collectAbstractMethods(Vector v) {
    for (int i = 0; i < methods.length; i++) {
      CMethod	current = methods[i];

      if (hasFlag(current.getModifiers(), ACC_ABSTRACT)) {
	v.addElement(current);
      }
    }

    if (superClass != null) {
      // java.lang.object
      getSuperClass().collectAbstractMethods(v);
    }
  }

  /**
   * collectInterfaceMethods
   * search for a matching method with the provided type parameters
   * look in parent hierarchy as needed
   * @param	name		method name
   * @param	params		method parameters
   */
  public void collectInterfaceMethods(Vector v) {
    if (hasFlag(getModifiers(), ACC_INTERFACE)) {
      for (int i = 0; i < methods.length; i++) {
	v.addElement(methods[i]);
      }
    }

    for (int i = 0; i < interfaces.length; i++) {
      CClass		parent;
      parent = interfaces[i].getCClass();
      assert(parent != null);
      parent.collectInterfaceMethods(v);
    }
  }

  /**
   * @return the short name of this class
   */
  protected static String getIdent(String qualifiedName) {
    assert(qualifiedName != null);
    String	syntheticName;
    int		cursor = qualifiedName.lastIndexOf('/');
    if (cursor > 0) {
      syntheticName = qualifiedName.substring(qualifiedName.lastIndexOf('/') + 1);
    } else {
      syntheticName = qualifiedName;
    }

    cursor = syntheticName.lastIndexOf('$');
    if (cursor > 0) {
      return syntheticName.substring(syntheticName.lastIndexOf('$') + 1);
    } else {
      return syntheticName;
    }
  }

  /**
   * @return	true if this member is accessible
   */
  public boolean isAccessible(CClass from) {
    if ((from.isDefineInside(this)) || isPublic() || isDefineInside(from)) {
      return true;
    }

    if (from.isInnerClass()) {
      if (isAccessible(from.getOwner())) {
	return true;
      }
    }

    if (!isPrivate() && getPackage() == from.getPackage()) {
      return true;
    }

    if (isProtected() && from.descendsFrom(this)) {
      return true;
    }

    return false;
  }

  /**
   * Add inner fields
   */
  public void addField(CSourceField field) {
    field.setPosition(fields_H.size());
    fields_H.put(field.getIdent(), field);
  }

  // ----------------------------------------------------------------------
  // GENERATE CLASSFILE INFO
  // ----------------------------------------------------------------------

  /**
   * Generate the code in a class file
   * @param	classes		a vector to add inner classes
   */
  public void genClassFile(java.io.File destination) {
    ClassConstant[]		interfaces = new ClassConstant[this.interfaces.length];
    for (int i = 0; i < this.interfaces.length; i++) {
      interfaces[i] = new ClassConstant(this.interfaces[i].getQualifiedName());
    }

    int			count = 0;
    MethodInfo[]	pmethods = new MethodInfo[methods.length];

    for (int i = 0; i < methods.length; i++) {
      pmethods[count++] = methods[i].genMethodInfo();
    }

    FieldInfo[]		pfields = getFieldInfo(fields_H, fields);

    InnerClassInfo[]	inners = new InnerClassInfo[innerClasses.length];
    for (int i = 0; i < inners.length; i++) {
      inners[i] = new InnerClassInfo(innerClasses[i].getCClass().getQualifiedName(),
				     getQualifiedName(),
				     innerClasses[i].getCClass().getIdent(),
				     (short)innerClasses[i].getCClass().getModifiers());
    }

    writeClassfile(new ClassInfo((short)getModifiers(),
				 getQualifiedName(),
				 superClass == null ? null : superClass.getCClass().getQualifiedName(),
				 interfaces,
				 pfields,
				 pmethods,
				 inners,
				 sourceFile,
				 false),
		   destination);
  }

  /**
   * Returns the field infos
   */
  protected abstract FieldInfo[] getFieldInfo(Hashtable fields_H, CField[] fields);

  /**
   * Creates a class file from class info
   *
   */
  protected void writeClassfile(ClassInfo classInfo, File destination) {
    String[]	classPath = Utils.splitQualifiedName(classInfo.getName());

    if (classPath[0] != null) {		// the class is part of a package
      String	classDir = classPath[0].replace('/', File.separatorChar);

      if (destination != null) {
	destination = new File(destination, classDir);
      }
    }

    if (destination != null) {		// check that destination exists or else create it
      if (! destination.exists()) {
	destination.mkdirs();
      }

      if (! destination.isDirectory()) {
	throw new RuntimeException("File " + destination + " is not a directory.");
      }
    }
    File	outputFile = destination == null ?
      new File(classPath[1] + ".class") :
      new File(destination, classPath[1] + ".class");

    try {
      MyOutputStream myos = new MyOutputStream();
      classInfo.write(new DataOutputStream(myos));
      FileOutputStream fos = new FileOutputStream(outputFile);
      fos.write(myos.getBuffer(), 0, myos.size());
      fos.close();
    } catch (at.dms.classfile.ClassFileFormatException e) {
      System.err.println("Format Exception on " + outputFile.getPath() + ": " + e.getMessage());
    } catch (java.io.IOException e) {
      System.err.println("I/O Exception on " + outputFile.getPath() + ": " + e.getMessage());
    }
  }

  static class MyOutputStream extends ByteArrayOutputStream {
    MyOutputStream() {
      buf = test;
    }
    public byte[] getBuffer() {
      return test = buf;
    }
    static byte[]	test;
    static {
      test = new byte[10000];
    }
  }

  // ----------------------------------------------------------------------
  // INNER CLASS SUPPORT
  // ----------------------------------------------------------------------

  /**
   * add synthetic parameters to method def
   */
  public CType[] genConstructorArray(CType[] params) {
    throw new RuntimeException();
  }

  /**
   * add synthetic parameters to method call
   */
  public void genSyntheticParams(CodeSequence code) {
    if (isInnerClass() && !isStatic() && hasOuterThis()) {
      // $$$ hasOuterThis is enougth ?
      code.plantThisLocalVar(opc_aload);
    }
  }

  /**
   * add synthetic parameters to method call
   */
  public void genSyntheticParamsFromExplicitSuper(JExpression[] params, CodeSequence code) {
    CClass		sup = this;
    if (sup.isInnerClass() && !sup.isStatic() && sup.hasOuterThis()) {
      JGeneratedLocalVariable var = new JGeneratedLocalVariable(null, 0, getOwner().getType(), JAV_OUTER_THIS, null);
      var.setPosition(params.length + 1);
      new JLocalVariableExpression(TokenReference.NO_REF, var).genCode(code);
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE INITIALIZATIONS
  // ----------------------------------------------------------------------

  protected static CClass	CLS_UNDEFINED = new CSourceClass(null, TokenReference.NO_REF, 0, "<gen>", "<gen>", false);
  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private	long		timestamp;
  private	int		version;
  private	String		sourceFile;
  private	String		qualifiedName;
  private	String		packageName;
  private	CClassType[]	interfaces;
  private	CClassType	type;
  private	boolean		hasOuterThis;
  private	CField[]	fields;
  private	CMethod[]	methods;

  // May be null if there is less than a certain amount of members
  private	Hashtable	fields_H;
  private	Hashtable	methods_H;

  private	CClassType[]	innerClasses;
  private	CClassType	superClass;
}
