/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CHandLexer.java,v 1.28 1999/10/28 12:45:47 vgp Exp $
 */

package at.dms.kjc;

import java.util.Vector;

import antlr.Token;

import at.dms.compiler.CToken;
import at.dms.compiler.Compiler;
import at.dms.compiler.InputBuffer;
import at.dms.compiler.PositionedError;
import at.dms.compiler.JavaStyleComment;

/**
 * This class describes the capabilities of lexers (token streams)
 * to handle line numbers and comments.
 */
public abstract class CHandLexer extends at.dms.compiler.HandLexer {

  /**
   * Constructs a new hand written lexer
   */
  protected CHandLexer(Compiler compiler, InputBuffer buffer) {
    super(compiler, buffer);

    comments = new Vector();
  }

  /**
   * >>> COMMENT
   */
  protected abstract Token getToken() throws PositionedError;

  // --------------------------------------------------------------------
  // SPECIAL TOKEN SCANNERS
  // --------------------------------------------------------------------
  // NOTE: scanXxx procedures do not return tokens

  /**
   * Scans an escape sequence
   *
   * @return	the scanned character
   */
  protected char scanEscapeSequence() throws PositionedError {
    if (buffer.data[buffer.pos++] != '\\') {
      throw new PositionedError(getTokenReference(), "illegal-char", "\\\\");
    }

    switch (buffer.data[buffer.pos]) {
    case 'b':
      buffer.pos++; return '\b';
    case 't':
      buffer.pos++; return '\t';
    case 'n':
      buffer.pos++; return '\n';
    case 'f':
      buffer.pos++; return '\f';
    case 'r':
      buffer.pos++; return '\r';
    case '\"':
      buffer.pos++; return '\"';
    case '\'':
      buffer.pos++; return '\'';
    case '\\':
      buffer.pos++; return '\\';
    case 'u':
      buffer.pos++;
      return (char)scanHexCharacter();
    default:
      return (char)scanOctCharacter();
    }
/*
    switch (buffer.data[buffer.pos]) {
    case 'b':
    case 't':
    case 'n':
    case 'f':
    case 'r':
    case '\"':
    case '\'':
    case '\\':
      buffer.pos++; return buffer.data[buffer.pos - 1];
    case 'u':
      buffer.pos++;
      return (char)scanHexCharacter();
    default:
      return (char)scanOctCharacter();
    }
*/
  }

  private char scanOctCharacter() throws PositionedError {
    int		val = 0;
    int		i;

    for (i = 0; i < 3 && (Character.digit(buffer.data[buffer.pos], 8) != -1); i++) {
      val = (8*val) + Character.digit(buffer.data[buffer.pos++], 8);
    }

    if (i == 0 || val > 0xFF) {
      throw new PositionedError(getTokenReference(), "illegal-char", "" + buffer.data[buffer.pos]);
    }

    return (char)val;
  }

  private char scanHexCharacter() throws PositionedError {
    int		val = 0;
    int		i;

    for (i = 0; i < 4 && (Character.digit(buffer.data[buffer.pos], 16) != -1); i++) {
      val = (16*val) + Character.digit(buffer.data[buffer.pos++], 16);
    }

    if (i == 0) {
      throw new PositionedError(getTokenReference(), "illegal-char", "" + buffer.data[buffer.pos]);
    }

    return (char)val;
  }

  /**
   * Determines if the specified character is permissible as
   * the first character in a Java identifier.
   *
   * @return	true iff the character may start a Java identifier
   */
  protected static boolean isJavaIdentifierStart(char c) {
    switch (c) {
    case '$':  case 'A':  case 'B':  case 'C':
    case 'D':  case 'E':  case 'F':  case 'G':
    case 'H':  case 'I':  case 'J':  case 'K':
    case 'L':  case 'M':  case 'N':  case 'O':
    case 'P':  case 'Q':  case 'R':  case 'S':
    case 'T':  case 'U':  case 'V':  case 'W':
    case 'X':  case 'Y':  case 'Z':  case '_':
    case 'a':  case 'b':  case 'c':  case 'd':
    case 'e':  case 'f':  case 'g':  case 'h':
    case 'i':  case 'j':  case 'k':  case 'l':
    case 'm':  case 'n':  case 'o':  case 'p':
    case 'q':  case 'r':  case 's':  case 't':
    case 'u':  case 'v':  case 'w':  case 'x':
    case 'y':  case 'z':
      return true;
    }
    return Character.isJavaIdentifierStart(c);
  }

  // --------------------------------------------------------------------
  // COMMENT HANDLING
  // --------------------------------------------------------------------

  /**
   * >>> COMMENT
   */
  protected boolean getJavaComment(boolean store) {
    // line.charAt(line_buffer.pos+0) is '/'
    if (buffer.data[buffer.pos + 1] == '/') {
      // EndOfLineComment
      buffer.pos += 2;

      int	start = buffer.pos;

      while (buffer.data[buffer.pos] != '\n') {
	buffer.pos++;
      }
      
      if (store) {
	addComment(new JavaStyleComment(new String(buffer.data, start, buffer.pos - start),
					true,
					spaceBefore(start),
					spaceAfter(buffer.pos)));
      }
      return true;
    } else if (buffer.data[buffer.pos + 1] == '*') {
      // TraditionalComment or DocumentationComment
      if (buffer.data[buffer.pos + 2] == '*') { 
	// DocumentationComment
	snarfJavaComment(store, true);
      } else { 
	// TraditionalComment
	snarfJavaComment(store, false);
      }
      return true;
    } else {
      // it's a token, not a comment.
      return false;
    }
  }

  private void snarfJavaComment(boolean store, boolean doc) {
    int		start = buffer.pos;

    buffer.pos += 2;
    while (true) { // Grab CommentTail
    loop:
      while (true) {
	switch(buffer.data[buffer.pos]) {
	case '*':
	  break loop;
	case '\n':
	  buffer.line++;
	default:
	  buffer.pos++;
	}
      }

      // At this point, line.charAt(line_buffer.pos)=='*'
      // Grab CommentTailStar starting at line_buffer.pos+1.

      if (buffer.data[buffer.pos + 1]=='/') {
	buffer.pos += 2;

	if (store) {
	  if (doc) {
	    addComment(new JJavadocComment(new String(buffer.data, start + 3, buffer.pos - start - 6),
					   spaceBefore(start),
					   spaceAfter(buffer.pos)));
	  } else {
	    addComment(new JavaStyleComment(new String(buffer.data, start + 2, buffer.pos - start - 4), 
					    false,
					    spaceBefore(start),
					    spaceAfter(buffer.pos)));
	  }
	}
	return;
      }
      buffer.pos++;
    }
  }

  /**
   *
   */
  public JJavadocComment getJavadocComment() {
    for (int i = comments.size() - 1; i >= 0; i--) {
      if (comments.elementAt(i) instanceof JJavadocComment) {
	JJavadocComment		jdoc = (JJavadocComment)comments.elementAt(i);

	comments.setElementAt(null, i);
	return jdoc;
      }
    }
    return null;
  }
}
