/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CMethod.java,v 1.24 1999/09/11 13:46:49 vgp Exp $
 */

package at.dms.kjc;

import java.util.Hashtable;
import java.util.Enumeration;

import at.dms.classfile.MethodInfo;

/**
 * This class represents an exported member of a class (fields)
 */
public class CMethod extends CMember {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Constructs a field export
   * @param	owner		the owner of this method
   * @param	modifiers	the modifiers on this method
   * @param	ident		the ident of this method
   * @param	returnType	the return type of this method
   * @param	parameters	the parameters type of this method
   * @param	exceptions	a list of all exceptions of the throws list
   * @param	javadoc	is this method deprecated
   */
  public CMethod(CClass owner,
		 int modifiers,
		 String ident,
		 CType returnType,
		 CType[] parameters,
		 CClassType[] exceptions,
		 boolean deprecated) {
    super(owner, modifiers, ident,deprecated);

    this.returnType = returnType;
    this.parameters = parameters;
    this.exceptions = exceptions;
  }

  /**
   * Constructs a field export
   * @param	owner		the owner of this method
   * @param	methodInfo	a method info from a class file
   */
  public CMethod(CClass owner, MethodInfo methodInfo) {
    super(owner,
	  methodInfo.getModifiers(),
	  methodInfo.getName(),
	  methodInfo.isDeprecated());

    CType[] types = CType.parseMethodSignature(methodInfo.getSignature());

    parameters = new CType[types.length - 1];
    for (int i = 0; i < parameters.length; i++) {
      parameters[i] = types[i];
    }

    returnType = types[types.length - 1];

    String[]	exceptions = methodInfo.getExceptions();
    if (exceptions != null) {
      this.exceptions = new CClassType[exceptions.length];
      for (int i = 0; i < exceptions.length; i++) {
	this.exceptions[i] = CClassType.lookup(exceptions[i]);
      }
    } else {
      this.exceptions = new CClassType[0];
    }
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * @return	the interface
   */
  public CMethod getMethod() {
    return this;
  }

  /**
   * @return the type of this field
   */
  public CType getReturnType() {
    return returnType;
  }

  /**
   * @return the type of this field
   */
  public CType[] getParameters() {
    return parameters;
  }

  /**
   * @return the type of this field
   */
  public String getSignature() {
    return CType.genMethodSignature(returnType, parameters);
  }

  /**
   * @return	the exceptions that can be thrown by this method
   */
  public CClassType[] getThrowables() {
    return exceptions;
  }

  /**
   * This method is used by initializers that knows throwables exceptions only
   * after body was checked.
   * @param	throwables	the exceptions that can be thrown by this method
   */
  public void setThrowables(Hashtable throwables) {
    if (throwables != null) {
      Enumeration	enum = throwables.elements();
      int		count = 0;

      exceptions = new CClassType[throwables.size()];
      while (enum.hasMoreElements()) {
	exceptions[count++] = ((CThrowableInfo)enum).getThrowable();
      }
    }
  }

  // ----------------------------------------------------------------------
  // ACCESSORS (QUICK)
  // ----------------------------------------------------------------------

  /**
   * @return	true if this member is native
   */
  public boolean isNative() {
    return hasFlag(getModifiers(), ACC_NATIVE);
  }

  /**
   * @return	true if this member is abstract
   */
  public boolean isAbstract() {
    return hasFlag(getModifiers(), ACC_ABSTRACT);
  }

  /**
   *
   */
  public boolean isConstructor() {
    return getIdent() == JAV_CONSTRUCTOR;
  }

  // ----------------------------------------------------------------------
  // CHECK MATCHING
  // ----------------------------------------------------------------------

  /**
   * equals
   * search if two methods has identique signature
   * @param	params		an other method
   */
  public boolean equals(CMethod method) {
    if (parameters.length != method.parameters.length) {
      return false;
    }

    for (int i = 0; i < parameters.length; i++) {
      if (!parameters[i].equals(method.parameters[i])) {
	return false;
      }
    }
    if (getIdent() != method.getIdent()) {
      return false;
    }

    return true;
  }
  
  /**
   * isMatching
   * search if this matches with the provided type parameters
   * @param	params		method parameters
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public boolean isMatching(String ident, CType[] params) {
    return parameters.length == params.length &&
      ident == getIdent() &&      
      isMatchingParams(params);
  }

  /**
   * isMatching
   * search if this matches with the provided type parameters
   * @param	params		method parameters
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public boolean isMatchingParams(CType[] params) {
    CType[]	parameters = getParameters();

    for (int i = 0; i < params.length; i++) {

      if (parameters[i] == CStdType.Char && params[i] != CStdType.Char) {
	return false;
      }

      if (!params[i].implicitlyCastTo(parameters[i])) {
	return false;
      }
    }

    return true;
  }

  /**
   * isMoreSpecificThan
   * search if this method is more specific than other
   * @param	other		the method to compare to
   * @return	true if this method is more specific than other
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public boolean isMoreSpecificThan(CMethod other) {
    CType[]	parameters = getParameters();
    CType[]	otherP = other.getParameters();

    for (int i = 0; i < otherP.length; i++) {
      if (!parameters[i].implicitlyCastTo(otherP[i])) {
	return false;
      }
    }

    return true;
  }

  public String toString() {
    return getIdent() + "(" + parameters.length + ")";
  }

  // ----------------------------------------------------------------------
  // GENERATE CLASSFILE INFO
  // ----------------------------------------------------------------------

  /**
   * Generate the code in a class file
   */
  public MethodInfo genMethodInfo() {
    fail();
    String[]	exceptions = new String[this.exceptions.length];
    for (int i = 0; i < this.exceptions.length; i++) {
      exceptions[i] = this.exceptions[i].getSignature();
    }

    return new MethodInfo((short)getModifiers(),
			  getIdent(),
			  getSignature(),
			  exceptions,
			  null,
			  isDeprecated(),
			  false);
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private	CType			returnType;
  private	CType[]			parameters;
  private	CClassType[]		exceptions;
}
