/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CType.java,v 1.29 1999/10/08 11:18:58 vgp Exp $
 */

package at.dms.kjc;

import java.util.Stack;
import java.util.Vector;

import at.dms.compiler.UnpositionedError;
import at.dms.compiler.FastStringBuffer;

/**
 * Root for type hierarchy
 */
public abstract class CType extends at.dms.compiler.Utils implements Constants {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Constructs a type signature
   */
  protected CType(int type) {
    this.type = type;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * equals
   */
  public boolean equals(CType other) {
    return this == other;
  }

  /**
   * Transforms this type to a string
   */
  public abstract String toString();

  /**
   * Transforms this type to a string
   * @return	the class file form of thid type
   */
  public String getSignature() {
    FastStringBuffer buff = StringBuffers.getStringBuffer();

    appendSignature(buff);
    String s = buff.toString();
    StringBuffers.release(buff);

    return s;
  }

  /**
   * Transforms this type to a string
   */
  protected abstract void appendSignature(FastStringBuffer buff);

  /**
   * Returns the size used in stack by value of this type
   */
  public abstract int getSize();

  /**
   * Check if a type is a numeric type
   * @return is it a numeric type ?
   */
  public boolean isNumeric() {
    return false;
  }

  public boolean isCodedHasInt() {
    return false;
  }

  /**
   * Check if a type is an integer type
   * @return is it a integer type ?
   */
  public boolean isOrdinal() {
    return false;
  }

  /**
   * Check if a type is an integer type
   * @return is it a real number type ?
   */
  public boolean isFloatingPoint() {
    return false;
  }

  /**
   * Check if a type is a class type
   * @return is it a subtype of ClassType ?
   */
  public boolean isPrimitive() {
    return isNumeric() || (this == CStdType.Boolean);
  }

  /**
   * Check if a type is a class type
   * @return is it a subtype of ClassType ?
   */
  public boolean isReference() {
    return false;
  }

  /**
   * Check if a type is a class type
   * @return is it a subtype of ClassType ?
   */
  public boolean isClassType() {
    return false;
  }

  /**
   * @return is this type an array ?
   */
  public boolean isArrayType() {
    return false;
  }

  /**
   * @return true if this type is valid
   */
  public boolean checked() {
    return true;
  }

  /**
   * Returns the ID of this type
   */
  public int getTypeID() {
    return type;
  }

  // ----------------------------------------------------------------------
  // INTERFACE CHECKING
  // ----------------------------------------------------------------------

  /**
   * check that type is valid
   * necessary to resolve String into java/lang/String
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public abstract void checkType(CContext context) throws UnpositionedError;

  /**
   * isCastableTo
   * @param dest  the expression type to cast to
   * @return  cast is valid ?
   */
  public abstract boolean isCastableTo(CType dest);

  /**
   * implicitlyCastTo
   * @param	dest	the expression type to cast to
   * @return	cast is valid ?
   */
  public boolean implicitlyCastTo(CType dest) {
    return dest.equals(this);
  }

  /**
   * isAssignableTo
   * @param dest  the destination type
   * @return  assignment is valid ?
   */
  public abstract boolean isAssignableTo(CType dest);

  /**
   * @return	true if this type corrspond to a checked exception
   */
  public boolean isCheckedException() {
    return false;
  }

  /**
   * @return the object class of this type
   */
  public CClass getCClass() {
    fail();
    return null;
  }

  // ----------------------------------------------------------------------
  // UTILITIES
  // ----------------------------------------------------------------------

  /**
   * Parse a java type signature
   *  Description : Attempts to parse the provided string as if it started with
   *    the Java VM-standard signature for a type.
   */
  public static CType parseSignature(String signature) {
    return instance.parseSignature(signature);
  }

  /**
   * Returns an array of types represented by the type signature
   * For methods, the return type is the last element of the array
   */
  public static CType[] parseMethodSignature(String sig) {
    return instance.parseMethodSignature(sig);
  }

  /**
   * generate a method signature
   */
  public static String genMethodSignature(CType returnType, CType[] parameters) {
    FastStringBuffer buff = StringBuffers.getStringBuffer();

    buff.append('(');

    for (int i = 0; i < parameters.length; i++) {
      parameters[i].appendSignature(buff);
    }
    buff.append(')');

    returnType.appendSignature(buff);

    String s = buff.toString();
    StringBuffers.release(buff);
    return s;
  }

  // optimization
  public static class StringBuffers {
    static FastStringBuffer getStringBuffer() {
      if (!ENV_USE_CACHE || stack.empty()) {
	return new FastStringBuffer();
      }
      FastStringBuffer buff = (FastStringBuffer)stack.pop();
      buff.reset();

      return buff;
    }

    static void release(FastStringBuffer buff) {
      if (ENV_USE_CACHE) {
	stack.push(buff);
      }
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private static Stack	stack = new Stack();
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  protected int	type;

  public static class MethodSignatureParser {
    /**
     * Parse a java type signature
     *  Description : Attempts to parse the provided string as if it started with
     *    the Java VM-standard signature for a type.
     */
    public final CType parseSignature(String signature) {
      return parseSignature(signature, 0, signature.length());
    }

    public CType parseSignature(String signature, int from, int to) {
      CType	type;
      int	cursor = from;
      while (signature.charAt(cursor) == '[') {
	cursor++;
      }

      switch (signature.charAt(cursor)) {
      case 'V':
	type = CStdType.Void;
	break;
      case 'B':
	type = CStdType.Byte;
	break;
      case 'C':
	type = CStdType.Char;
	break;
      case 'D':
	type = CStdType.Double;
	break;
      case 'F':
	type = CStdType.Float;
	break;
      case 'I':
	type = CStdType.Integer;
	break;
      case 'J':
	type = CStdType.Long;
	break;
      case 'L':
	type = CClassType.lookup(signature.substring(cursor + 1, to - 1));
	break;
      case 'S':
	type = CStdType.Short;
	break;
      case 'Z':
	type = CStdType.Boolean;
	break;
      default:
	fail();
	return null;
      }

      cursor -= from;
      return cursor > 0 ? new CArrayType(type, cursor) : type;
    }

    /**
     * Returns an array of types represented by the type signature
     * For methods, the return type is the last element of the array
     */
    public synchronized CType[] parseMethodSignature(String signature) {
      // assert(sig.charAt(0) == '(');
      char[]	sig = signature.toCharArray();
      int	current = 1;

      while (sig[current] != ')') {
	int	end = current;
	while (sig[end] == '[') {
	  end += 1;
	}
	if (sig[end] != 'L') {
	  end += 1;
	} else {
	  while (sig[end] != ';') {
	    end += 1;
	  }
	  end += 1;
	}
	vec.addElement(parseSignature(signature, current, end));
	current = end;
      }
      vec.addElement(parseSignature(signature, current + 1, signature.length()));

      // store into array
      CType[]	arr = new CType[vec.size()];

      vec.copyInto(arr);
      vec.setSize(0);

      return arr;
    }
    Vector	vec = new Vector();
  }

  // !!! remove
  protected static MethodSignatureParser instance = new MethodSignatureParser();
}
