/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JAddExpression.java,v 1.24 1999/10/08 11:18:59 vgp Exp $
 */

package at.dms.kjc;

import at.dms.classfile.MethodRefInstruction;
import at.dms.classfile.ClassRefInstruction;
import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * This class implements '+ - * /' specific operations
 * Plus operand may be String, numbers
 */
public class JAddExpression extends JBinaryArithmeticExpression {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * This method is directly called by the parser
   * @param	where		the line of this node in the source code
   * @param	left		the left operand
   * @param	right		the right operand
   */
  public JAddExpression(TokenReference where,
			JExpression left,
			JExpression right) {
    super(where, left, right);
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public JExpression checkExpression(CExpressionContext context) throws PositionedError {
    left = left.checkExpression(context);
    right = right.checkExpression(context);

    check(context, left.getType() != CStdType.Void && right.getType() != CStdType.Void,
	  "add-badtype", left.getType(), right.getType());

    JExpression expr = checkBinaryBody(context);
    if ((expr == this) && !type.equals(CStdType.String)) {
      left = left.convertType(type, context);
      right = right.convertType(type, context);
    }
    return expr;
  }

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	UnpositionedError Error catched as soon as possible
   */
  public JExpression checkBinaryBody(CExpressionContext context) throws PositionedError {
    try {
      type = computeType(left.getType(), right.getType());
    } catch (UnpositionedError e) {
      throw e.getError(getTokenReference());
    }

    // no cst fold on string to save constant pool size ???
    if (left.isConstant() && right.isConstant() && !type.equals(CStdType.String)) {
      return constantFolding(left.getNumberLiteral(), right.getNumberLiteral());
    } else {
      return this;
    }
  }

  /**
   * compute the type of this expression according to operands
   * @param	leftType		the type of left operand
   * @param	rightType		the type of right operand
   * @return	the type computed for this binary operation
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public static CType computeType(CType	leftType, CType rightType) throws UnpositionedError {
    if (leftType.equals(CStdType.String) || rightType.equals(CStdType.String)) {
      return CStdType.String;
    }

    if (leftType.isNumeric() && rightType.isNumeric()) {
      return CNumericType.binaryPromote(leftType, rightType);
    }

    throw new UnpositionedError("add-badtype", leftType, rightType);
  }

  /**
   * @param	left		the left value
   * @param	right		the right value
   * @return	a literal resulting of an operation over two literals
   */
  public JExpression constantFolding(JNumberLiteral left,
				     JNumberLiteral right) {
    if (type == CStdType.Double) {
      return new JRealLiteral(getTokenReference(),
			      new Double(left.numberValue().doubleValue() +
					 right.numberValue().doubleValue()),
			      CStdType.Double);
    } else if (type == CStdType.Float) {
      return new JRealLiteral(getTokenReference(),
			      new Float(left.numberValue().floatValue() +
					right.numberValue().floatValue()),
			      CStdType.Float);
    } else if (type == CStdType.Long) {
      return new JOrdinalLiteral(getTokenReference(),
				 new Long(left.numberValue().longValue() +
					  right.numberValue().longValue()));
    } else if (type.implicitlyCastTo(CStdType.Integer)) {
      return new JOrdinalLiteral(getTokenReference(),
				 new Integer(left.numberValue().intValue() +
					     right.numberValue().intValue()));
    } else {
      fail();
    }

    return left;
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Generate the code in pure java form
   * It is usefull to debug and tune compilation process
   * @param p the printwriter into the code is generated
   */
  public void genJavaCode(CJavaPrettyPrinter p) {
    p.printBinaryExpression("+", left, right);
  }

  /**
   * @param	type		the type of result
   * @return	the type of opcode for this operation
   */
  public static int getOpcode(CType type) {
    switch (type.getTypeID()) {
    case TID_FLOAT:
      return opc_fadd;
    case TID_LONG:
      return opc_ladd;
    case TID_DOUBLE:
      return opc_dadd;
    default:
      return opc_iadd;
    }
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    if (type.equals(CStdType.String)) {
      code.plantInstruction(new ClassRefInstruction(opc_new, "java/lang/StringBuffer"));
      code.plantInstruction(new NoArgInstruction(opc_dup));
      code.plantInstruction(new MethodRefInstruction(opc_invokespecial,
						     "java/lang/StringBuffer",
						     JAV_CONSTRUCTOR,
						     "()V"));
      appendToStringBuffer(code, left);
      appendToStringBuffer(code, right);
      code.plantInstruction(new MethodRefInstruction(opc_invokevirtual,
						     "java/lang/StringBuffer",
						     "toString",
						     "()Ljava/lang/String;"));
    } else {
      left.genCode(code);
      right.genCode(code);

      code.plantInstruction(new NoArgInstruction(getOpcode(getType())));
    }
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  private void appendToStringBuffer(CodeSequence code, JExpression expr) {
    if ((expr instanceof JAddExpression) && (expr.getType().equals(CStdType.String))) {
      ((JAddExpression)expr).appendToStringBuffer(code, ((JAddExpression)expr).left);
      ((JAddExpression)expr).appendToStringBuffer(code, ((JAddExpression)expr).right);
    } else {
      expr.genCode(code);

      CType	type = expr.getType();
      if (!type.isReference() || type.equals(CStdType.String) || type.equals(CStdType.CharArray)) {
	if (type == CStdType.Byte || type == CStdType.Short) {
	  type = CStdType.Integer;
	}

	code.plantInstruction(new MethodRefInstruction(opc_invokevirtual,
						       "java/lang/StringBuffer",
						       "append",
						       "(" + type.getSignature() + ")Ljava/lang/StringBuffer;"));
      } else {
	code.plantInstruction(new MethodRefInstruction(opc_invokevirtual,
						       "java/lang/StringBuffer",
						       "append",
						       "(Ljava/lang/Object;)Ljava/lang/StringBuffer;"));
      }
    }
  }
}
