/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JCharLiteral.java,v 1.21 1999/10/08 11:18:59 vgp Exp $
 */

package at.dms.kjc;

import at.dms.compiler.NumberParser;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * A simple character constant
 */
public class JCharLiteral extends JOrdinalLiteral {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   * @param	image		the string representation of this literal
   */
  public JCharLiteral(TokenReference where, String image) {
    super(where, image);
  }

  /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   * @param	image		the string representation of this literal
   */
  public JCharLiteral(TokenReference where, char val) {
    super(where, "");

    value = new Character(val);
  }

  /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   * @param	image		the string representation of this literal
   */
  public JCharLiteral(TokenReference where, Object val) {
    super(where, "");
    if (val instanceof Character) {
      value = (Character)val;
    } else {
      value = new Character((char)((Number)val).intValue());
    }
  }

  /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   * @param	image		the string representation of this literal
   * @param	quoted		there is quote around image
   */
  public JCharLiteral(TokenReference where, String image, boolean quoted) {
    super(where, image.substring(1, image.length() - 1));
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * Compute the type of this expression (called after parsing)
   * @return the type of this expression
   */
  public CType getType() {
    return CStdType.Char;
  }

  /**
   * get a Number that represents the value of this literal
   * Do not call before checkBody
   */
  public Number numberValue() {
    return new Integer(value.charValue());
  }

  /**
   * get a Number that represents the value of this literal
   * Do not call before checkBody
   */
  public Object getValue() {
    return value;
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   */
  public JExpression checkExpression(CExpressionContext context) throws PositionedError {
    if (image == null) {
      return this;
    }
    if (image.startsWith("\\u")) {
      value = new Character((char)NumberParser.decodeHexInt(false, image.substring(2)));
    } else if (image.startsWith("\\") && image.length() > 1) {
      if ((image.charAt(1) >= '0') && (image.charAt(1) <= '9')) {
	value = new Character((char)NumberParser.decodeOctInt(false, image.substring(1)));
	check(context, value.charValue() <= 0377,
	      "invalid-octal-char", image);
      } else if (image.equals("\\b")) {
	value = new Character('\b');
      } else if (image.equals("\\r")) {
	value = new Character('\r');
      } else if (image.equals("\\t")) {
	value = new Character('\t');
      } else if (image.equals("\\n")) {
	value = new Character('\n');
      } else if (image.equals("\\f")) {
	value = new Character('\f');
      } else if (image.equals("\\\"")) {
	value = new Character('\"');
      } else if (image.equals("\\\'")) {
	value = new Character('\'');
      } else if (image.equals("\\\\")) {
	value = new Character('\\');
      } else {
	check(context, false, "invalid-escape-sequence", image);
      }
    } else {
      value = new Character(image.charAt(0));
    }

    image = null;
    return this;
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Generate the code in pure java form
   * It is usefull to debug and tune compilation process
   * @param p the printwriter into the code is generated
   */
  public void genJavaCode(CJavaPrettyPrinter p) {
    if (value != null) {
      p.printCharLiteral(value);
    } else {
      p.print("'" + image + "'");
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private	Character		value;
}
