/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JClassFieldExpression.java,v 1.46 1999/10/08 11:18:59 vgp Exp $
 */

package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.classfile.FieldRefInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * a field access ('getExpression().t' or 't')
 */
public class JClassFieldExpression extends JExpression implements CInitializable {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   */
  public JClassFieldExpression(TokenReference where, String ident) {
    super(where);

    this.ident = ident;
  }

  /**
   * Construct a node in the parsing tree
   * @param where the line of this node in the source code
   */
  public JClassFieldExpression(TokenReference where,
			       JExpression left,
			       String ident) {
    this(where, ident);

    this.left = left;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  public String getIdent() {
    return ident;
  }

  /**
   * Compute the type of this expression (called after parsing)
   * @return the type of this expression
   */
  public CType getType() {
    return field.getType();
  }

  /**
   * @return is this expression constant ?
   */
  public boolean isConstant() {
    return field.isFinal() && field.getValue() != null && field.getValue().isConstant();
  }

  /**
   * Returns true if this field accept assignment
   */
  public boolean isLValue(CExpressionContext context) {
    if (!field.isFinal() || !(field instanceof CSourceField)) {
      return true;
    }

    if (context.getBodyContext().getClassContext().getCClass() == field.getOwner() && !((CSourceField)field).isFullyDeclared()) {
      return !CVariableInfo.mayBeInitialized(context.getBodyContext().getFieldInfo(((CSourceField)field).getPosition()));
    }

    return true;
  }

  /**
   * Returns true if this field is already initialized
   */
  public boolean isInitialized(CExpressionContext context) {
    if (!(field instanceof CSourceField) || field.isStatic()) {
      return true;
    }
    if (context.getBodyContext().getClassContext().getCClass() == field.getOwner() &&
	!((CSourceField)field).isFullyDeclared()) {
      return CVariableInfo.isInitialized(context.getBodyContext().getFieldInfo(((CSourceField)field).getPosition()));
    }

    return true;
  }

  /**
   * Returns true if this field is already initialized
   */
  public void initialize(CBodyContext context) {
    if ((field instanceof CSourceField) && (context.getClassContext().getCClass() == field.getOwner() && !((CSourceField)field).isFullyDeclared())) {
      context.setFieldInfo(((CSourceField)field).getPosition(), CVariableInfo.INF_INITIALIZED + CVariableInfo.INF_MAYBE_INITIALIZED);
    }
  }

  public CField getField() {
    return field;
  }

  /**
   * Returns the literal value of this field
   */
  public JLiteral getLiteral() {
    return (JLiteral)field.getValue();
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public JExpression checkExpression(CExpressionContext context) throws PositionedError {
    CClass	local = context.getClassContext().getCClass();

    findPrefix(local, context);
    checkAccess(local, context);

    if (field instanceof CSourceField) {
      ((CSourceField)field).use();
    }

    if (isConstant()) {
      return field.getValue();
    }

    return this;
  }

  /**
   * Find the type of the prefix
   *
   * @exception	PositionedError	Error catched as soon as possible
   * @exception	UnpositionedError	Error catched as soon as possible
   */
  protected void findPrefix(CClass local, CExpressionContext context) throws PositionedError {
    if (left != null) {
      left = left.checkExpression(context);
      check(context, left.getType().isClassType(),
	    "field-badaccess", left.getType());
    } else {
      try {
	field = context.lookupField(ident);
	if (field == null) {
	  field = context.getClassContext().lookupOuterField(ident); // $$$ Why searching again
	}
      } catch (UnpositionedError e) {
	throw e.getError(getTokenReference());
      }
      check(context, field != null, "field-unknown", ident);

      if (local.descendsFrom(field.getOwner())) {
	if (!field.isStatic()) {
	  left = new JThisExpression(getTokenReference(), field.getOwner()).checkExpression(context);
	} else {
	  left = new JTypeNameExpression(getTokenReference(), field.getOwner().getType());
	}
      } else { // access to outer class
	if (!field.getOwner().isStatic()) {
	  if (!field.isStatic()) {
	    check(context, !local.isStatic(), "field-staticerr", ident);
	    left = new JThisExpression(getTokenReference(), field.getOwner()).checkExpression(context);
	  } else {
	    left = new JTypeNameExpression(getTokenReference(), field.getOwner().getType());
	  }
	  left = left.checkExpression(context);
	} else {
	  left = new JTypeNameExpression(getTokenReference(), field.getOwner().getType());
	  left = left.checkExpression(context);
	}
      }
    }
  }

  /**
   * Checks is access to prefix is okay
   *
   * @exception	PositionedError	Error catched as soon as possible
   * @exception	UnpositionedError	Error catched as soon as possible
   */
  public void checkAccess(CClass local, CExpressionContext context) throws PositionedError {
    CClass	access = left.getType().getCClass();
    try {
      field = access.lookupField(ident);
    } catch (UnpositionedError e) {
      throw e.getError(getTokenReference());
    }
    check(context, field != null, "field-unknown", ident);

    access = field.getOwner();

    check(context, field.isAccessible(local), "field-noaccess", ident);
    check(context, field.getOwner().isAccessible(local), "class-noaccess", field.getOwner().getIdent());

    if ((context.getMethodContext() instanceof CConstructorContext) &&
	(left instanceof JThisExpression) &&
	!field.isStatic()) {
      if (field.getType().isClassType() &&
	  field.getType().getCClass() == context.getClassContext().getCClass()) {
	check(context, ((CConstructorContext)context.getMethodContext()).isSuperConstructorCalled(),
	      "constructor-explicit-call", field.getIdent());
      }
    }

    if (access == local) {
      /* !!!
      boolean	declared = !((field instanceof CSourceField) && !((CSourceField)field).isFullyDeclared());
      check(context, declared || (left != null && !(left instanceof JThisExpression)),
	    "field-not-declared-yet", field.getIdent());
      */
    }
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Generate the code in pure java form
   * It is usefull to debug and tune compilation process
   * @param p the printwriter into the code is generated
   */
  public void genJavaCode(CJavaPrettyPrinter p) {
    p.printFieldExpression(left, getIdent());
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());
    boolean		discardValue = code.discardValue();

    if (field.isStatic()) {
      code.setDiscardValue(true);
      if (left != null) {
	left.genCode(code);
      }

      if (!discardValue) {
	code.plantInstruction(new FieldRefInstruction(opc_getstatic,
						      field.getPrefixName(),
						      field.getIdent(),
						      field.getType().getSignature()));
      }
    } else {
      code.setDiscardValue(discardValue);
      left.genCode(code);
      if (!discardValue) {
	code.plantInstruction(new FieldRefInstruction(opc_getfield,
						      field.getPrefixName(),
						      field.getIdent(),
						      field.getType().getSignature()));
      }
    }
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genStartStoreCode(CodeSequence code) {
    if (!field.isStatic()) {
      left.genCode(code);
    }
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genEndStoreCode(CodeSequence code, boolean discardValue) {
    if (!discardValue) {
      if (field.isStatic()) {
	if (getType().getSize() == 2) {
	  code.plantInstruction(new NoArgInstruction(opc_dup2));
	} else {
	  code.plantInstruction(new NoArgInstruction(opc_dup));
	}
      } else {
	if (getType().getSize() == 2) {
	  code.plantInstruction(new NoArgInstruction(opc_dup2_x1));
	} else {
	  code.plantInstruction(new NoArgInstruction(opc_dup_x1));
	}
      }
    }

    if (field.isStatic()) {
      //if (!(left instanceof JTypeNameExpression))
      //code.plantInstruction(new NoArgInstruction(opc_pop));

      code.plantInstruction(new FieldRefInstruction(opc_putstatic,
						    field.getPrefixName(),
						    field.getIdent(),
						    field.getType().getSignature()));
    } else {
      code.plantInstruction(new FieldRefInstruction(opc_putfield,
						    field.getPrefixName(),
						    field.getIdent(),
						    field.getType().getSignature()));
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  protected	String		ident;
  protected	CField		field;
  protected	JExpression	left;
}
