/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JCompilationUnit.java,v 1.40 1999/10/21 12:23:09 vgp Exp $
 */

package at.dms.kjc;

import java.util.Hashtable;
import java.util.Vector;
import java.io.File;

import at.dms.compiler.CWarning;
import at.dms.compiler.Compiler;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * This class represents a virtual file and is the main entry point in java grammar
 */
public class JCompilationUnit extends JPhylum {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Constructs a CompilationUnit with the specified top level context
   * @param	where		the position of this token
   */
  public JCompilationUnit(TokenReference where,
			  JPackageName packageName,
			  JPackageImport[] importedPackages,
			  JClassImport[] importedClasses,
			  JTypeDeclaration[] typeDeclarations) {
    super(where);
    this.packageName = packageName;
    this.importedPackages = importedPackages;
    this.importedClasses = importedClasses;
    this.typeDeclarations = typeDeclarations;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * @return the package name of this compilation unit
   */
  public String getPackageName() {
    assert(packageName != null);
    return packageName.getName();
  }

  /**
   * @return	the name of the file associated with this compilation unit
   */
  public String getFileName() {
    return getTokenReference().getFile();
  }

  // ----------------------------------------------------------------------
  // INTERFACE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Second pass (quick), check interface looks good
   * @exception	PositionedError	an error with reference to the source file
   */
  public void checkInterface(Compiler compiler) throws PositionedError {
    // 7.5.1 Single-Type-Import Declaration
    for (int i = 0; i < importedClasses.length; i++) {
      JClassImport	ic = importedClasses[i];

      if (! CTopLevel.hasClassFile(ic.getClassName())) {
	compiler.reportTrouble(new PositionedError(getTokenReference(),
						   "class-unknown",
						   ic.getClassName()));
      }

      CClassType	impClass = CClassType.lookup(ic.getClassName());

      Object		clazz = allLoadedClasses.put(ic.getClassIdent(), impClass);
      if (clazz != null) {
	if (impClass != clazz) {
	  compiler.reportTrouble(new CWarning(getTokenReference(), "cunit-rename", impClass.getIdent()));
	}
      }
    }

    // 7.5.2 Type-Import-on-Demand Declaration

    // check uniquness of classes
    for (int i = 0; i < typeDeclarations.length ; i++) {
      CClass	object = typeDeclarations[i].getCClass();

      Object		clazz = allLoadedClasses.get(object.getIdent());

      if (clazz == null) {
	allLoadedClasses.put(object.getIdent(), object.getType());
      } else {
	if (clazz != object) {
	  compiler.reportTrouble(new PositionedError(getTokenReference(),
						     "cunit-rename",
						     object.getQualifiedName()));
	}
      }
    }

    export = new CCompilationUnit(packageName.getName(),
				  importedClasses,
				  importedPackages,
				  allLoadedClasses);

    CCompilationUnitContext	context = new CCompilationUnitContext(compiler, export);
    for (int i = 0; i < typeDeclarations.length ; i++) {
      typeDeclarations[i].checkInterface(context);
    }
  }

  /**
   * Second pass (quick), check interface looks good
   * Exceptions are not allowed here, this pass is just a tuning
   * pass in order to create informations about exported elements
   * such as Classes, Interfaces, Methods, Constructors and Fields
   * @return	true iff sub tree is correct enought to check code
   * @exception	PositionedError	an error with reference to the source file
   */
  public void checkInitializers(Compiler compiler, Vector classes) throws PositionedError {
    CCompilationUnitContext	context = new CCompilationUnitContext(compiler, export, classes);
    for (int i = 0; i < typeDeclarations.length ; i++) {
      typeDeclarations[i].checkInitializers(context);
    }
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @exception	PositionedError Error catched as soon as possible
   */
  public void checkBody(Compiler compiler, Vector classes) throws PositionedError {
    CCompilationUnitContext	context = new CCompilationUnitContext(compiler, export, classes);
    for (int i = 0; i < typeDeclarations.length ; i++) {
      typeDeclarations[i].checkTypeBody(context);
    }

    // Check for unused class imports
    for (int i = 0; i < importedClasses.length; i++) {
      importedClasses[i].checkBody(compiler);
    }
    // Check for unused package imports
    for (int i = 0; i < importedPackages.length; i++) {
      importedPackages[i].checkBody(compiler);
    }
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Generate the code in pure java form
   * It is usefull to debug and tune compilation process
   */
  public void genJavaCode(String destination) {
    CJavaPrettyPrinter  pp;

    if (destination == null || destination.equals("")) {
      genJavaCode(pp = Main.getPrettyPrinter(getTokenReference().getName() + ".gen"));
    } else {
      genJavaCode(pp = Main.getPrettyPrinter(destination + File.separatorChar + getTokenReference().getName()));
    }
    pp.close();
  }

  /**
   * Generate the code in pure java form
   * It is usefull to debug and tune compilation process
   * @param	p		the printwriter into the code is generated
   */
  public void genJavaCode(CJavaPrettyPrinter p) {
    p.printCompilationUnit(packageName, importedPackages, importedClasses, typeDeclarations);
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private JPackageName			packageName;
  private JClassImport[]		importedClasses;
  private JPackageImport[]		importedPackages;
  private JTypeDeclaration[]		typeDeclarations;

  private Hashtable			allLoadedClasses = new Hashtable(); // $$$ DEFAULT VALUE IS OKAY ???
  private CCompilationUnit		export;
}
