/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JCompoundAssignmentExpression.java,v 1.20 1999/10/08 11:18:59 vgp Exp $
 */

package at.dms.kjc;

import at.dms.classfile.MethodRefInstruction;
import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * This class implements the assignment operation
 */
public class JCompoundAssignmentExpression extends JAssignmentExpression {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * This method is directly called by the parser
   * @param	where		the line of this node in the source code
   * @param	oper		the assignment operator
   * @param	left		the left operand
   * @param	right		the right operand
   */
  public JCompoundAssignmentExpression(TokenReference where,
				       int oper,
				       JExpression left,
				       JExpression right) {
    super(where, left, right);
    this.oper = oper;
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * 15.25.2 Compound Assignment Operators
   * // x += 3;   ===>   x = (x.getType()) (x + 3)
   * @param	context		the actual context of analyse
   * @return	a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public JExpression checkExpression(CExpressionContext context) throws PositionedError {
    context.setLeftSide(true);
    left = left.checkExpression(context);
    check(context, left instanceof CInitializable &&
	  ((CInitializable)left).isLValue(context), "assignment-notlvalue");

    context.setLeftSide(false);

    try {
      ((CInitializable)left).initialize(context.getBodyContext());
    } catch (UnpositionedError e) {
      throw e.getError(getTokenReference());
    }

    right = right.checkExpression(context);
    boolean convertRight = true;

    try {
      switch (oper) {
      case OPE_STAR:
	type = JMultExpression.computeType(left.getType(), right.getType());
	break;
      case OPE_SLASH:
	type = JDivideExpression.computeType(left.getType(), right.getType());
	break;
      case OPE_PERCENT:
	type = JModuloExpression.computeType(left.getType(), right.getType());
	break;
      case OPE_PLUS:
	type = JAddExpression.computeType(left.getType(), right.getType());
	break;
      case OPE_MINUS:
	type = JMinusExpression.computeType(left.getType(), right.getType());
	break;
      case OPE_SL:
      case OPE_SR:
      case OPE_BSR:
	type = JShiftExpression.computeType(left.getType(), right.getType());
	convertRight = false;
	break;
      case OPE_BAND:
      case OPE_BXOR:
      case OPE_BOR:
	type = JBitwiseExpression.computeType(left.getType(), right.getType());
	break;
      }
    } catch (UnpositionedError e) {
      throw e.getError(getTokenReference());
    }
    check(context, type.isCastableTo(left.getType()), "assignment-badtype");

    type = left.getType();

    if (convertRight) {
      right = right.convertType(type, context);
    }

    return this;
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Generate the code in pure java form
   * It is usefull to debug and tune compilation process
   * @param	p		the printwriter into the code is generated
   */
  public void genJavaCode(CJavaPrettyPrinter p) {
    p.printCompoundAssignmentExpression(oper, left, right);
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    boolean	discardValue = code.discardValue();
    code.setLineNumber(getTokenReference().getLine());

    code.setDiscardValue(false);

    ((CInitializable)left).genStartStoreCode(code);

    code.setDiscardValue(false);

    if (oper == OPE_PLUS && type.equals(CStdType.String)) {
      left.genCode(code);
      right.genCode(code);
      if (!right.getType().equals(CStdType.String)) {
	code.plantInstruction(new MethodRefInstruction(opc_invokestatic,
						       "java/lang/String",
						       "valueOf",
						       "(" + right.getType().getSignature() + ")Ljava/lang/String;"));
      }

      code.plantInstruction(new MethodRefInstruction(opc_invokevirtual,
						     "java/lang/String",
						     "concat",
						     "(Ljava/lang/String;)Ljava/lang/String;"));

      ((CInitializable)left).genEndStoreCode(code, discardValue);
      return;
    }

    left.genCode(code);
    right.genCode(code);

    int		opcode = -1;

    switch (oper) {
    case OPE_STAR:
      opcode = JMultExpression.getOpcode(getType());
      break;
    case OPE_SLASH:
      opcode = JDivideExpression.getOpcode(getType());
      break;
    case OPE_PERCENT:
      opcode = JModuloExpression.getOpcode(getType());
      break;
    case OPE_PLUS:
      opcode = JAddExpression.getOpcode(getType());
      break;
    case OPE_MINUS:
      opcode = JMinusExpression.getOpcode(getType());
      break;
    case OPE_SL:
    case OPE_SR:
    case OPE_BSR:
      opcode = JShiftExpression.getOpcode(oper, getType());
      break;
    case OPE_BAND:
    case OPE_BXOR:
    case OPE_BOR:
      opcode = JBitwiseExpression.getOpcode(oper, getType());
      break;
    }

    code.plantInstruction(new NoArgInstruction(opcode));

    ((CInitializable)left).genEndStoreCode(code, discardValue);
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  protected  int		oper;
}
