/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JExpression.java,v 1.21 1999/10/08 11:19:00 vgp Exp $
 */

package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * Root class for all expressions
 */
public abstract class JExpression extends JPhylum {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

 /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   */
  public JExpression(TokenReference where) {
    super(where);
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * Compute the type of this expression (called after parsing)
   * @return the type of this expression
   */
  public abstract CType getType();

  /**
   * @return true if this expression is constant
   */
  public boolean isConstant() {
    return false;
  }

  // ----------------------------------------------------------------------
  // LITERAL HANDLING
  // ----------------------------------------------------------------------

  /**
   * Returns the literal value of this expression
   */
  public JLiteral getLiteral() {
    throw new RuntimeException("THIS IS NOT A LITERAL");
  }

  /**
   * Returns a typed literal for this expression
   *
   * The expression must be a constant
   */
  public final JOrdinalLiteral getOrdinalLiteral() {
    return (JOrdinalLiteral)getLiteral();
  }

  /**
   * Returns a typed literal for this expression
   *
   * The expression must be a constant
   */
  public final JRealLiteral getRealLiteral() {
    return (JRealLiteral)getLiteral();
  }

  /**
   * Returns a typed literal for this expression
   *
   * The expression must be a constant
   */
  public final JBooleanLiteral getBooleanLiteral() {
    return (JBooleanLiteral)getLiteral();
  }

  /**
   * Returns a typed literal for this expression
   *
   * The expression must be a constant
   */
  public final JNumberLiteral getNumberLiteral() {
    return (JNumberLiteral)getLiteral();
  }

  /**
   * Returns a typed literal for this expression
   *
   * The expression must be a constant
   */
  public final JStringLiteral getStringLiteral() {
    return (JStringLiteral)getLiteral();
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public abstract JExpression checkExpression(CExpressionContext context) throws PositionedError;

  /**
   * Adds a compiler error.
   * @param	context		the context in which the error occurred
   * @param	key		the message ident to be displayed
   * @param	params		the array of parameters
   *
   */
  protected void fail(CContext context, String key, Object[] params)
    throws PositionedError
  {
    throw new CLineError(getTokenReference(), key, params);
  }

  /**
   * convertType
   * changes the type of this expression to an other
   * @param  type the destination type
   */
  public JExpression convertType(CType type, CExpressionContext context) throws PositionedError {
    if (getType().equals(type)) {
      return this;
    } else if (isConstant() && getType().isNumeric()) {
      // not null
      return JLiteral.createLiteral(type, (this.getLiteral()).getValue());
    } else {
      return new JUnaryPromote(this, type).checkExpression(context);
    }
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Generate the code in pure java form
   * It is usefull to debug and tune compilation process
   * @param p the printwriter into the code is generated
   */
  public abstract void genJavaCode(CJavaPrettyPrinter p);

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    fail();
  }

  /**
   * Generates a sequence of bytescodes to branch on a label
   * This method helps to handle heavy optimizables conditions
   * @param	code		the code list
   */
  public void genBranch(boolean cond, CodeSequence code, CodeLabel label) {
    genCode(code);
    code.plantLabelRef(cond ? opc_ifne : opc_ifeq, label);
  }

  /**
   * Generates a sequence of bytescodes to branch on a label
   * This method helps to handle heavy optimizables conditions
   * @param	code		the code list
   */
  public JExpression unParenthesize() {
    return this;
  }

  // ----------------------------------------------------------------------
  // PUBLIC CONSTANTS
  // ----------------------------------------------------------------------

  public final static JExpression[]		EMPTY = new JExpression[0];
}
