/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JForStatement.java,v 1.32 1999/10/21 12:23:09 vgp Exp $
 */

package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.JavaStyleComment;

/**
 * JLS 14.10: While Statement
 *
 * The while statement executes an expression and a statement repeatedly
 * until the value of the expression is false.
 */
public class JForStatement extends JLoopStatement implements CBreakable {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   * @param	init		the init part
   * @param	cond		the cond part
   * @param	incr		the increment part
   * @param	body		the loop body.
   */
  public JForStatement(TokenReference where,
		       JStatement init,
		       JExpression cond,
		       JStatement incr,
		       JStatement body,
		       JavaStyleComment[] comments)
  {
    super(where, comments);
    this.init = init;
    this.cond = cond;
    this.incr = incr;
    if (body instanceof JBlock) {
      this.body = new JCompoundStatement(where, ((JBlock)body).getBody());
    } else {
      this.body = body;
    }
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check statement.
   *
   * @param	context		the actual context of analyse
   * @exception	PositionedError		if the check fails
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public void checkStatement(CBodyContext context) throws PositionedError {
    try {
      if (init != null) {
	init.checkBody(context);
      }

      // ------------------------------------------------------------
      CBlockContext self = new CBlockContext(context,
					     init instanceof JVariableDeclarationStatement ? 
					     ((JVariableDeclarationStatement)init).getVars().length :
					     0);
      CBodyContext  inside = new CBodyContext(self);

      if (cond != null) {
	cond = cond.checkExpression(new CExpressionContext(inside));
	check(context, cond.getType() == CStdType.Boolean, "for-cond-notboolean");
      }

      CBodyContext	neverContext = null;
      if (cond != null && (!cond.isConstant() || !((JBooleanLiteral)cond.getLiteral()).booleanValue())) {
	neverContext = inside.cloneContext();
      }

      CLoopContext	selfLoop = new CLoopContext(inside, this);
      CBodyContext	inside2 = new CBodyContext(selfLoop);
      body.checkStatement(inside2);

      if (neverContext == null && hasBreak()) {
	inside2.adopt(getBreakContextSummary());
      }
      
      inside2.close(getTokenReference());      
      selfLoop.close(getTokenReference());      

      if (incr != null) {
	incr.checkStatement(inside);
      }

      inside.close(getTokenReference());
      self.close(getTokenReference());

      if (neverContext != null) {
	context.merge(neverContext);
      } else if (!hasBreak()) {
	context.setReachable(false);
      }

      // ------------------------------------------------------------

    } catch (CBlockError e) {
      context.reportTrouble(e);
    }

    if (init instanceof JVariableDeclarationStatement) {
      ((JVariableDeclarationStatement)init).dereferenceVariable(context);
    }
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Generate the code in pure java form
   * It is usefull to debug and tune compilation process
   * @param	p		the printwriter into the code is generated
   */
  public void genJavaCode(CJavaPrettyPrinter p) {
    p.printForStatement(init, cond, incr, body);
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    CodeLabel		startLabel = new CodeLabel();
    CodeLabel		condLabel = new CodeLabel();

    code.pushContext(this);
    if (init != null) {
      init.genCode(code);			//		INIT
    }
    code.plantLabelRef(opc_goto, condLabel);	//		GOTO cond
    code.plantLabel(startLabel);		//	start:
    if (body != null) {
      body.genCode(code);			//		BODY
    }
    code.plantLabel(getContinue());		//	incr:
    if (incr != null) {
      incr.genCode(code);			//		INCR
    }
    code.plantLabel(condLabel);			//	cond:
    if (cond != null) {
      cond.genBranch(true, code, startLabel);	//		COND IFNE start
    } else {
      code.plantLabelRef(opc_goto, startLabel);	//		GOTO start
    }
    code.plantLabel(getEnd());			//	end:

    code.popContext(this);
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private JStatement		init;
  private JExpression		cond;
  private JStatement		incr;
  private JStatement		body;
}

