/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JLabeledStatement.java,v 1.20 1999/10/21 12:23:10 vgp Exp $
 */

package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.compiler.JavaStyleComment;

/**
 * JLS 14.6: Labeled Statement
 *
 * Statements may have label prefixes.
 */
public class JLabeledStatement extends JStatement implements CBreakable {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * @param where the line of this node in the source code
   * @param	label		the label of the enclosing labeled statement
   */
  public JLabeledStatement(TokenReference where, String label, JStatement stmt, JavaStyleComment[] comments) {
    super(where, comments);

    this.label = label;
    this.stmt = stmt;
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check statement.
   *
   * @param	context		the actual context of analyse
   * @exception	PositionedError		if the check fails
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public void checkStatement(CBodyContext context) throws PositionedError {
    try {
      context.getMethodContext().addLabel(label, this);
    } catch (UnpositionedError e) {
      throw e.getError(getTokenReference());
    }
    stmt.checkBody(context);
    context.getMethodContext().removeLabel(label);

    if (mayBeBreaked) {
      context.setReachable(true);
    }
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Generate the code in pure java form
   * It is usefull to debug and tune compilation process
   * @param	p		the printwriter into the code is generated
   */
  public void genJavaCode(CJavaPrettyPrinter p) {
    super.genJavaCode(p);
    p.printLabeledStatement(label, stmt);
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    stmt.genCode(code);
    code.plantLabel(endLabel);

    endLabel = null;
  }

  /**
   * Prevent statement that there is at least one break
   */
  public void addBreak(CBodyContext context) {
    mayBeBreaked = true;
    if (stmt instanceof CBreakable) {
      ((CBreakable)stmt).addBreak(context);
    }
  }

  /**
   * Return the end of this block (for break statement)
   */
  public CodeLabel getEnd() {
    if (stmt instanceof CBreakable) {
      return ((CBreakable)stmt).getEnd();
    } else {
      return endLabel;
    }
  }

  /**
   * Return the beginning of this block (for continue statement)
   */
  public CodeLabel getContinue() {
    return ((CBreakable)stmt).getContinue();
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private CodeLabel		endLabel = new CodeLabel();
  private String		label;
  private JStatement		stmt;
  private boolean		mayBeBreaked;
}
