/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JMethodCallExpression.java,v 1.43 1999/10/08 11:19:00 vgp Exp $
 */

package at.dms.kjc;

import at.dms.classfile.MethodRefInstruction;
import at.dms.classfile.InvokeinterfaceInstruction;
import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * This class represents methods call
 * <pre> methodname(explist) </pre>
 */
public class JMethodCallExpression extends JExpression {

  /**
   * Construct a node in the parsing tree
   * This method is directly called by the parser
   * @param	where		the line of this node in the source code
   * @param	prefix		an expression that is a field of a class representing a method
   * @param	ident		the method ident
   * @param	args		the argument of the call
   */
  public JMethodCallExpression(TokenReference where,
			       JExpression prefix,
			       JExpression[] args)
  {
    super(where);

    if (prefix instanceof JNameExpression) {
      this.ident = ((JNameExpression)prefix).getName().intern();
      this.prefix = ((JNameExpression)prefix).getPrefix();
    } else {
      throw new RuntimeException();
    }
    this.args = args;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * @return the type of this expression
   */
  public CType getType() {
    return method.getReturnType();
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public JExpression checkExpression(CExpressionContext context) throws PositionedError {
    CType[] argsType = new CType[args.length];

    context.setDiscardValue(false);
    for (int i = 0; i < argsType.length; i++) {
      args[i] = args[i].checkExpression(context);
      argsType[i] = args[i].getType();
    }

    if (ident == null) {
      if (prefix instanceof JClassFieldExpression) {
	check(context, false, "method-not-found", ((JClassFieldExpression)prefix).getIdent());
      }
      if (prefix instanceof JTypeNameExpression) {
	check(context, false, "method-not-found", ((JTypeNameExpression)prefix).getType());
      }
      check(context, false, "bad-method-name");
    }

    if (prefix != null) {
      prefix = prefix.checkExpression(context);
      if (prefix instanceof JNameExpression) {
	check(context, false, "bad-method-name", ((JNameExpression)prefix).getName());
      }
      check(context, prefix.getType().isReference(),
	    "method-badprefix", ident, prefix.getType());
      try {
	method = prefix.getType().getCClass().lookupMethod(ident, argsType);
      } catch (UnpositionedError e) {
	throw e.getError(getTokenReference());
      }
    } else {
      try {
	method = context.lookupMethod(ident, argsType);
      } catch (UnpositionedError e) {
	throw e.getError(getTokenReference());
      }
    }

    if (method == null) {
      if (prefix != null) {
	if (prefix instanceof JNameExpression) {
	  throw new CMethodNotFoundError(getTokenReference(),
					 this,
					 ((JNameExpression)prefix).getQualifiedName() + "." + ident,
					 argsType);
	} else if (prefix instanceof JTypeNameExpression) {
	  throw new CMethodNotFoundError(getTokenReference(),
					 this,
					 ((JTypeNameExpression)prefix).getQualifiedName() + "." + ident,
					 argsType);
	} else {
	  throw new CMethodNotFoundError(getTokenReference(), this, ident, argsType);
	}
      } else {
	throw new CMethodNotFoundError(getTokenReference(), this, ident, argsType);
      }
    }

    // check access
    CClass		local = context.getClassContext().getCClass();
    CClass		access = method.getOwner();

    check(context, method.isAccessible(local), "method-noaccess", method.getIdent());
    check(context, access.isAccessible(local), "class-noaccess", access.getIdent());

    CClassType[]	exceptions = method.getThrowables();
    for (int i = 0; i < exceptions.length; i++) {
      if (exceptions[i].isCheckedException()) {
	if (prefix == null ||				// special case of clone
	    !prefix.getType().isArrayType() ||
	    ident != "clone" ||
	    !exceptions[i].getCClass().getQualifiedName().
	    equals("java/lang/CloneNotSupportedException")) {
	  context.getBodyContext().addThrowable(new CThrowableInfo(exceptions[i],
								   this,
								   context.getBodyContext().makeSnapshot()));
	}
      }
    }

    if ((prefix == null) && !method.isStatic()) {
      if (access == local) {
	prefix = new JThisExpression(getTokenReference()).checkExpression(context);
      } else {
	prefix = new JThisExpression(getTokenReference(), access).checkExpression(context);
      }
    }

    check(context,
	  method.isStatic() || !(prefix instanceof JTypeNameExpression),
	  "method-static-bad", method.getIdent());

    argsType = method.getParameters();
    for (int i = 0; i < argsType.length; i++) {
      if (args[i] instanceof JTypeNameExpression) {
	check(context, false,   "var-unknown", ((JTypeNameExpression)args[i]).getQualifiedName());
      }
      args[i] = args[i].convertType(argsType[i], context);
    }

    return this;
  }

  private int getArgSize() {
    int		size = 0;
    CType[]	argsType = method.getParameters();

    for (int i = 0; i < argsType.length; i++) {
      size += argsType[i].getSize();
    }
    return  size;
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Generate the code in pure java form
   * It is usefull to debug and tune compilation process
   * @param p the printwriter into the code is generated
   */
  public void genJavaCode(CJavaPrettyPrinter p) {
    if (ident == null) {
      this.ident = ((JNameExpression)prefix).getName();
      prefix = ((JNameExpression)prefix).getPrefix();
    }
    p.printMethodCallExpression(prefix, ident, args);
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    boolean		forceNonVirtual = false;
    boolean		discardValue = code.discardValue();

    code.setDiscardValue(false);

    if (prefix != null) {
      if (method.isStatic()) {
	code.setDiscardValue(true);
      }
      prefix.genCode(code);

      if (prefix instanceof JSuperExpression) {
	forceNonVirtual = true;
      }
    }

    code.setDiscardValue(false);

    for (int i = 0; i < args.length; i++) {
      args[i].genCode(code);
    }

    if (method.isStatic()) {
      code.plantInstruction(new MethodRefInstruction(opc_invokestatic,
						     method.getPrefixName(),
						     method.getIdent(),
						     method.getSignature()));
    } else {
      if (forceNonVirtual || method.isPrivate()) {
	// JDK1.2 don't like... || method.isFinal())
	code.plantInstruction(new MethodRefInstruction(opc_invokespecial,
						       method.getPrefixName(),
						       method.getIdent(),
						       method.getSignature()));

      } else {
	if (method.getOwner().isInterface()) {
	  code.plantInstruction(new InvokeinterfaceInstruction(method.getPrefixName(),
							       method.getIdent(),
							       method.getSignature(),
							       getArgSize() + 1)); // this
	} else {
	  code.plantInstruction(new MethodRefInstruction(opc_invokevirtual,
							 method.getPrefixName(),
							 method.getIdent(),
							 method.getSignature()));
	}
      }
    }

    if (discardValue) {
      if (getType().getSize() == 2) {
	code.plantInstruction(new NoArgInstruction(opc_pop2));
      } else if (getType().getSize() == 1) {
	code.plantInstruction(new NoArgInstruction(opc_pop));
      }
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  protected	CMethod		method;
  protected	JExpression	prefix;
  protected	String		ident;
  protected     JExpression[]	args;
}
