/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JMethodDeclaration.java,v 1.44 1999/10/21 12:23:10 vgp Exp $
 */

package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.compiler.JavaStyleComment;

/**
 * This class represents a java class in the syntax tree
 */
public class JMethodDeclaration extends JMemberDeclaration {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * This method is directly called by the parser
   * @param	where		the line of this node in the source code
   * @param	parent		parent in which this methodclass is built
   * @param	modifiers	list of modifiers
   * @param	returnType	the return type of this method
   * @param	ident		the name of this method
   * @param	parameters	the parameters of this method
   * @param	exceptions	the exceptions throw by this method
   * @param	body		the body of this method
   * @param	javadoc		is this method deprecated
   */
  public JMethodDeclaration(TokenReference where,
			    int modifiers,
			    CType returnType,
			    String ident,
			    JFormalParameter[] parameters,
			    CClassType[] exceptions,
			    JBlock body,
			    JJavadocComment javadoc,
			    JavaStyleComment[] comments) {
    super(where, javadoc, comments);

    this.modifiers = modifiers;
    this.returnType = returnType;
    this.ident = ident.intern();
    this.body = body;

    this.parameters = parameters;
    this.exceptions = exceptions;
    assert(parameters != null);
    assert(exceptions != null);
  }

  // ----------------------------------------------------------------------
  // INTERFACE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Second pass (quick), check interface looks good
   * Exceptions are not allowed here, this pass is just a tuning
   * pass in order to create informations about exported elements
   * such as Classes, Interfaces, Methods, Constructors and Fields
   * @return true iff sub tree is correct enough to check code
   * @exception	PositionedError	an error with reference to the source file
   */
  public CSourceMethod checkInterface(CClassContext context) throws PositionedError {
    boolean	inInterface = context.getCClass().isInterface();
    String	ident = (this instanceof JConstructorDeclaration) ? JAV_CONSTRUCTOR : this.ident;

    // Collect all parsed data
    if (inInterface) {
      modifiers |= ACC_PUBLIC | (this instanceof JInitializerDeclaration ? 0 : ACC_ABSTRACT);
    }

    // 8.4.3 Method Modifiers
    check(context,
	  !hasOtherFlags(modifiers,
			 ACC_PUBLIC + ACC_PROTECTED + ACC_PRIVATE +
			 ACC_ABSTRACT + ACC_FINAL + ACC_STATIC +
			 ACC_NATIVE + ACC_SYNCHRONIZED),
	  "method-flags");

    if (inInterface && !(this instanceof JInitializerDeclaration)) {
      check(context,
	    !hasOtherFlags(modifiers, ACC_PUBLIC + ACC_ABSTRACT),
	    "method-flags-in-interface", this.ident);
    }

    try {
      returnType.checkType(context);

      CType[]	parameterTypes = new CType[parameters.length];
      for (int i = 0; i < parameterTypes.length; i++) {
	parameterTypes[i] = parameters[i].checkInterface(context);
      }

      for (int i = 0; i < this.exceptions.length; i++) {
	this.exceptions[i].checkType(context);
      }

      setInterface(new CSourceMethod(context.getCClass(), modifiers, ident, returnType,
				     parameterTypes, exceptions, isDeprecated(), body));
      
      return (CSourceMethod)getMethod();
    } catch (UnpositionedError cue) {
      throw cue.getError(getTokenReference());
    }
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception PositionedError Error catched as soon as possible
   */
  public void checkStatement(CClassContext context) throws PositionedError {
    try {
      check(context, context.getCClass().isAbstract() || !getMethod().isAbstract(),
	    "method-abstract-classnot", ident);

      CMethod sub = context.getCClass().lookupSuperMethod(ident, getMethod().getParameters());

      if (sub != null) {
	// check redefinition
	boolean	exactSignature = true;
	CType[]	params1 = sub.getParameters();
	CType[] params2 = ((CSourceMethod)getMethod()).getParameters();
	for (int i = 0; i < params1.length; i++) {
	  if (!params1[i].equals(params2[i])) {
	    exactSignature = false;
	    break;
	  }
	}
	if (exactSignature) {
	  check(context, returnType.equals(sub.getReturnType()),
		"method-return-different", ident);

	  if (!(this instanceof JInitializerDeclaration)) {
	    check(context,
		  !(getMethod().isPrivate() && !sub.isPrivate()),
		  "method-access-different", ident, null);
	    check(context,
		  !(getMethod().isProtected() && !(sub.isProtected() || sub.isPrivate() || (!sub.isPublic()))),
		  "method-access-different", ident, null);
	    check(context,
		  !(!getMethod().isPublic() && sub.isPublic()),
		  "method-access-different", ident, null);
	  }

	  if (!sub.isPrivate()) {
	    CClassType[]	exc = sub.getThrowables();

	  loop:
	    for (int i = 0; i < exceptions.length; i++) {
	      for (int j = 0; j < exc.length; j++) {
	        if (exceptions[i].isAssignableTo(exc[j])) {
		  continue loop;
		}
	      }
  	      check(context, false, "method-throws-different", ident, exceptions[i]);
            }
 	  }
        }
      }

      if (body != null) {
	check(context,
	      !context.getCClass().isInterface() || (this instanceof JInitializerDeclaration),
	      "method-body-in-interface", ident);

	if (!(this instanceof JInitializerDeclaration)) {
	  check(context,
		!getMethod().isNative() && !getMethod().isAbstract(),
		"method-body-native-abstract", ident);
	}

	CMethodContext	self = new CMethodContext(context, getMethod());
	CBlockContext	block = new CBlockContext(self, parameters.length);
	CBodyContext	inside = new CBodyContext(block);

	if (!getMethod().isStatic()) {
	  // add this local var
	  block.addThisVariable();
	}

	for (int i = 0; i < parameters.length; i++) {
	  parameters[i].checkBody(inside);
	}

	body.checkStatement(inside); // USE A COMPOUND STATEMENT !!!

	inside.close(getTokenReference());
	block.close(getTokenReference());
	self.close(getTokenReference());

	if (block.isReachable() && getMethod().getReturnType() != CStdType.Void) {
	  context.reportTrouble(new CLineError(getTokenReference(), "method-need-return", getMethod().getIdent()));
	}
      } else {
	check(context, getMethod().isAbstract() || getMethod().isNative() || context.getClassContext().getCClass().isInterface(),
	      "method-nobody-noabstract", ident);
      }
    } catch (UnpositionedError ce) {
      context.reportTrouble(ce.getError(getTokenReference()));
    }
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Generate the code in pure java form
   * It is usefull to debug and tune compilation process
   * @param	p		the printwriter into the code is generated
   */
  public void genJavaCode(CJavaPrettyPrinter p) {
    super.genJavaCode(p);
    p.printMethodDeclaration(modifiers,
			     returnType,
			     ident,
			     parameters,
			     exceptions,
			     body);
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    fail(); // nothing to do here
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  // $$$ MOVE TO BE PRIVATE
  protected int				modifiers;
  protected CType			returnType;
  protected String			ident;
  protected JFormalParameter[]		parameters;
  protected CClassType[]		exceptions;
  protected JBlock			body;
}
