/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JMinusExpression.java,v 1.15 1999/10/08 11:19:00 vgp Exp $
 */

package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * This class implements '-' specific operations
 * Plus operand may be String, numbers
 */
public class JMinusExpression extends JBinaryArithmeticExpression {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * This method is directly called by the parser
   * @param	where		the line of this node in the source code
   * @param	left		the left operand
   * @param	right		the right operand
   */
  public JMinusExpression(TokenReference where,
			  JExpression left,
			  JExpression right) {
    super(where, left, right);
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param	context		the actual context of analyse
   * @return	a pure java expression including promote node
   * @exception	UnpositionedError Error catched as soon as possible
   */
  public JExpression checkBinaryBody(CExpressionContext context) throws PositionedError {
    try {
      type = computeType(left.getType(), right.getType());
    } catch (UnpositionedError e) {
      throw e.getError(getTokenReference());
    }
    
    if (left.isConstant() && right.isConstant()) {
      return constantFolding(left.getNumberLiteral(), right.getNumberLiteral());
    } else {
      return this;
    }
  }

  /**
   * compute the type of this expression according to operands
   * @param	leftType		the type of left operand
   * @param	rightType		the type of right operand
   * @return	the type computed for this binary operation
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public static CType computeType(CType	leftType, CType rightType) throws UnpositionedError {
    if (leftType.isNumeric() && rightType.isNumeric()) {
      return CNumericType.binaryPromote(leftType, rightType);
    }
    throw new UnpositionedError("minus-badtype", leftType, rightType);
  }

  /**
   * @param	left		the left value
   * @param	right		the right value
   * @return	a literal resulting of an operation over two literals
   */
  public JExpression constantFolding(JNumberLiteral left, JNumberLiteral right) {
    if (type == CStdType.Double) {
      return new JRealLiteral(getTokenReference(),
			      (new Double(left.numberValue().doubleValue() -
					  (right.getNumberLiteral()).numberValue().doubleValue())),
			      CStdType.Double);
    } else if (type == CStdType.Float) {
      return new JRealLiteral(getTokenReference(),
			      (new Float(left.numberValue().floatValue() -
					 (right.getNumberLiteral()).numberValue().floatValue())),
			      CStdType.Float);
    } else if (type == CStdType.Long) {
      return new JOrdinalLiteral(getTokenReference(),
				 (new Long(left.numberValue().longValue() -
					   (right.getNumberLiteral()).numberValue().longValue())));
    } else if (type.implicitlyCastTo(CStdType.Integer)) {
      return new JOrdinalLiteral(getTokenReference(),
				 (new Integer(left.numberValue().intValue() -
					      (right.getNumberLiteral()).numberValue().intValue())));
    } else {
      fail();
    }

    return left;
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Generate the code in pure java form
   * It is usefull to debug and tune compilation process
   * @param p the printwriter into the code is generated
   */
  public void genJavaCode(CJavaPrettyPrinter p) {
    p.printBinaryExpression("-", left, right);
  }

  /**
   * @param	type		the type of result
   * @return	the type of opcode for this operation
   */
  public static int getOpcode(CType type) {
    switch (type.getTypeID()) {
    case TID_FLOAT:
      return opc_fsub;
    case TID_LONG:
      return opc_lsub;
    case TID_DOUBLE:
      return opc_dsub;
    default:
      return opc_isub;
    }
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    left.genCode(code);
    right.genCode(code);

    code.plantInstruction(new NoArgInstruction(getOpcode(getType())));
  }
}
