/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JOrdinalLiteral.java,v 1.21 1999/10/08 11:19:01 vgp Exp $
 */

package at.dms.kjc;

import at.dms.classfile.PushLiteralInstruction;

import at.dms.compiler.NumberParser;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * This class represents ordinal literals (byte, short, int, long)
 */
public class JOrdinalLiteral extends JNumberLiteral {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * @param where  the line of this node in the source code
   * @param image  the string representation of this literal
   */
  public JOrdinalLiteral(TokenReference where, String image) {
    super(where);
    this.image = image;
  }

  /**
   * Construct a node in the parsing tree
   * @param where  the line of this node in the source code
   * @param image  the string representation of this literal
   */
  public JOrdinalLiteral(TokenReference where, Number value) {
    super(where);
    setValue(value);
  }

  /**
   * Construct a node in the parsing tree
   * @param where  the line of this node in the source code
   * @param image  the string representation of this literal
   */
  public JOrdinalLiteral(TokenReference where, long value, CNumericType type) {
    super(where);
    setValue(new Long(value));
    this.type = type;
  }

  /**
   * Construct a node in the parsing tree
   * @param where  the line of this node in the source code
   * @param image  the string representation of this literal
   */
  public JOrdinalLiteral(TokenReference where, long value) {
    super(where);
    setValue(new Long(value));
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * This is needed to handle -2147483648 that is allowed agains 2147483648
   * that is not allowed, so we need to know if there is an unary minus before
   * us
   */
  public void setSignToMinus() {
    if (image != null) {
      image = "-" + image;
    }
    if (value != null) {
      value = new Long(-value.longValue());
    }
  }

  /**
   * Compute the type of this expression (called after parsing)
   * @return the type of this expression
   */
  public CType getType() {
    assert(type != null);
    return type;
  }

  /**
   * get a Number that represents the value of this literal
   * Do not call before checkBody
   */
  public Number numberValue() {
    assert(value != null);
    return value;
  }

  /**
   * get a Number that represents the value of this literal
   * Do not call before checkBody
   */
  public Object getValue() {
    if (getType() == CStdType.Long) {
      return value;
    } else {
      return new Integer(value.intValue());
    }
  }

  /**
   * @param	newValue		the new value
   */
  private void setValue(Number newValue) {
    long	val = newValue.longValue();

    if (val == newValue.byteValue()) {
      type = CStdType.Byte;
      value = new Byte(newValue.byteValue());
    } else if (val == newValue.shortValue()) {
      type = CStdType.Short;
      value = new Short(newValue.shortValue());
    } else if (val <= Character.MAX_VALUE && val >= Character.MIN_VALUE) {
      type = CStdType.Char;
      value = new Integer(newValue.intValue());
    } else if (val == newValue.intValue()) {
      type = CStdType.Integer;
      value = new Integer(newValue.intValue());
    } else {
      type = CStdType.Long;
      value = newValue;
    }
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   */
  public JExpression checkExpression(CExpressionContext context) throws PositionedError {
    Long	longObject = null;

    if (value != null) {
      return this; // already checked
    }

    try {
      if ((image.indexOf("L") > 0) || (image.indexOf("l") > 0)) {
	image = image.substring(0, image.length() - 1);
	type = CStdType.Long;
	value = new Long(NumberParser.decodeLong(image));
	image = null;

	return this;
      }
      longObject = new Long(NumberParser.decodeInt(image));
      image = null;
    } catch (NumberFormatException ex) {
      check(context, false, "liter-int-bad", image);
    }

    setValue(longObject);
    return this;
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Generate the code in pure java form
   * It is usefull to debug and tune compilation process
   * @param p the printwriter into the code is generated
   */
  public void genJavaCode(CJavaPrettyPrinter p) {
    if (value != null) {
      p.printNumberLiteral(value);
    } else {
      p.print(image);
    }
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    if (type == CStdType.Long) {
      code.plantInstruction(new PushLiteralInstruction(numberValue().longValue()));
    } else {
      code.plantInstruction(new PushLiteralInstruction(numberValue().intValue()));
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  protected String		image;
  private Number		value;
  private CType			type;
}
