/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JPostfixExpression.java,v 1.24 1999/10/08 11:19:01 vgp Exp $
 */

package at.dms.kjc;

import at.dms.classfile.IincInstruction;
import at.dms.classfile.NoArgInstruction;
import at.dms.classfile.PushLiteralInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;

/**
 * This class represents unary expressions.
 */
public class JPostfixExpression extends JExpression {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   * @param	oper		the operator
   * @param	expr		the operand
   */
  public JPostfixExpression(TokenReference where, int oper, JExpression expr) {
    super(where);
    this.oper = oper;
    this.expr = expr;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * Compute the type of this expression (called after parsing)
   * @return the type of this expression
   */
  public CType getType() {
    return type;
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * 15.13.2 Postfix Increment Operator ++
   * 15.13.3 Postfix Decrement Operator --
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public JExpression checkExpression(CExpressionContext context) throws PositionedError {
    boolean	leftSide = context.isLeftSide();

    context.setLeftSide(true);
    expr = expr.checkExpression(context);
    context.setLeftSide(leftSide);
    check(context, expr.getType().isNumeric(), "postfix-badtype", expr.getType());
    check(context,
	  (expr instanceof CInitializable) && ((CInitializable)expr).isLValue(context),
	  "postfix-notlvalue");
    check(context, ((CInitializable)expr).isInitialized(context), "postfix-notinitialized");

    type = expr.getType();

    return this;
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Generate the code in pure java form
   * It is usefull to debug and tune compilation process
   * @param p the printwriter into the code is generated
   */
  public void genJavaCode(CJavaPrettyPrinter p) {
    p.printPostfixExpression(oper, expr);
  }


  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());
    boolean		discardValue = code.discardValue();
    int			val = oper == OPE_POSTINC ? 1 : -1;

    if ((expr.getType() == CStdType.Integer) &&
	(expr instanceof JLocalVariableExpression)) {
      expr.genCode(code);
      code.plantInstruction(new IincInstruction(((JLocalVariableExpression)expr).getPosition(),
						val));
    } else {
      if (discardValue) {
	code.setDiscardValue(false);
	((CInitializable)expr).genStartStoreCode(code);
	expr.genCode(code);
      }  else {
	((CInitializable)expr).genStartStoreCode(code);
	expr.genCode(code);
	if (getType().getSize() == 2) {
	  if ((expr instanceof JLocalVariableExpression) ||
	      ((expr instanceof JClassFieldExpression) &&
	       hasFlag(((JClassFieldExpression)expr).getField().getModifiers(), ACC_STATIC))) {
	    code.plantInstruction(new NoArgInstruction(opc_dup2));
	  } else if (expr instanceof JArrayAccessExpression) {
	    code.plantInstruction(new NoArgInstruction(opc_dup2_x2));
	  } else {
	    code.plantInstruction(new NoArgInstruction(opc_dup2_x1));
	  }
	} else if (getType().getSize() == 1) {
	  if ((expr instanceof JLocalVariableExpression) ||
	      ((expr instanceof JClassFieldExpression) &&
	       hasFlag(((JClassFieldExpression)expr).getField().getModifiers(), ACC_STATIC))) {
	    code.plantInstruction(new NoArgInstruction(opc_dup));
	  } else if (expr instanceof JArrayAccessExpression) {
	    code.plantInstruction(new NoArgInstruction(opc_dup_x2));
	  } else {
	    code.plantInstruction(new NoArgInstruction(opc_dup_x1));
	  }
	}
      }

      switch (expr.getType().getTypeID()) {
      case TID_FLOAT:
	code.plantInstruction(new PushLiteralInstruction((float)val));
	code.plantInstruction(new NoArgInstruction(opc_fadd));
	break;
      case TID_LONG:
	code.plantInstruction(new PushLiteralInstruction((long)val));
	code.plantInstruction(new NoArgInstruction(opc_ladd));
	break;
      case TID_DOUBLE:
	code.plantInstruction(new PushLiteralInstruction((double)val));
	code.plantInstruction(new NoArgInstruction(opc_dadd));
	break;

      case TID_BYTE:
      case TID_CHAR:
      case TID_SHORT:
      case TID_INT:
	code.plantInstruction(new PushLiteralInstruction(val));
	code.plantInstruction(new NoArgInstruction(opc_iadd));
	switch (expr.getType().getTypeID()) {
	case TID_BYTE:
	  code.plantInstruction(new NoArgInstruction(opc_i2b));
	  break;
	case TID_CHAR:
	  code.plantInstruction(new NoArgInstruction(opc_i2c));
	  break;
	case TID_SHORT:
	  code.plantInstruction(new NoArgInstruction(opc_i2s));
	}
      }

      ((CInitializable)expr).genEndStoreCode(code, true);
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  protected int			oper;
  protected JExpression		expr;
  protected CType		type;
}

