/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JReturnStatement.java,v 1.26 1999/10/21 12:23:10 vgp Exp $
 */

package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.JavaStyleComment;

/**
 * JLS 14.15: Return Statement
 *
 * A return statement returns control to the invoker of a method or constructor.
 */
public class JReturnStatement extends JStatement {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

 /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   * @param	expr		the expression to return.
   */
  public JReturnStatement(TokenReference where, JExpression expr, JavaStyleComment[] comments) {
    super(where, comments);
    this.expr = expr;
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * Returns the type of this return statement
   */
  public CType getType() {
    return expr != null ? expr.getType() : CStdType.Void;
  }


  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check statement.
   *
   * @param	context		the actual context of analyse
   * @exception	PositionedError		if the check fails
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public void checkStatement(CBodyContext context) throws PositionedError {
    if (expr != null) {
      CExpressionContext	expressionContext = context.createExpressionContext();
      expr = expr.checkExpression(expressionContext);
      check(context, expr.getType().implicitlyCastTo(context.getMethodContext().getCMethod().getReturnType()) &&
	    expr.getType() != CStdType.Void,
	    "return-badtype", expr.getType(), context.getMethodContext().getCMethod().getReturnType());
      expr = expr.convertType(context.getMethodContext().getCMethod().getReturnType(), expressionContext);
    } else {
      check(context, context.getMethodContext().getCMethod().getReturnType() == CStdType.Void,
	    "return-empty-nonvoid");
    }

    context.addReturn();
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Generate the code in pure java form
   * It is usefull to debug and tune compilation process
   * @param	p		the printwriter into the code is generated
   */
  public void genJavaCode(CJavaPrettyPrinter p) {
    super.genJavaCode(p);
    p.printReturnStatement(expr);
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    if (expr != null) {
      expr.genCode(code);

      code.plantReturn(this);

      switch (expr.getType().getTypeID()) {
      case TID_CLASS:
      case TID_ARRAY:
	code.plantInstruction(new NoArgInstruction(opc_areturn));
	break;

      case TID_FLOAT:
	code.plantInstruction(new NoArgInstruction(opc_freturn));
	break;

      case TID_DOUBLE:
	code.plantInstruction(new NoArgInstruction(opc_dreturn));
	break;

      case TID_LONG:
	code.plantInstruction(new NoArgInstruction(opc_lreturn));
	break;

      default:
	code.plantInstruction(new NoArgInstruction(opc_ireturn));
      }
    } else {
      code.plantReturn(this);
      code.plantInstruction(new NoArgInstruction(opc_return));
    }
  }

  /**
   * Load the value from a local var (after finally)
   */
  public void load(CodeSequence code, JLocalVariable var) {
    switch (expr.getType().getTypeID()) {
    case TID_CLASS:
    case TID_ARRAY:
      code.plantLocalVar(opc_aload, var);
      break;

    case TID_FLOAT:
      code.plantLocalVar(opc_fload, var);
      break;

    case TID_DOUBLE:
      code.plantLocalVar(opc_dload, var);
      break;

    case TID_LONG:
      code.plantLocalVar(opc_lload, var);
      break;

    default:
      code.plantLocalVar(opc_iload, var);
    }
  }

  /**
   * Load the value from a local var (after finally)
   */
  public void store(CodeSequence code, JLocalVariable var) {
    switch (expr.getType().getTypeID()) {
    case TID_CLASS:
    case TID_ARRAY:
      code.plantLocalVar(opc_astore, var);
      break;

    case TID_FLOAT:
      code.plantLocalVar(opc_fstore, var);
      break;

    case TID_DOUBLE:
      code.plantLocalVar(opc_dstore, var);
      break;

    case TID_LONG:
      code.plantLocalVar(opc_lstore, var);
      break;

    default:
      code.plantLocalVar(opc_istore, var);
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  protected JExpression		expr;
}
