/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JShiftExpression.java,v 1.16 1999/10/08 11:19:01 vgp Exp $
 */

package at.dms.kjc;

import at.dms.classfile.NoArgInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * This class implements '+ - * /' specific operations
 * Plus operand may be String, numbers
 */
public class JShiftExpression extends JBinaryArithmeticExpression {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * This method is directly called by the parser
   * @param	where		the line of this node in the source code
   * @param	oper		the operator
   * @param	left		the left operand
   * @param	right		the right operand
   */
  public JShiftExpression(TokenReference where,
			  int oper,
			  JExpression left,
			  JExpression right) {
    super(where, left, right);
    this.oper = oper;
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public JExpression checkExpression(CExpressionContext context) throws PositionedError {
    left = left.checkExpression(context);
    right = right.checkExpression(context);

    JExpression expr = checkBinaryBody(context);
    if (expr == this) {
      left = left.convertType(type, context);
      right = right.convertType(CStdType.Integer, context);
    }
    return expr;
  }

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	UnpositionedError Error catched as soon as possible
   */
  public JExpression checkBinaryBody(CExpressionContext context) throws PositionedError {
    try {
      type = computeType(left.getType(), right.getType());
    } catch (UnpositionedError e) {
      throw e.getError(getTokenReference());
    }

    if (left.isConstant() && right.isConstant()) {
      return constantFolding(left.getOrdinalLiteral(), right.getOrdinalLiteral());
    } else {
      return this;
    }
  }

  /**
   * compute the type of this expression according to operands
   * @param	leftType		the type of left operand
   * @param	rightType		the type of right operand
   * @return	the type computed for this binary operation
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public static CType computeType(CType	leftType, CType rightType) throws UnpositionedError {
    //if (rightType == CStdType.Long)
    // check the spec  throw new UnpositionedError("shift-badtype", leftType, rightType);

    if (leftType.isOrdinal() && rightType.isOrdinal()) {
      return leftType = CNumericType.unaryPromote(leftType);
    }

    throw new UnpositionedError("shift-badtype", leftType, rightType);
  }

  /**
   * @param	left		the left literal
   * @param	right		the right literal
   * @return	a literal resulting of an operation over two literals
   */
  public JExpression constantFolding(JOrdinalLiteral left, JOrdinalLiteral right) {
    if (type == CStdType.Long) {
      switch (oper) {
      case OPE_SL:
	left = new JOrdinalLiteral(getTokenReference(),
				    (new Long(left.numberValue().longValue() <<
					      right.numberValue().longValue())));
	break;
      case OPE_SR:
	left = new JOrdinalLiteral(getTokenReference(),
				   (new Long(left.numberValue().longValue() >>
					     right.numberValue().longValue())));
	break;
      case OPE_BSR:
	left = new JOrdinalLiteral(getTokenReference(),
				   (new Long(left.numberValue().longValue() >>>
					     right.numberValue().longValue())));
	break;
      }
    } else if (type.implicitlyCastTo(CStdType.Integer)) {
      switch (oper) {
      case OPE_SL:
	left = new JOrdinalLiteral(getTokenReference(),
				   (new Integer(left.numberValue().intValue() <<
						right.numberValue().intValue())));
	break;
      case OPE_SR:
	left = new JOrdinalLiteral(getTokenReference(),
				   (new Integer(left.numberValue().intValue() >>
						right.numberValue().intValue())));
	break;
      case OPE_BSR:
	left = new JOrdinalLiteral(getTokenReference(),
				   (new Integer(left.numberValue().intValue() >>>
						right.numberValue().intValue())));
	break;
      }
    } else {
      fail();
    }

    return left;
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Generate the code in pure java form
   * It is usefull to debug and tune compilation process
   * @param p the printwriter into the code is generated
   */
  public void genJavaCode(CJavaPrettyPrinter p) {
    p.printShiftExpression(oper, left, right);
  }


  /**
   * @param	type		the type of result
   * @return	the type of opcode for this operation
   */
  public static int getOpcode(int oper, CType type) {
    if (type == CStdType.Long) {
      switch (oper) {
      case OPE_SL:
	return opc_lshl;
      case OPE_SR:
	return opc_lshr;
      case OPE_BSR:
	return opc_lushr;
      }
    } else {
      switch (oper) {
      case OPE_SL:
	return opc_ishl;
      case OPE_SR:
	return opc_ishr;
      case OPE_BSR:
	return opc_iushr;
      }
    }
    return -1;
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    left.genCode(code);
    right.genCode(code);

    code.plantInstruction(new NoArgInstruction(getOpcode(oper, getType())));
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  protected	int			oper;
}
