/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JSwitchStatement.java,v 1.42 1999/10/21 12:23:10 vgp Exp $
 */

package at.dms.kjc;

import java.util.Vector;

import at.dms.classfile.SwitchInstruction;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.JavaStyleComment;
import at.dms.util.VectorCache;

/**
 * JLS 14.9: Switch Statement
 */
public class JSwitchStatement extends JStatement implements CBreakable {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   * @param	expr		the expr part
   * @param	body		the different part of body.
   */
  public JSwitchStatement(TokenReference where,
			  JExpression expr,
			  JSwitchGroup[] body,
			  JavaStyleComment[] comments) {
    super(where, comments);

    this.expr = expr;
    this.body = body;
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check statement.
   *
   * @param	context		the actual context of analyse
   * @exception	PositionedError		if the check fails
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public void checkStatement(CBodyContext context) throws PositionedError {
    expr = expr.checkExpression(context.createExpressionContext());
    check(context, expr.getType().implicitlyCastTo(CStdType.Integer), "switch-badexpr");

    CSwitchBodyContext	self = new CSwitchBodyContext(this, context, expr.getType());
    CBodyContext	inside = new CBodyContext(self);
    CBodyContext	bodyContext = inside.cloneContext();
    CBodyContext[]	bodys = new CBodyContext[body.length];

    for (int i = 0; i < body.length; i++) {
      bodys[i] = body[i].checkBody(self, i == 0 || bodys[i - 1].isReachable() ? 
				   bodyContext : 
				   inside.cloneContext());
    }

    boolean	isFirst = self.defaultExists();
    for (int i = 0; i < bodys.length; i++) {
      if (isFirst && (bodys[i].isReachable() || bodys[i].isBreaked())) {
	inside.adopt(bodys[i]);
	isFirst = false;
      } else if (bodys[i].isBreaked() || 
		 i == bodys.length - 1 && bodys[i].isReachable()) {
	inside.merge(bodys[i]);
      }
    }
    inside.close(getTokenReference());
    self.setReachable(bodys[bodys.length - 1].isReachable() ||
		      hasBreak ||
		      !self.defaultExists());
    self.close(getTokenReference());
  }

  /**
   * Prevent statement that there is at least one break
   */
  public void addBreak(CBodyContext context) {
    hasBreak = true;
  }

  /**
   * Return the end of this block (for break statement)
   */
  public CodeLabel getEnd() {
    return endLabel;
  }

  /**
   * Return the beginning of this block (for continue statement)
   */
  public CodeLabel getContinue() {
    fail();
    return null;
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Generate the code in pure java form
   * It is usefull to debug and tune compilation process
   * @param	p		the printwriter into the code is generated
   */
  public void genJavaCode(CJavaPrettyPrinter p) {
    super.genJavaCode(p);
    p.printSwitchStatement(expr, body);
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    CodeLabel		defaultLabel = new CodeLabel();

    Vector		matches = VectorCache.request(VectorCache.BIG);
    Vector		targets = VectorCache.request(VectorCache.BIG);

    for (int i = 0; i < body.length; i++) {
      body[i].collectLabels(defaultLabel, matches, targets);
    }

    expr.genCode(code);

    code.pushContext(this);
    code.plantInstruction(new SwitchInstruction(defaultLabel, matches, targets));
    for (int i = 0; i < body.length; i++) {
      body[i].genCode(code);
    }

    if (!defaultLabel.hasAddress()) {
      code.plantLabel(defaultLabel);
    }

    code.plantLabel(endLabel);
    code.popContext(this);

    VectorCache.release(matches);
    VectorCache.release(targets);

    endLabel = null;
  }


  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private JExpression		expr;
  private JSwitchGroup[]	body;
  private boolean		hasBreak;
  private CodeLabel		endLabel = new CodeLabel();
}
