/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JTypeDeclaration.java,v 1.63 1999/10/21 12:23:10 vgp Exp $
 */

package at.dms.kjc;

import java.util.Hashtable;
import at.dms.compiler.CWarning;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.compiler.JavaStyleComment;

/**
 * This class represents a java class or interface in the syntax tree
 */
public abstract class JTypeDeclaration extends JMemberDeclaration {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Constructs a type declaration in the parsing tree.
   * @param	where		the line of this node in the source code
   * @param	parent		the parent in which this class is defined
   * @param	modifiers	the list of modifiers of this class
   * @param	ident		the short name of this class
   * @param	interfaces	the names of this types's interfaces
   * @param	javadoc	is this type declaration deprecated
   */
  public JTypeDeclaration(TokenReference where,
			  int modifiers,
			  String ident,
			  CClassType[] interfaces,
			  JFieldDeclaration[] fields,
			  JMethodDeclaration[] methods,
			  JTypeDeclaration[] inners,
			  JPhylum[] initializers,
			  JJavadocComment javadoc,
			  JavaStyleComment[] comments) {
    super(where, javadoc, comments);

    this.modifiers = modifiers;
    this.ident = ident.intern();
    this.interfaces = interfaces;
    this.fields = fields;
    this.methods = methods;
    this.inners = inners;
    this.body = initializers;
  }

  public void generateInterface(CClass owner, String prefix) {
    // We can now define an intermediate external representation of this class to use internally
    setInterface(sourceClass = new CSourceClass(owner, getTokenReference(), modifiers, ident, prefix + ident, isDeprecated()));

    CClassType[]	innerClasses = new CClassType[inners.length];
    for (int i = 0; i < inners.length; i++) {
      inners[i].generateInterface(sourceClass, sourceClass.getQualifiedName() + "$");
      //if (!inners[i].getCClass().isStatic() && !getCClass().isInterface() && !inners[i].getCClass().isInterface()) {
      //inners[i].addOuterThis();
      //}
      innerClasses[i] = inners[i].getCClass().getType();
    }

    sourceClass.setInnerClasses(innerClasses); // prevent interface
    uniqueSourceClass = CTopLevel.addSourceClass(sourceClass);
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * @return	true if this class is at top level
   */
  public boolean isAtTopLevel() {
    return getCClass().getOwner() == null;
  }

  public JFieldDeclaration[] getFields() {
    return fields;
  }

  public JConstructorDeclaration getDefaultConstructor() {
    return defaultConstructor;
  }

  public void setDefaultConstructor(JConstructorDeclaration defaultConstructor) {
    this.defaultConstructor = defaultConstructor;
  }

  public void setIdent(String ident) {
    this.ident = ident;
  }

  // ----------------------------------------------------------------------
  // INTERFACE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Second pass (quick), check interface looks good
   * @exception	PositionedError	an error with reference to the source file
   */
  public abstract void checkInterface(CContext context) throws PositionedError;

  /**
   * Second pass (quick), check interface looks good
   * @exception	PositionedError	an error with reference to the source file
   */
  protected void checkInterface(CContext context, CClassType superClass) throws PositionedError {
    CClassContext	self = new CClassContext(context, sourceClass, this);
    Hashtable		hashField;
    CMethod[]		methodList;
    Hashtable		hashMethod;

    if (!uniqueSourceClass) {
      context.reportTrouble(new CWarning(getTokenReference(), "top-rename", sourceClass.getQualifiedName()));
    }

    // 8.1.2 Class Modifiers
    if (isAtTopLevel()) {
      check(context,
	    !hasFlag(getCClass().getModifiers(), ACC_PROTECTED + ACC_PRIVATE), "class-flags");
    }

    check(context,
	  !hasFlag(getCClass().getModifiers(), ACC_SYNCHRONIZED + ACC_VOLATILE + ACC_TRANSIENT + ACC_NATIVE),
	  "class-flags");

    if (isAtTopLevel() && sourceClass.isPublic() && !getTokenReference().getName().startsWith(ident + ".")) {
      context.reportTrouble(new CWarning(getTokenReference(), "class-name-filename",
					 ident, getTokenReference().getName()));
    }

    // resolve interfaces
    try {
      for (int i = 0; i < interfaces.length; i++) {
	interfaces[i].checkType(self);
      }
    } catch (UnpositionedError cue) {
      throw cue.getError(getTokenReference());
    }

    // Add fields of this class
    int		generatedFields = getCClass().hasOuterThis() ? 1 : 0;
    hashField = new Hashtable(fields.length + generatedFields + 1, 1.5f);
    for (int i = fields.length - 1; i >= 0 ; i--) {
      CSourceField	field = fields[i].checkInterface(self);

      field.setPosition(i);
      check(context, hashField.put(field.getIdent(), field) == null,
	    "field-rename", field.getIdent());
    }
    if (generatedFields > 0) {
      CSourceField	field = outerThis.checkInterface(self);

      field.setPosition(hashField.size());
      check(context, hashField.put(JAV_OUTER_THIS, field) == null,
	    "field-rename", JAV_OUTER_THIS);
    }

    // Add methods of this class
    int		generatedMethods = 0;
    generatedMethods += defaultConstructor != null ? 1 : 0;
    generatedMethods += statInit != null && !statInit.isDummy() ? 1 : 0;
    generatedMethods += instanceInit != null && !instanceInit.isDummy() ? 1 : 0;

    hashMethod = new Hashtable(methods.length + generatedMethods + 1, 1.1f);
    methodList = new CMethod[methods.length + generatedMethods];

    int		count = 0;
    for (int i = 0; i < methods.length; i++) {
      CMethod			other, selfM;

      methodList[count++] = selfM = methods[i].checkInterface(self);

      if ((other = (CMethod)hashMethod.put(selfM.getIdent(), selfM)) != null) {
	check(context, !other.equals(selfM), "method-redefine", methods[i].getMethod().getIdent());
      }
    }
    if (defaultConstructor != null) {
      methodList[count] = defaultConstructor.checkInterface(self);
      hashMethod.put(methodList[count], methodList[count]);
      count++;
    }
    if (statInit != null) {
      if (!statInit.isDummy()) {
	methodList[count] = statInit.checkInterface(self);
	hashMethod.put(methodList[count], methodList[count]);
	count++;
      } else {
	statInit.checkInterface(self);
      }
    }
    if (instanceInit != null) {
      if (!instanceInit.isDummy()) {
	methodList[count] = instanceInit.checkInterface(self);
	hashMethod.put(methodList[count], methodList[count]);
	count++;
      } else {
	instanceInit.checkInterface(self);
      }
    }

    // Check inners
    for (int i = inners.length - 1; i >= 0 ; i--) {
      inners[i].checkInterface(self);
    }

    sourceClass.close(this.interfaces, superClass, hashField, methodList, hashMethod);
  }

  /**
   * Checks that interface are not duplicated
   *
   * @exception	PositionedError	Error catched as soon as possible
   */
  public void checkInitializers(CContext context) throws PositionedError {
    for (int i = 0; i < interfaces.length; i++) {
      for (int j = 0; j < i - 1; j++) {
	check(context, !interfaces[i].equals(interfaces[j]),
	      "interfaces-duplicate", interfaces[i]);
      }
    }

    sourceClass.setInitializerChecked();
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * checkBody
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception PositionedError Error catched as soon as possible
   */
  public void addOuterThis() {
    if (outerThis == null) {
      sourceClass.setHasOuterThis(true);
      outerThis = new JFieldDeclaration(getTokenReference(),
					new JVariableDefinition(getTokenReference(),
								ACC_PRIVATE,
								getOwner().getCClass().getType(),
								JAV_OUTER_THIS,
								null),
					null,
					null);
      ((CSourceClass)getCClass()).addField(new CSourceField(getCClass(),
							    ACC_PRIVATE,
							    JAV_OUTER_THIS,
							    getOwner().getCClass().getType(),
							    false));
    }
  }

  /**
   * createContext
   * @param	parent		the parent context or null
   * @return	returns a CClassContext that represents this context
   * @exception	UnpositionedError	Error catched as soon as possible
   */
  public abstract CClassContext createContext(CContext parent) throws UnpositionedError;

  /**
   * checkTypeBody
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception PositionedError Error catched as soon as possible (for subclasses)
   */
  public void checkTypeBody(CContext context) throws PositionedError {
    context.addSourceClass(sourceClass);
  }

  // ----------------------------------------------------------------------
  // PROTECTED UTILITIES
  // ----------------------------------------------------------------------

  public CClass getOwner() {
    return getCClass().getOwner();
  }


  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBER
  // ----------------------------------------------------------------------

  protected	int			modifiers;
  protected	String			ident;

  protected	JPhylum[]		body; // !!! explain
  protected	JFieldDeclaration[]	fields;
  protected	JMethodDeclaration[]	methods;
  protected	JTypeDeclaration[]	inners;
  protected	CClassType[]		interfaces;

  private	JFieldDeclaration	outerThis;
  private	JConstructorDeclaration defaultConstructor;
  protected	JInitializerDeclaration	statInit;
  protected	JInitializerDeclaration	instanceInit;

  // Definitive data
  protected	CSourceClass		sourceClass;
  private	boolean			uniqueSourceClass = true;
}
