/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JWhileStatement.java,v 1.32 1999/10/21 12:23:11 vgp Exp $
 */

package at.dms.kjc;

import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.JavaStyleComment;

/**
 * JLS 14.10: While Statement
 *
 * The while statement executes an expression and a statement repeatedly
 * until the value of the expression is false.
 */
public class JWhileStatement extends JLoopStatement implements CBreakable {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * @param	where		the line of this node in the source code
   * @param	cond		the expression to evaluate.
   * @param	body		the loop body.
   */
  public JWhileStatement(TokenReference where, JExpression cond, JStatement body, JavaStyleComment[] comments) {
    super(where, comments);
    this.cond = cond;
    this.body = body;
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check statement.
   *
   * @param	context		the actual context of analyse
   * @exception	PositionedError		if the check fails
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public void checkStatement(CBodyContext context) throws PositionedError {
    try {
      cond = cond.checkExpression(context.createExpressionContext());
      check(context, cond.getType() == CStdType.Boolean,
	    "while-cond-notboolean", cond.getType());

      if (!cond.isConstant()) {
	CBodyContext	neverContext = context.cloneContext();
	CLoopContext	self = new CLoopContext(context, this);
	CBodyContext	bodyContext = new CBodyContext(self);

	body.checkBody(bodyContext);
	bodyContext.close(getTokenReference());
	self.close(getTokenReference());

	context.merge(neverContext);
      } else {
	if (! (cond.getBooleanLiteral()).booleanValue()) {
	  // while (false)
	  throw new CLineError(body.getTokenReference(), "unreachable-statement");
	}
	CLoopContext	self = new CLoopContext(context, this);
	CBodyContext	bodyContext = new CBodyContext(self);

	body.checkBody(bodyContext);

	if (hasBreak()) {
	  bodyContext.adopt(getBreakContextSummary());
	} 

	bodyContext.close(getTokenReference());
	self.close(getTokenReference());
	context.setReachable(hasBreak());
      }
    } catch (CBlockError e) {
      context.reportTrouble(e);
    }
  }

  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Generate the code in pure java form
   * It is usefull to debug and tune compilation process
   * @param	p		the printwriter into the code is generated
   */
  public void genJavaCode(CJavaPrettyPrinter p) {
    super.genJavaCode(p);
    p.printWhileStatement(cond, body);
  }

  /**
   * Generates a sequence of bytescodes
   * @param	code		the code list
   */
  public void genCode(CodeSequence code) {
    code.setLineNumber(getTokenReference().getLine());

    code.pushContext(this);

    if (cond.isConstant() && (cond.getBooleanLiteral()).booleanValue()) { // while true
      code.plantLabel(getContinue());
      body.genCode(code);
      code.plantLabelRef(opc_goto, getContinue());
      code.plantLabel(getEnd());
    } else {
      CodeLabel		bodyLabel = new CodeLabel();
      code.plantLabelRef(opc_goto, getContinue());	//		GOTO cont
      code.plantLabel(bodyLabel);			//	cont:
      body.genCode(code);				//		BODY CODE
      code.plantLabel(getContinue());			//	cont:
      cond.genBranch(true, code, bodyLabel);      	//		IFNE body
      code.plantLabel(getEnd());			//	end:	...
    }

    code.popContext(this);
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private JExpression		cond;
  private JStatement		body;
}
