/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Main.java,v 1.104 1999/10/16 23:25:35 vgp Exp $
 */

package at.dms.kjc;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Vector;

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;

import at.dms.compiler.Compiler;
import at.dms.compiler.Console;
import at.dms.compiler.CWarning;
import at.dms.util.Message;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * This class implements the entry point of the Java compiler
 */
public class Main extends Compiler {

  /**
   * The constructor, you use main() to compile
   */
  public Main() {
    // each compile session is identified by this timestamp
    timestamp = System.currentTimeMillis();
  }

  // ----------------------------------------------------------------------
  // RUN FROM COMMAND LINE
  // ----------------------------------------------------------------------

  /*
   * Runs a compilation session
   *
   * @param	dir		the working directory
   * @param	err		the error output stream
   * @param	console		the console to show the progess of the compilation (!!! not used)
   * @param	args		the arguments to the compiler
   * @return	true iff the compilation succeeded
   */
  public boolean run(String dir, PrintWriter err, Console console, String[] args) {
    this.err = err;
    setWorkingDirectory(dir);
    this.console = console;

    return run(args);
  }

  /**
   * Runs a compilation session
   *
   * @param	args		the command line arguments
   */
  public boolean run(String[] args) {
    Message.addBundle("kjc");

    if (!parseArgs(args)) {
      return false;
    }

    initialize();

    if (infiles.size() == 0) {
      usage(Message.format("usage-nofile", null));
      return false;
    }

    if (verboseMode()) {
      if (console == null) {
	console = new Console(err, true);
      }
      console.compilationStarted(infiles.size());
    }

    try {
      infiles = verifyFiles(infiles);
    } catch (UnpositionedError e) {
      err.println(e.getMessage());
      return false;
    }

    destination = checkDestination(destination);

    CTopLevel.initSession(this);	//!!! add end session
    JCompilationUnit[]	tree = new JCompilationUnit[infiles.size()];

    if (nbProc > tree.length) {
      nbProc = tree.length;
    }

    if (optionMultiThread) {
      parseMultiProc(tree);
    } else {
      for (int count = 0; count < tree.length; count++) {
	tree[count] = parseFile((File)infiles.elementAt(count));
      }
    }

    infiles = null;

    if (errorFound) {
      return false;
    }

    if (!optionBeautify) {
      long	lastTime = System.currentTimeMillis();
      for (int count = 0; count < tree.length; count++) {
	checkInterface(tree[count]);
      }
      if (verboseMode()) {
	console.checkInterface(lastTime);
      }

      if (errorFound) {
	return false;
      }

      for (int count = 0; count < tree.length; count++) {
	checkInitializers(tree[count]);
      }

      if (errorFound) {
	return false;
      }

      if (optionMultiThread) {
	checkBodyMultiProc(tree);
      } else {
	for (int count = 0; count < tree.length; count++) {
	  checkBody(tree[count]);
	  if (!optionGenerateJava && !optionBeautify) {
	    tree[count] = null;
	  }
	}
      }

      if (errorFound) {
	return false;
      }
    }

    if (!optionNoWrite) {
      if (optionGenerateJava || optionBeautify) {
	if (optionMultiThread) {
	  genJavaCodeMultiProc(tree);
	} else {
	  for (int count = 0; count < tree.length; count++) {
	    tree[count].genJavaCode(destination);
	    tree[count] = null;
	  }
	}
      } else {
	File dir = destination == null ? null : new File(destination);
	genCode(dir);
      }
    }

    if (verboseMode()) {
      console.compilationEnded();
    }

    CodeSequence.endSession();

    return true;
  }

  /**
   * Generates the code from an array of compilation unit and
   * a destination
   *
   * @param	tree		an array of compilation unit
   * @param	destination	the directory where to write classfiles
   */
  public void genCode(File destination) {
    CSourceClass[]	classes = getClasses();
    this.classes.setSize(0);

    try {
      if (optionMultiThread) {
	genCodeMultiProc(classes, destination);
      } else {
	for (int count = 0; count < classes.length; count++) {
	  long last = System.currentTimeMillis();
	  classes[count].genCode(destination);
	  if (verboseMode()) {
	    console.generateClassFile(last, classes[count].getQualifiedName());
	  }
	  classes[count] = null;
	}
      }
    } catch (IOException ioe) {
      System.err.println(ioe.getMessage());
    }
  }

  /*
   * Parse command line arguments.
   */
  protected boolean parseArgs(String[] args) {
    Getopt	g = new Getopt("kjc", args, SHORTOPTS, LONGOPTS, true);

  _loop_:
    for (;;) {
      switch (g.getopt()) {
      case 'b':
	optionBeautify = true;
	break;

      case 'd':
	destination = g.getOptarg();
	break;

      case 'g':
	//optionDebug = true;
	break;

      case 'h':
	usage("");
	help();
	err.flush();
	System.exit(0);

      case 'j':
	optionGenerateJava = true;
	break;

      case 'm':
	optionMultiThread = true;
	break;

      case 'n':
	optionNoWrite = true;
	break;

      case 'O':
	try {
	  optionOptimize = g.getOptarg() != null ?
	    new Integer(g.getOptarg()).intValue() * 10 :
	    1;
	} catch (Exception e) {
	  err.println("malformed option: " + g.getOptarg());
	  err.flush();
	  System.exit(0);
	}
	break;

      case 'V':
	err.println("KOPI Java Compiler Version: " + VERSION);
	err.flush();
	System.exit(0);
	break;

      case 'v':
	optionVerbose = true;
	break;

      case 'w':
	optionWarning = true;
	break;

      case 500:
	classpath = g.getOptarg();
	break;

      case 501:
	optionDeprecation = true;
	break;

      case 503:
	optionWarning = false;
	break;

      case -1:
	break _loop_;

      default:
	return false;
      }
    }

    for (int i = g.getOptind(); i < args.length; i++) {
      infiles.addElement(args[i]);
    }

    return true;
  }

  /**
   * Initialize the compiler (read classpath, check classes.zip)
   */
  protected void initialize() {
    if (classpath == null) {
      classpath = System.getProperty("java.class.path");
      if (classpath == null) {
	classpath = ".";
      }
    }
    at.dms.classfile.ClassPath.init(classpath);
    CStdType.init(this);
  }

  /**
   * shows an usage message
   */
  protected void usage(String msg) {
    err.println("kjc: " + msg);
    err.println(Message.format("usage-line", null));
    err.println();
    err.println(Message.format("info1", null));
    err.println(Message.format("info2", null));
    err.println(Message.format("info3", null));
    err.println(Message.format("info4", null));
  }

  /**
   * shows a help message
   */
  protected void help() {
    err.println(Message.format("usage-verbose", null));
    err.println(Message.format("usage-version", null));
    err.println(Message.format("usage-java", null));
    err.println(Message.format("usage-beautify", null));
    err.println(Message.format("usage-classpath", null));
    err.println(Message.format("usage-dest", null));
    err.println(Message.format("usage-multi", null));
    err.println(Message.format("usage-nowrite", null));
    err.println(Message.format("usage-warning", null));
    err.println(Message.format("usage-strict", null));
    err.println(Message.format("usage-fast", null));
    err.println(Message.format("usage-deprecation", null));
    err.println(Message.format("usage-optimize1", null));
    err.println(Message.format("usage-optimize2", null));
    err.println();
    err.println(Message.format("info1", null));
    err.println(Message.format("info2", null));
    err.println(Message.format("info3", null));
    err.println(Message.format("info4", null));
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * @return the timestamp for this compile process
   * CClassType use an  internal cache for class that need to be flushed
   * after each compilation process
   */
  public static long getTimeStamp() {
    return instance == null ? 0 : instance.timestamp;
  }

  /**
   * @return true if compilation in verbose mode
   */
  public boolean verboseMode() {
    return optionVerbose;
  }

  /**
   * @return true if compilation in verbose mode
   */
  public static boolean deprecationMode() {
    return instance == null ? false : instance.optionDeprecation;
  }

  /**
   * Returns the console
   */
  public static Console getConsole() {
    if (instance.console == null) {
      instance.console = new Console(err, instance.verboseMode());
    }
    return instance.console;
  }

  /**
   * @return the optimization level
   */
  public static int optimizeCode() {
    return instance == null ? 10 : instance.optionOptimize;
  }

  /**
   * @return true if compiler should check for warnings
   */
  public static boolean wantsWarning() {
    return instance.optionWarning;
  }

  /**
   * Shows an error/warning and set error flag to true
   */
  public static CJavaPrettyPrinter getPrettyPrinter(String fileName) {
    return instance != null ?
      instance.getPrettyPrinterOverided(fileName) :
      new CJavaPrettyPrinter(fileName);
  }

  /**
   * Entry point
   *
   * @param	args		the command line arguments
   */
  public static void main(String[] args) {
    boolean	success;

    success = compile(args);

    System.exit(success ? 0 : 1);
  }

  /**
   * Second entry point
   */
  public static boolean compile(String[] args) {
    if (instance != null) {
      // only one instance of Main at the same time
      throw new RuntimeException("DO NOT CALL MAIN WHILE THERE IS A COMPILE PROCESS");
    }

    boolean	success;

    instance = new Main();
    success = instance.run(args);
    instance = null;
    err.flush();

    return success;
  }

  /**
   * Return the currently compiling compiler
   */
  public Main getCurrentInstance() {
    return instance;
  }

  // ----------------------------------------------------------------------
  // PROTECTED METHODS
  // ----------------------------------------------------------------------

  /**
   * parse the givven file and return a compilation unit
   * side effect: increment error number
   * @param	file		the name of the file (assert exists)
   * @return	the compilation unit defined by this file
   */
  protected JCompilationUnit parseFile(File file) {
    long		lastTime = System.currentTimeMillis();
    CHandLexer		lexer = null;
    ParseManager	parsers = null;
    try {
      parsers = ParseManager.getParseManager(this, file, file.getPath());
      // Create a parser that reads from the scanner
      lexer = parsers.getJavaLexer();
      JCompilationUnit unit = parsers.getJavaParser().jCompilationUnit();
      if (verboseMode()) {
	console.parseFile(lastTime, file.getPath());
      }
      parsers.close();
      return unit;
    } catch (antlr.ParserException e) {
      err.println(file + ":" + beautifyParseError(e, lexer.getLine()));
      errorFound = true;
      parsers.close();
    } catch (Exception e) {
      e.printStackTrace();
      errorFound = true;
    }

    if (verboseMode()) {
      console.parseFile(lastTime, file.getPath());
    }
    return null;
  }

  /**
   * Shows an error/warning and set error flag to true
   */
  protected CJavaPrettyPrinter getPrettyPrinterOverided(String fileName) {
    return new CJavaPrettyPrinter(fileName);
  }

  /**
   * check that interface of a given compilation unit is correct
   * side effect: increment error number
   * @param	cunit		the compilation unit
   */
  protected void checkInterface(JCompilationUnit cunit) {
    try {
      cunit.checkInterface(this);
    } catch (PositionedError e) {
      reportTrouble(e);
    }
  }

  /**
   * check that interface of a given compilation unit is correct
   * side effect: increment error number
   * @param	cunit		the compilation unit
   */
  protected void checkInitializers(JCompilationUnit cunit) {
    try {
      cunit.checkInitializers(this, classes);
    } catch (PositionedError e) {
      reportTrouble(e);
    }
  }

  /**
   * check that body of a given compilation unit is correct
   * side effect: increment error number
   * @param	cunit		the compilation unit
   */
  protected void checkBody(JCompilationUnit cunit) {
    long	lastTime = System.currentTimeMillis();
    try {
      cunit.checkBody(this, classes);
    } catch (PositionedError e) {
      reportTrouble(e);
    }
    if (verboseMode()) {
      console.checkBody(lastTime, cunit.getFileName());
    }
  }

  /**
   * generate the source code of parsed compilation unit
   * @param	cunit		the compilation unit
   */
  protected void generateJavaCode(JCompilationUnit cunit) {
    if (verboseMode()) {
      console.generateJavaCode(cunit.getFileName());
    }
    cunit.genJavaCode(destination);
  }

  /**
   * Parse each file in multi thread
   */
  protected void parseMultiProc(final JCompilationUnit[] tree) {

    try {
      Thread[]		parsers = new Thread[nbProc];
      int		length = tree.length / nbProc;

      for (int i = 0; i < nbProc; i++) {
	parsers[i] = new ThreadParser(this, infiles, tree, i * length, i == nbProc - 1 ? tree.length : (i + 1) * length);
	parsers[i].start();
      }

      for (int i = 0; i < nbProc; i++) {
	parsers[i].join();
      }
    } catch (java.lang.InterruptedException ie) {
      ie.printStackTrace();
    }
  }

  /**
   *
   */
  protected void checkBodyMultiProc(final JCompilationUnit[] tree) {
    try {
      ThreadChecker[]	checkers = new ThreadChecker[tree.length];
      int		length = tree.length / nbProc;

     for (int i = 0; i < nbProc; i++) {
       checkers[i] = new ThreadChecker(this, tree, i * length, i == nbProc - 1 ? tree.length : (i + 1) * length);
       checkers[i].start();
      }

      for (int i = 0; i < nbProc; i++) {
	checkers[i].join();
      }
    } catch (java.lang.InterruptedException ie) {
      ie.printStackTrace();
    }
  }

  /**
   *
   */
  protected void genJavaCodeMultiProc(final JCompilationUnit[] tree) {
    try {
      Thread[]		writers = new Thread[tree.length];
      int		length = tree.length / nbProc;

      for (int i = 0; i < nbProc; i++) {
	writers[i] = new JavaThreadWriter(destination, tree, i * length, i == nbProc - 1 ? tree.length : (i + 1) * length);
	writers[i].start();
      }

      for (int i = 0; i < nbProc; i++) {
	writers[i].join();
      }
    } catch (java.lang.InterruptedException ie) {
      ie.printStackTrace();
    }
  }

  /**
   *
   */
  protected void genCodeMultiProc(final CSourceClass[] tree, final File destination) {
    try {
      ThreadWriter[]	writers = new ThreadWriter[tree.length];
      int		length = tree.length / nbProc;

     for (int i = 0; i < nbProc; i++) {
       writers[i] = new ThreadWriter(tree, destination, i * length, i == nbProc - 1 ? tree.length : (i + 1) * length);
       writers[i].start();
      }

      for (int i = 0; i < nbProc; i++) {
	writers[i].join();
      }
    } catch (java.lang.InterruptedException ie) {
      ie.printStackTrace();
    }
  }

  // --------------------------------------------------------------------
  // COMPILER
  // --------------------------------------------------------------------

  /**
   * Reports a trouble (error or warning).
   *
   * @param	trouble		a description of the trouble to report.
   */
  public void reportTrouble(PositionedError trouble) {
    if (trouble instanceof CWarning) {
      if ((optionWarning) && ((CWarning)trouble).isEnabled()) {
	err.println(trouble.getErrorMessage());
      }
    } else {
      if (trouble.getTokenReference() != TokenReference.NO_REF) {
	err.println(trouble.getErrorMessage());
	errorFound = true;
      }
    }
  }

  /**
   * Returns true iff comments should be parsed (false if to be skipped)
   */
  public boolean parseComments() {
    return optionDeprecation || (optionBeautify && !optionNoWrite);
  }

  /**
   * Returns the classes to generate
   */
  public CSourceClass[] getClasses() {
    return (CSourceClass[])at.dms.compiler.Utils.vectorToArray(classes, CSourceClass.class);
  }

  // ----------------------------------------------------------------------
  // PROTECTED METHODS
  // ----------------------------------------------------------------------

  /**
   * Generate an human readable error message
   */
  protected static String beautifyParseError(antlr.ParserException e, int line) {
    String temp = e.toString(); // can do better
    if (temp == null) {
      temp = "unknown";
    }

    int		idx = temp.indexOf(",");
    if (idx >= 0) {
      temp = temp.substring(idx + 1);
    }

    return line + ":" + temp;
  }

  // ----------------------------------------------------------------------
  // PROTECTED DATA MEMBERS
  // ----------------------------------------------------------------------

  protected static Main		instance;		// the sub instance

  protected Vector		infiles			= new Vector();
  protected boolean		errorFound;
  protected String		classpath;
  protected long		timestamp;

  // options
  protected String		destination;
  protected boolean		optionVerbose		= false;
  protected boolean		optionGenerateJava	= false;
  protected boolean		optionBeautify		= false;
  protected boolean		optionNoWrite		= false;
  protected boolean		optionWarning		= false;
  protected int			optionOptimize		= 1;
  protected boolean		optionMultiThread	= false;
  protected boolean		optionDeprecation	= false;
  protected boolean		optionFast		= false;
  protected static int		nbProc			= 2;
  protected static PrintWriter	err			= new PrintWriter(System.err);
  protected Console		console;

  // all generated classes
  private Vector		classes = new Vector(100);

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private static final String	VERSION			= "1.3D";

  public static final String	SHORTOPTS = "bd:ghjmnO::Vvw";
  public static final
    LongOpt[]			LONGOPTS = {
      new LongOpt("beautify",		LongOpt.NO_ARGUMENT,		null, 'b'),
      new LongOpt("classpath",		LongOpt.REQUIRED_ARGUMENT,	null, 500),
      new LongOpt("debug",		LongOpt.NO_ARGUMENT,		null, 'g'),
      new LongOpt("destination",	LongOpt.REQUIRED_ARGUMENT,	null, 'd'),
      new LongOpt("deprecation",	LongOpt.NO_ARGUMENT,		null, 501),
      new LongOpt("help",		LongOpt.NO_ARGUMENT,		null, 'h'),
      new LongOpt("java",		LongOpt.NO_ARGUMENT,		null, 'j'),
      new LongOpt("multi",		LongOpt.NO_ARGUMENT,		null, 'm'),
      new LongOpt("O",			LongOpt.OPTIONAL_ARGUMENT,	null, 'O'),
      new LongOpt("nowrite",		LongOpt.NO_ARGUMENT,		null, 'n'),
      new LongOpt("verbose",		LongOpt.NO_ARGUMENT,		null, 'v'),
      new LongOpt("version",		LongOpt.NO_ARGUMENT,		null, 'V'),
      new LongOpt("warning",		LongOpt.NO_ARGUMENT,		null, 'w'),
      new LongOpt("nowarn",		LongOpt.NO_ARGUMENT,		null, 503)
    };
}

// deported here because of a javac bug
class ThreadParser extends Thread {
  ThreadParser(Main compiler, Vector infiles, JCompilationUnit[] tree, int start, int end) {
    this.compiler = compiler;
    this.start = start;
    this.end = end;
    this.tree = tree;
    this.infiles = infiles;
  }

  public void run() {
    for (int count = start; count < end; count++) {
      tree[count] = compiler.parseFile((File)infiles.elementAt(count));
    }
  }

  private Main			compiler;
  private int			start;
  private int			end;
  private JCompilationUnit[]	tree;
  private Vector		infiles;
}

class ThreadChecker extends Thread {
  ThreadChecker(Main compiler, JCompilationUnit[] tree, int start, int end) {
    this.compiler = compiler;
    this.start = start;
    this.end = end;
    this.tree = tree;
  }

  public void run() {
    for (int count = start; count < end; count++) {
      compiler.checkBody(tree[count]);
      tree[count] = null;
    }
  }

  private Main			compiler;
  private int			start;
  private int			end;
  private JCompilationUnit[]	tree;
}

class JavaThreadWriter extends Thread {
  JavaThreadWriter(String destination, JCompilationUnit[] tree, int start, int end) {
    this.start = start;
    this.end = end;
    this.tree = tree;
    this.destination = destination;
  }

  public void run() {
    for (int count = start; count < end; count++) {
      tree[count].genJavaCode(destination);
    }
  }

  private int			start;
  private int			end;
  private JCompilationUnit[]	tree;
  private String		destination;
}

class ThreadWriter extends Thread {
  ThreadWriter(CSourceClass[] tree, File destination, int start, int end) {
    this.start = start;
    this.end = end;
    this.tree = tree;
    this.destination = destination;
  }

  public void run() {
    try {
      for (int count = start; count < end; count++) {
	tree[count].genCode(destination);
	tree[count] = null;
      }
    } catch (java.io.IOException ioe) {
      System.err.println(ioe.getMessage());
    }
  }

  private int			start;
  private int			end;
  private CSourceClass[]	tree;
  private File			destination;
}
