/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Assembler.java,v 1.11 1999/10/08 11:19:02 vgp Exp $
 */

package at.dms.ksm;

import java.io.*;

import at.dms.classfile.ClassFileFormatException;
import at.dms.classfile.ClassInfo;
import at.dms.compiler.Utils;

/**
 * This class is the entry point for the KOPI assembler.
 */
public class Assembler {

  // --------------------------------------------------------------------
  // CONSTRUCTORS
  // --------------------------------------------------------------------

  // --------------------------------------------------------------------
  // ENTRY POINT
  // --------------------------------------------------------------------

  // --------------------------------------------------------------------
  // ACCESSORS
  // --------------------------------------------------------------------

  /**
   * Assembles an assembly language file to a class file
   *
   * @param	sourceFile		the name of the assembly language file
   * @param	destination		the directory root to place the classfile
   * @exception	AsmError		an error occurred
   */
  public void assemble(String sourceFile, String destination) throws AsmError {
    ClassInfo		classInfo;

    classInfo = parseInput(sourceFile);
    writeClassfile(classInfo, destination);
  }

  /**
   * Parses an assembly language file and creates a class info structure from it
   *
   * @param	sourceFile		the name of the source file
   * @return	a class info structure holding the information from the source
   */
  private ClassInfo parseInput(String sourceFile) throws AsmError {
    try {
      InputStream	input = new BufferedInputStream(new FileInputStream(sourceFile));
      KSMLexer		lexer = new KSMLexer(input);
      KSMParser		parser = new KSMParser(lexer);
      ClassInfo		classInfo = parser.aCompilationUnit();

      input.close();

      return classInfo;
    } catch (java.io.FileNotFoundException e) {
      throw new AsmError(null, "unknown-file", sourceFile);
    } catch (java.io.IOException e) {
      throw new AsmError(null, "io-exception", sourceFile, e.getMessage());
    } catch (AsmError e) {
      // caught and rethrown because it is a subclass of antl.ParserException
      throw e;
    } catch (antlr.ParserException e) {
      throw new AsmError(null, "parsing-error",
			 sourceFile,
			 e.getLine() + " : " + e.getErrorMessage());
    }
  }

  /**
   * Creates a class file from class info
   */
  private void writeClassfile(ClassInfo classInfo, String destination)
    throws AsmError
  {
    String[]	classPath = Utils.splitQualifiedName(classInfo.getName());

    if (!classPath[0].equals("")) {		// the class is part of a package
      String	classDir = classPath[0].replace('/', File.separatorChar);

      if (destination != null) {
	destination += File.separator + classDir;
      }
    }

    if (destination != null) {		// check that destination exists or else create it
      File	destDir = new File(destination);

      if (! destDir.exists()) {
	destDir.mkdirs();
      }

      if (! destDir.isDirectory()) {
	throw new AsmError(null, "can-t-create", destination);
      }
    }

    File	outputFile = new File(destination, classPath[1] + ".class");

    try {
      DataOutputStream		out =
	new DataOutputStream(new BufferedOutputStream(new FileOutputStream(outputFile)));

      classInfo.write(out);
      out.flush();
      out.close();

      //System.out.println(Message.format("generated", new Object[] { outputFile.getPath() }));
    } catch (ClassFileFormatException e) {
      throw new AsmError(null, "semantic-error", outputFile.getPath(), e.getMessage());
    } catch (java.io.IOException e) {
      throw new AsmError(null, "io-exception", outputFile.getPath(), e.getMessage());
    }
  }

  // --------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // --------------------------------------------------------------------

}
