/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Main.java,v 1.9 1999/10/16 23:25:35 vgp Exp $
 */

package at.dms.lexgen;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.Hashtable;
import java.util.Vector;

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;

import at.dms.jperf.JPerf;
import at.dms.util.Message;

/**
 * This class is the entry point for the KOPI assembler.
 */
public class Main {

  // --------------------------------------------------------------------
  // CONSTRUCTORS
  // --------------------------------------------------------------------

  /**
   * Only main can construct Main
   */
  private Main(String[] args) {
    parseArgs(args);

    parseSourceFiles();
    checkIdentifiers();
    if (buildDefinition) {
      buildDefinitionFile();
    }
    if (buildInterface) {
      buildInterfaceFile();
    }
    if (buildKeyword) {
      buildKeywordFile();
    }
    System.exit(0);
  }

  /**
   *
   */
  private void parseSourceFiles() {
    boolean		errorsFound = false;

    definitions = new DefinitionFile[sourceFiles.length];
    for (int i = 0; i < sourceFiles.length; i++) {
      try {
	definitions[i] = DefinitionFile.read(sourceFiles[i]);
      } catch (LexgenError e) {
	System.err.println(e.getMessage());
	errorsFound = true;
      }
    }

    if (errorsFound) {
      System.exit(1);
    }
  }


  /**
   *
   */
  private void checkIdentifiers() {
    Hashtable		identifiers = new Hashtable();
    boolean		errorsFound = false;
    String		prefix = definitions[definitions.length - 1].getPrefix();
    int			cnt = MIN_TOKEN_ID;

    for (int i = 0; i < definitions.length; i++) {
      try {
	cnt = definitions[i].checkIdentifiers(identifiers, prefix, cnt);
      } catch (LexgenError e) {
	System.err.println(e.getMessage());
	errorsFound = true;
      }
    }

    if (errorsFound) {
      System.exit(1);
    }
  }

  /**
   *
   */
  private void buildDefinitionFile() {
    String		vocabulary = definitions[definitions.length - 1].getVocabulary();
    String		prefix = definitions[definitions.length - 1].getPrefix();
    File		outputFile = new File(vocabulary + "TokenTypes.txt");

    try {
      PrintWriter	out = new PrintWriter(new BufferedWriter(new FileWriter(outputFile)));

      // !!! print file header

      // print vocabulary name
      out.println(vocabulary + " // output vocabulary");

      for (int i = 0; i < definitions.length; i++) {
	definitions[i].printDefinition(out, prefix);
      }

      out.flush();
      out.close();
    } catch (java.io.IOException e) {
      System.err.println("I/O Exception on " + outputFile.getPath() + ": " + e.getMessage());
      System.exit(1);
    }
  }

  /**
   *
   */
  private void buildInterfaceFile() {
    DefinitionFile	definition = definitions[definitions.length - 1];
    String		vocabulary = definition.getVocabulary();
    File		outputFile = new File(vocabulary + "TokenTypes.java");

    try {
      PrintWriter	out = new PrintWriter(new BufferedWriter(new FileWriter(outputFile)));

      // !!! print file header

      definition.printInterface(out,
				definitions.length == 1
				? null
				: definitions[definitions.length - 2].getClassName());

      out.flush();
      out.close();
    } catch (java.io.IOException e) {
      System.err.println("I/O Exception on " + outputFile.getPath() + ": " + e.getMessage());
      System.exit(1);
    }
  }

  /**
   *
   */
  private void buildKeywordFile() {
    String		vocabulary = definitions[definitions.length - 1].getVocabulary();
    String		prefix = definitions[definitions.length - 1].getPrefix();
    String		packageName = definitions[definitions.length - 1].getPackageName();

    Vector		keywords = new Vector();
    Vector		types = new Vector();
    Vector		header = new Vector();
    Vector		footer = new Vector();

    // retrieve keywords to hash
    for (int i = 0; i < definitions.length; i++) {
      definitions[i].putKeywords(keywords, types, prefix);
    }

    // header
    header.addElement("// Generated by JPerf");
    header.addElement("package " + packageName + ";");
    header.addElement("import at.dms.compiler.CToken;\n");
    header.addElement("/*package*/ final class " + vocabulary + "Keywords implements " + vocabulary + "TokenTypes {");
    header.addElement("  public static CToken lookup(final char[] data, int offset, int length) {");
    header.addElement("    return tokens[find(data, offset, length) + 1];");
    header.addElement("  }");
    header.addElement("  private static final CToken[] tokens = new CToken[] {\n    null,");
    for (int i = 0; i < types.size(); i++) {
      header.addElement("    new CToken(" + (String)types.elementAt(i) + ", \"" + (String)keywords.elementAt(i) + "\"),");
    }
    header.addElement("  };");

    // footer
    footer.addElement("}");

    // generate perfect hash function
    JPerf		perf = new JPerf(makeStringArray(keywords),
					 makeStringArray(header),
					 makeStringArray(footer));
    perf.build();

    try {
      perf.genCode(vocabulary + "Keywords.java");
    } catch (java.io.IOException e) {
      System.err.println("I/O Exception on " + vocabulary + "Keywords.java" + ": " + e.getMessage());
      System.exit(1);
    }
  }

  /**
   * Transforms a vector of strings into a string array.
   */
  private static String[] makeStringArray(Vector vec) {
    String[]	arr;

    arr = new String[vec.size()];
    vec.copyInto(arr);
    return arr;
  }

  // --------------------------------------------------------------------
  // ENTRY POINT
  // --------------------------------------------------------------------

  /**
   * Entry point to the assembler
   */
  public static void main(String[] args) {
    Message.addBundle("lexgen");

    new Main(args);
  }

  // --------------------------------------------------------------------
  // ACCESSORS
  // --------------------------------------------------------------------

  /*
   * Parse command line arguments.
   */
  private void parseArgs(String[] args) {
    Getopt	g = new Getopt("lexgen", args, shortopts, longopts, true);

  _loop_:
    for (;;) {
      switch (g.getopt()) {
      case 'd':
	buildDefinition = true;
	break;

      case 'i':
	buildInterface = true;
	break;

      case 'k':
	buildKeyword = true;
	break;

      case 'h':
	printUsage();
	System.exit(0);

      case 'V':
	printVersion();
	break;

      case -1:
	break _loop_;
      }
    }

    if (! buildDefinition && !buildInterface && !buildKeyword) {
      printUsage();
      System.exit(1);
    }

    if (args.length == g.getOptind()) {
      printUsage();
      System.exit(1);
    }

    this.sourceFiles = new String[args.length - g.getOptind()];
    System.arraycopy(args, g.getOptind(), this.sourceFiles, 0, this.sourceFiles.length);
  }

  private static void printUsage() {
    System.err.println(Message.format("usage-line-lexgen", null));
    System.err.println();
    System.err.println(Message.format("info1", null));
    System.err.println(Message.format("info2", null));
    System.err.println(Message.format("info3", null));
    System.err.println(Message.format("info4", null));
  }

  private static void printVersion() {
    System.err.println("KOPI Token Definition Generator Version: " + VERSION);
  }

  // --------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // --------------------------------------------------------------------

  private static final String	VERSION		= "1.3D";

  private static final String	shortopts = "dhikV";
  private static final
    LongOpt[]			longopts = {
      new LongOpt("definition", LongOpt.NO_ARGUMENT, null, 'd'),
      new LongOpt("interface", LongOpt.NO_ARGUMENT, null, 'i'),
      new LongOpt("keywords", LongOpt.NO_ARGUMENT, null, 'k'),
      new LongOpt("help", LongOpt.NO_ARGUMENT, null, 'h'),
      new LongOpt("version", LongOpt.NO_ARGUMENT, null, 'V')
    };

  // token ids start with this number
  private static final int	MIN_TOKEN_ID = 4;

  private boolean		buildDefinition;
  private boolean		buildInterface;
  private boolean		buildKeyword;
  private String[]		sourceFiles;
  private DefinitionFile[]	definitions;
}
