/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: InstructionHandle.java,v 1.3 1999/10/07 08:59:47 vgp Exp $
 */

package at.dms.optimize;

import java.util.Enumeration;
import java.util.Vector;
import at.dms.classfile.AccessorContainer;
import at.dms.classfile.AbstractInstructionAccessor;
import at.dms.classfile.Instruction;
import at.dms.classfile.InstructionAccessor;
import at.dms.classfile.JumpInstruction;
import at.dms.classfile.LineNumberInfo;
import at.dms.classfile.SwitchInstruction;
import at.dms.classfile.HandlerInfo;

/**
 * 
 */
class InstructionHandle extends AbstractInstructionAccessor {

  // --------------------------------------------------------------------
  // CONSTRUCTORS
  // --------------------------------------------------------------------

  /**
   * Creates a new instruction handle.
   *
   * @param	insn		the instruction
   * @param	prev		the handle of the next instruction
   *				in textual order
   */
  public InstructionHandle(Instruction insn, InstructionHandle prev) {
    this.insn = insn;
    this.prev = prev;
    this.next = null;

    if (prev != null) {
      prev.next = this;
    }
  }

  // --------------------------------------------------------------------
  // ACCESSORS
  // --------------------------------------------------------------------

  /**
   * Returns the enclosed instruction.
   */
  public Instruction getInstruction() {
    return insn;
  }

  /**
   * Returns the handle of the next instruction in textual order.
   */
  public InstructionHandle getNext() {
    return next;
  }

  /**
   * Returns the handle of the next instruction in textual order.
   */
  public InstructionHandle getPrevious() {
    return prev;
  }

  /**
   * Sets the handle of the next instruction in textual order.
   */
  public void setNext(InstructionHandle next) {
    this.next = next;
  }

  /**
   * Notifies this handle that is has been attached to the specified container.
   */
  public void attachTo(AccessorContainer container) {
    if (container instanceof LineNumberInfo) {
      if (lineNumbers == null) {
	lineNumbers = new Vector();
      }
      lineNumbers.addElement(new Integer(((LineNumberInfo)container).getLine()));
    }
    //!!! nothing to do for now
  }

  /**
   * isReached
   */
  public boolean isReached() {
    return reached;
  }

  /**
   * isTarget
   */
  public boolean isTarget() {
    return accessors != null;
  }

  /**
   * isReached
   */
  public void set() {
    reached = prev == null || 
      accessors != null ||
      (prev.isReached() && prev.getInstruction().canComplete());
  }

  /**
   * remove
   */
  public boolean remove() {
    if (accessors != null) {
      if (getNext() == null) {
	return false;
      }
      for (int i = 0; i < accessors.size(); i++) {
	changeTarget(((AccessorContainer)accessors.elementAt(i)), next);
      }
    }
    
    if (prev != null) {
      prev.next = next;
    }
    if (next != null) {
      next.prev = prev;
    }

    return true;
  }

  /**
   * replaceBy
   */
  public void replaceBy(InstructionHandle handle) {
    if (accessors != null) {
      for (int i = 0; i < accessors.size(); i++) {
	changeTarget(((AccessorContainer)accessors.elementAt(i)), handle);
      }
    }

    handle.next = next;
    handle.set();
  }

  /**
   * reset
   */
  public void reset() {
    if (insn instanceof AccessorContainer) {
      if (insn instanceof JumpInstruction) {
	((InstructionHandle)((JumpInstruction)insn).getTarget()).addAccessor((JumpInstruction)insn); // !!!
      } else if (insn instanceof SwitchInstruction) {
	SwitchInstruction	current = (SwitchInstruction)insn;

	for (int i = -1; i < current.getSwitchCount(); i++) {
	  ((InstructionHandle)current.getTarget(i)).addAccessor(current);
	}
      } else {
	throw new RuntimeException(">>>>>>>" + insn);
      }
    }
  }

  /**
   * reset
   */
  public void clean() {
    reached = false;
    
    if (accessors != null) {
      at.dms.util.VectorCache.release(accessors);
      accessors = null;
    }
  }

  public void addAccessor(AccessorContainer accessor) {
    if (accessors == null) {
      accessors = at.dms.util.VectorCache.request(at.dms.util.VectorCache.SMALL);
    }

    accessors.addElement(accessor);
  }

  public void removeAccessor(AccessorContainer accessor) {
    accessors.removeElement(accessor);
    
    if (accessors.size() == 0) {
      at.dms.util.VectorCache.release(accessors);
      accessors = null;
    }
  }

  // --------------------------------------------------------------------
  // RECONSTRUCT A CodeInfo STRUCTURE
  // --------------------------------------------------------------------

  /**
   * Adds line number info stored for the instruction to the specified vector.
   */
  public void addLineNumberInfo(Vector lineNumberInfo) {
    if (lineNumbers != null) {
      for (Enumeration enum = lineNumbers.elements(); enum.hasMoreElements(); ) {
	int	line = ((Integer)enum.nextElement()).intValue();

	lineNumberInfo.addElement(new LineNumberInfo((short)line, this));
      }
    }
  }

  // --------------------------------------------------------------------
  // PRIVATE UTILITIES
  // --------------------------------------------------------------------

  private void changeTarget(AccessorContainer container, InstructionHandle target) {
    if (container instanceof JumpInstruction) {
      ((JumpInstruction)container).setTarget(target);
      target.addAccessor(container);
    } else if (container instanceof SwitchInstruction) {
      SwitchInstruction	current = (SwitchInstruction)container;

      for (int i = -1; i < current.getSwitchCount(); i++) {
	if (current.getTarget(i) == this) {
	  current.setTarget(i, target);
	}
      }
      target.addAccessor(container);
    } else if (container instanceof HandlerInfo) {
      HandlerInfo	handler = (HandlerInfo)container;

      if (handler.getStart() == this) {
	handler.setStart(target);
      }
      if (handler.getEnd() == this) {
	handler.setEnd(target);
      }
      if (handler.getHandler() == this) {
	handler.setHandler(target);
      }
    } else {
      throw new RuntimeException(">>>>>>>>>>>" + container);
    }
  }

  // --------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // --------------------------------------------------------------------

  private final Instruction	insn;
  private boolean		reached;

  private InstructionHandle	prev;
  private InstructionHandle	next;

  private Vector		lineNumbers;
  private Vector		accessors;
}
