/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Main.java,v 1.6 1999/10/16 23:25:36 vgp Exp $
 */

package at.dms.optimize;

import java.io.*;

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;

import at.dms.classfile.ClassFileFormatException;
import at.dms.classfile.ClassInfo;
import at.dms.classfile.CodeInfo;
import at.dms.classfile.MethodInfo;
import at.dms.util.Message;
import at.dms.compiler.UnpositionedError; // !!!

/**
 * This class is the entry point for the KOPI disassembler.
 */
public class Main {

  // --------------------------------------------------------------------
  // CONSTRUCTORS
  // --------------------------------------------------------------------

  /**
   * Only main can construct Main
   */
  private Main(String[] args) {
    parseArgs(args);

    if (sourceFiles == null) {
      printUsage();
      System.exit(1);
    } else {
      boolean	errorsFound = false;

      for (int i = 0; i < sourceFiles.length; i++) {
	if (optionVerbose) {
	  System.err.println("Processing " + sourceFiles[i] + ".");
	}

	try {
	  optimizeClass(sourceFiles[i]);
	} catch (UnpositionedError e) {
	  System.err.println("Error: " + e.getMessage());
	  errorsFound = true;
	}
      }

      System.exit(errorsFound ? 1 : 0);
    }
  }

  // --------------------------------------------------------------------
  // ACTIONS
  // --------------------------------------------------------------------

  /**
   * Reads, optimizes and writes a class file
   * @exception	UnpositionedError	an error occurred
   */
  private void optimizeClass(String fileName) throws UnpositionedError {
    ClassInfo		info;
    MethodInfo[]	methods;

    info = readClassFile(fileName);
    methods = info.getMethods();
    for (int i = 0; i < methods.length; i++) {
      optimizeMethod(methods[i]);
    }
    writeClassFile(info, fileName);
  }

  private void optimizeMethod(MethodInfo method) throws UnpositionedError {
    CodeInfo		code;

    code = method.getCodeInfo();
    if (code != null) {
      code = Optimizer.optimize(code, optionLevel);
      method.setCodeInfo(code);
    }
  }
    
  private ClassInfo readClassFile(String fileName) throws UnpositionedError {
    try {
      DataInputStream	in;
      ClassInfo		info;

      in = new DataInputStream(new BufferedInputStream(new FileInputStream(fileName), 2048));
      info = new ClassInfo(in, false);
      in.close();

      return info;
    } catch (ClassFileFormatException e) {
      throw new UnpositionedError("semantic-error", new Object[] { fileName, e.getMessage() });
    } catch (IOException e) {
      throw new UnpositionedError("io-exception", new Object[] { fileName, e.getMessage() });
    }
  }

  private void writeClassFile(ClassInfo info, String fileName) throws UnpositionedError {
    try {
      DataOutputStream	out;

      out = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(fileName)));
	
      info.write(out);
      out.flush();
      out.close();
    } catch (ClassFileFormatException e) {
      throw new UnpositionedError("semantic-error", new Object[] { fileName, e.getMessage() });
    } catch (IOException e) {
      e.printStackTrace();
      throw new UnpositionedError("io-exception", new Object[] { fileName, e.getMessage() });
    }
  }

  // --------------------------------------------------------------------
  // ENTRY POINT
  // --------------------------------------------------------------------

  /**
   * Entry point to the assembler
   */
  public static void main(String[] args) {
    try {
      Message.addBundle("optimize");

      new Main(args);
    } catch (Exception e) {
      System.err.println(e.getMessage());
      e.printStackTrace();
    }
  }

  // --------------------------------------------------------------------
  // ACCESSORS
  // --------------------------------------------------------------------

  /*
   * Parse command line arguments.
   */
  private void parseArgs(String[] args) {
    Getopt	g = new Getopt("optimize", args, shortopts, longopts, true);

  _loop_:
    for (;;) {
      switch (g.getopt()) {
      case 'h':
	printUsage();
	System.exit(0);

      case 'V':
	printVersion();
	break;

      case 'v':
	optionVerbose = true;
	break;

      case 'l':
	try {
	  optionLevel = Integer.parseInt(g.getOptarg());
	} catch (NumberFormatException e) {
	  System.err.println(Message.format("malformed-option", new Object[]{ g.getOptarg() }));
	  System.exit(1);
	}
	break;


      case -1:
	break _loop_;
      }
    }

    if (args.length > g.getOptind()) {
      this.sourceFiles = new String[args.length - g.getOptind()];
      System.arraycopy(args, g.getOptind(), this.sourceFiles, 0, this.sourceFiles.length);
    }
  }

  private static void printUsage() {
    System.err.println(Message.format("usage-line", null));
    System.err.println(Message.format("usage-version", null));
    System.err.println(Message.format("usage-verbose", null));
    System.err.println();
    System.err.println(Message.format("info1", null));
    System.err.println(Message.format("info2", null));
    System.err.println(Message.format("info3", null));
    System.err.println(Message.format("info4", null));
  }

  private static void printVersion() {
    System.err.println("KOPI optimizer - Version " + VERSION);
  }

  // --------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // --------------------------------------------------------------------

  private static final String	VERSION		= "1.3D";

  private static final String	shortopts = "hVl:v";
  private static final
    LongOpt[]			longopts = {
      new LongOpt("help", LongOpt.NO_ARGUMENT, null, 'h'),
      new LongOpt("version", LongOpt.NO_ARGUMENT, null, 'V'),
      new LongOpt("level", LongOpt.NO_ARGUMENT, null, 'l'),
      new LongOpt("verbose", LongOpt.NO_ARGUMENT, null, 'v')
    };

  private int			optionLevel = 1;
  private boolean		optionVerbose;
  private String[]		sourceFiles;
}
