/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Main.java,v 1.32 1999/10/12 15:39:49 vgp Exp $
 */

package at.dms.sqlc;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Vector;

import at.dms.compiler.HandLexer;
import at.dms.compiler.CWarning;
import at.dms.util.CharArrayCache;
import at.dms.compiler.Compiler;
import at.dms.compiler.Console;
import at.dms.compiler.InputBuffer;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.compiler.UnpositionedError;

/**
 * This class implements the entry point of the Java compiler
 */
public class Main extends Compiler {

  // ----------------------------------------------------------------------
  // RUN FROM COMMAND LINE
  // ----------------------------------------------------------------------

  /*
   * Runs a compilation session
   *
   * @param	dir		the working directory
   * @param	err		the error output stream
   * @param	console		the console to show the progess of the compilation (!!! not used)
   * @param	args		the arguments to the compiler
   * @return	true iff the compilation succeeded
   */
  public boolean run(String dir, PrintWriter err, Console console, String[] args) {
    this.err = err;
    //!!!FIXME this.workingDirectory = dir;
    //!!!FIXME this.console = console;

    return run(args);
  }

  /**
   * Runs the compiler
   */
  public boolean run(String[] args) {
    if (!parseArguments(args)) {
      return false;
    }

    initialize();

    if (infiles.size() == 0) {
      usage("no source file given");
      return false;
    }

    if (verboseMode()) {
      err.println("[ START COMPILATION IN VERBOSE MODE ]");
    }

    try {
      infiles = verifyFiles(infiles);
    } catch (UnpositionedError e) {
      err.println(e.getMessage());
      return false;
    }

    //!!!NOT USED destination = checkDestination(destination);

    SCompilationUnit[]	tree = new SCompilationUnit[infiles.size()];

    for (int count = 0; count < tree.length; count++) {
      tree[count] = parseFile(this, (File)infiles.elementAt(count));
    }

    if (errorFound) {
      return false;
    }

    for (int count = 0; count < tree.length; count++) {
      tree[count].check();
    }

    if (errorFound) {
      return false;
    }
    
    try {
      for (int count = 0; count < tree.length; count++) {
	tree[count].genSQL();
      }
    } catch (PositionedError e) {
      e.printStackTrace();
    }

    return true;
  }

  /**
   * parse the argument list and set flags
   * @return parsing error occur
   */
  protected boolean parseArguments(String[] args) {
    for (int i = 0; i < args.length; i += 1) {
      if (args[i].equals("-help")) {
	help();
	System.exit(0);
      } else if (args[i].startsWith("-")) {
	usage("unrecognised option " + args[i]);
	return false;
      } else if (args[i].endsWith(".sql")) {
	infiles.addElement(args[i]);
      } else {
	usage("unrecognised argument " + args[i]);
	return false;
      }
    }

    return true;
  }

  /**
   * Initialize the compiler (read classpath, check classes.zip)
   */
  protected void initialize() {
    at.dms.util.Message.addBundle("sqlc");

    XUtils.setChecker(new DBChecker() {
	/**
	 * Checks if table exists
	 */
	public boolean tableExists(String table) {
	  return true;
	}

	/**
	 * Checks if a column exists in a table
	 */
	public boolean columnExists(String table, String column) {
	  return true;
	}});
  }

  /**
   * shows an usage message
   */
  protected void usage(String msg) {
    err.println("sqlc: " + msg);
    err.println("usage: sqlc [-help] <sql-files>");
  }

  /**
   * shows a help message
   */
  protected void help() {
    err.println("usage: sqlc [-help] <sql-files>");
    err.println("a simple parser for SQL Statement");
    err.println("Not any semantics analysis done here");
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * @return true if compilation in verbose mode
   */
  public boolean verboseMode() {
    return false;
  }

  /**
   * Entry point
   */
  public static void main(String[] args) {
    if (instance != null) {
      // only one instance of Main at the same time
      throw new RuntimeException("DO NOT CALL MAIN WHILE THERE IS A COMPILE PROCESS");
    }

    boolean	success;

    instance = new Main();
    success = instance.run(args);
    instance = null;
    System.exit(success ? 0 : 1);
  }

  /**
   *
   */
  public Main getCurrentInstance() {
    return instance;
  }

  // --------------------------------------------------------------------
  // COMPILER
  // --------------------------------------------------------------------

  /**
   * Reports a trouble (error or warning).
   *
   * @param	trouble		a description of the trouble to report.
   */
  public void reportTrouble(PositionedError trouble) {
    if (trouble instanceof CWarning) {
      if ((optionWarning) && ((CWarning)trouble).isEnabled()) {
	err.println(trouble.getErrorMessage());
      }
    } else {
      if (trouble.getTokenReference() != TokenReference.NO_REF) {
	err.println(trouble.getErrorMessage());
	errorFound = true;
      }
    }
  }

  /**
   * Returns true iff comments should be parsed (false if to be skipped)
   */
  public boolean parseComments() {
    return false;
  }

  // ----------------------------------------------------------------------
  // PROTECTED METHODS
  // ----------------------------------------------------------------------

  /**
   * parse the givven file and return a compilation unit
   * side effect: increment error number
   * @param	file		the name of the file (assert exists)
   * @return	the compilation unit defined by this file
   */
  protected SCompilationUnit parseFile(Compiler compiler, File file) {
    InputBuffer			buffer;

    try {
      buffer = new InputBuffer(file.getName(), file);
    } catch (IOException e) {
      err.println("I/O error on " + file.getName() + ": " + e.getMessage());
      errorFound = true;
      return null;
    }

    // to get line number if errors
    HandLexer	lexer = new SQLScanner(compiler, buffer);

    // Create a parser that reads from the scanner
    SQLParser	parser = new SQLParser(compiler, lexer, file.getName());

    // start parsing at the compilationUnit rule
    try {
      return parser.sCompilationUnit();
    } catch (Exception e) {
      e.printStackTrace();
      err.println(file + ":" + lexer.getLine() + " " + e.getMessage());
      errorFound = true;
      return null;
    }
  }

  // ----------------------------------------------------------------------
  // PROTECTED DATA MEMBERS
  // ----------------------------------------------------------------------

  protected static Main			instance;		// the sub instance
  protected static Vector		infiles = new Vector();
  protected static boolean		errorFound;

  protected static PrintWriter	err = new PrintWriter(System.err);
  protected boolean		optionWarning		= false;
}
