/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: SqlChecker.java,v 1.9 1999/11/01 00:02:55 graf Exp $
 */

package at.dms.sqlc;

import java.util.Vector;
import java.util.Stack;
import java.sql.Date;
import java.sql.Time;
import java.sql.Timestamp;

import at.dms.compiler.PositionedError;
import at.dms.compiler.UnpositionedError;

import at.dms.compiler.CWarning;

/**
 * This class implements a Java pretty printer
 */
public class SqlChecker implements SqlVisitor {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Constructs a new visitor
   */
  public SqlChecker(SqlContext context) {
    this.current = new StringBuffer();
    enter(context);
  }

  // ----------------------------------------------------------------------
  // ACCESSORS
  // ----------------------------------------------------------------------

  /**
   * Returns the current string
   */
  public String toString() {
    return current.toString();
  }

  // ----------------------------------------------------------------------
  // VISITORS
  // ----------------------------------------------------------------------

  /*
   * Visits AndCondition
   */
  public void visitAndCondition(SearchCondition left, SearchCondition right)
    throws PositionedError, UnpositionedError
  {
    left.accept(this);
    current.append(" AND ");
    right.accept(this);
  }

  /*
   * Visits ArrayPrecision
   */
  public void visitArrayPrecision(IntegerLiteral size)
    throws PositionedError, UnpositionedError
  {
    current.append("(");
    if (size != null) {
      size.accept(this);
    } else {
      current.append("*");
    }
    current.append(")");
  }

  /*
   * Visits Assignment
   */
  public void visitAssignment(Expression left, Expression right)
    throws PositionedError, UnpositionedError
  {
    left.accept(this);
    current.append(" = ");
    right.accept(this);

    // check if left is a text and if it exists in the table
    if (left instanceof SimpleIdentExpression) {
      TableReference table = (TableReference)(getContext().getTables()).elementAt(0);
      if (!table.hasColumn(((SimpleIdentExpression)left).getIdent())) {
	getContext().reportTrouble(new CWarning(left.getTokenReference(), "x-column-not-resolvable", ((SimpleIdentExpression)left).getIdent()));
      }
    }
  }

  /*
   * Visits BetweenPredicate
   */
  public void visitBetweenPredicate(Expression start,
				    String startQualifier,
				    Expression end,
				    String endQualifier)
    throws PositionedError, UnpositionedError
  {
    current.append(" BETWEEN ");
    start.accept(this);
    if (startQualifier != null) {
      current.append(" ");
      current.append(startQualifier);
    }
    current.append(" AND ");
    end.accept(this);
    if (endQualifier != null) {
      current.append(" ");
      current.append(endQualifier);
    }
  }

  /*
   * Visits BinCharType
   */
  public void visitBinCharType(ArrayPrecision prec)
    throws PositionedError, UnpositionedError
  {
    current.append("BINCHAR");
    if (prec != null) {
      prec.accept(this);
    }
  }

  /*
   * Visits BinaryExpression
   */
  public void visitBinaryExpression(String operator, Expression left, Expression right)
    throws PositionedError, UnpositionedError
  {
    left.accept(this);
    current.append(" ");
    current.append(operator);
    current.append(" ");
    right.accept(this);
  }

  /*
   * Visits BincharLiteral
   */
  public void visitBincharLiteral(String value)
    throws PositionedError, UnpositionedError
  {
    current.append(value);
  }

  /*
   * Visits BitLiteral
   */
  public void visitBitLiteral(String value)
    throws PositionedError, UnpositionedError
  {
    current.append(value);
  }

  /*
   * Visits BitsType
   */
  public void visitBitsType(ArrayPrecision prec)
    throws PositionedError, UnpositionedError
  {
    current.append("BITS");
    prec.accept(this);
  }

  /*
   * Visits BlobType
   */
  public void visitBlobType()
    throws PositionedError, UnpositionedError
  {
    current.append("BLOB");
  }

  /*
   * Visits BoolType
   */
  public void visitBoolType()
    throws PositionedError, UnpositionedError
  {
    current.append("BOOL");
  }

  /*
   * Visits BooleanLiteral
   */
  public void visitBooleanLiteral(boolean value)
    throws PositionedError, UnpositionedError
  {
    current.append(value ? "TRUE" : "FALSE");
  }

  /*
   * Visits CaseExpression
   */
  public void visitCaseExpression(AbstractCaseExpression cases)
    throws PositionedError, UnpositionedError
  {
    current.append("CASE ");
    cases.accept(this);
    current.append(" END");
  }

  /*
   * Visits CastPrimary
   */
  public void visitCastPrimary(Expression left, Type right, boolean hasAs)
    throws PositionedError, UnpositionedError
  {
    if (hasAs) {
      current.append(" CAST (");
      left.accept(this);
      current.append(" AS ");
      right.accept(this);
      current.append(")");
    } else {
      left.accept(this);
      if (right != null) {
	current.append(" CAST ");
	right.accept(this);
      }
    }
  }

  /*
   * Visits CharType
   */
  public void visitCharType(ArrayPrecision prec)
    throws PositionedError, UnpositionedError
  {
    current.append("CHAR");
    if (prec != null) {
      prec.accept(this);
    }
  }

  /*
   * Visits ComparaisonPredicate
   */
  public void visitComparaisonPredicate(Predicate left, Predicate right, String operator)
    throws PositionedError, UnpositionedError
  {
    left.accept(this);
    current.append(" ");
    current.append(operator);
    current.append(" ");
    right.accept(this);
  }

  /*
   * Visits CorrespondingSpec
   */
  public void visitCorrespondingSpec(FieldNameList fields)
    throws PositionedError, UnpositionedError
  {
    current.append("CORRESPONDING");
    if (fields != null) {
      current.append(" BY (");
      fields.accept(this);
      current.append(")");
    }
  }

  /*
   * Visits CountExpression
   */
  public void visitCountExpression(boolean hasDistinct, Expression expr)
    throws PositionedError, UnpositionedError
  {
    if (hasDistinct) {
      current.append("COUNT(DISTINCT ");
      expr.accept(this);
      current.append(")");
    } else {
      current.append("COUNT(*)");
    }
  }

  /*
   * Visits DefaultValuesInsertSource
   */
  public void visitDefaultValuesInsertSource()
    throws PositionedError, UnpositionedError
  {
    current.append("DEFAULT VALUES");
  }

  /*
   * Visits DeleteStatement
   */
  public void visitDeleteStatement(String schemaName, String tableName, SearchCondition searchCondition)
    throws PositionedError, UnpositionedError
  {
    current.append("DELETE FROM ");
    current.append(schemaName);
    if (tableName != null) {
      current.append(" ");
      current.append(tableName);
    }
    if (searchCondition != null) {
      current.append(" ");
      current.append("WHERE ");
      searchCondition.accept(this);
    }

    // the table exists ?
    if (!XUtils.tableExists(schemaName)) {
      getContext().reportTrouble(new CWarning(searchCondition.getTokenReference(), "x-table-not-found", schemaName));
    }

  }

  /*
   * Visits DoubleType
   */
  public void visitDoubleType()
    throws PositionedError, UnpositionedError
  {
    current.append("DOUBLE");
  }

  /*
   * Visits ExistsPredicate
   */
  public void visitExistsPredicate(SubTableExpression sub)
    throws PositionedError, UnpositionedError
  {
    current.append("EXISTS (");
    sub.accept(this);
    current.append(")");
  }

  /*
   * Visits ExpressionList
   */
  public void visitExpressionList(Vector exprs)
    throws PositionedError, UnpositionedError
  {
    for (int i = 0; i < exprs.size(); i++) {
      if (i != 0) {
	current.append(", ");
      }
      ((Expression)exprs.elementAt(i)).accept(this);
    }
  }

  /*
   * Visits ExpressionPredicate
   */
  public void visitExpressionPredicate(Expression expr)
    throws PositionedError, UnpositionedError
  {
    expr.accept(this);
  }

  /*
   * Visits ExtendedPredicate
   */
  public void visitExtendedPredicate(Predicate left,  String operator, Predicate right)
    throws PositionedError, UnpositionedError
  {
    left.accept(this);
    current.append(" ");
    if (operator != null) {
      current.append(operator);
      current.append(" ");
    }
    right.accept(this);
  }

  /*
   * Visits FieldNameList
   */
  public void visitFieldNameList(Vector exprs)
    throws PositionedError, UnpositionedError
  {
    for (int i = 0; i < exprs.size(); i++) {
      if (i != 0) {
	current.append(", ");
      }
      current.append(exprs.elementAt(i));
    }
  }

  /*
   * Visits FieldReference
   */
  public void visitFieldReference(String table, String ident)
    throws PositionedError, UnpositionedError
  {
    if (table != null) {
      current.append(table);
      current.append(".");
    }
    current.append(ident);

    // !!! checkColumnExists(getTokenReference(), getContext(), table, ident);
  }

  /*
   * Visits FieldReferencePrimary
   */
  public void visitFieldReferencePrimary(Expression fieldRef, Expression begin, Expression end)
    throws PositionedError, UnpositionedError
  {
    fieldRef.accept(this);
    if (begin!= null) {
      current.append(" SUBRANGE(");
      begin.accept(this);
      current.append(", ");
      end.accept(this);
      current.append(") ");
    }
  }

  /*
   * Visits FloatType
   */
  public void visitFloatType()
    throws PositionedError, UnpositionedError
  {
    current.append("FLOAT");
  }

  /*
   * Visits FromClause
   */
  public void visitFromClause(Vector tableRefs)
    throws PositionedError, UnpositionedError
  {
    current.append("FROM ");
    for (int i = 0; i < tableRefs.size(); i++) {
      if (i != 0) {
	current.append(", ");
      }
      ((SimpleTableReference)tableRefs.elementAt(i)).accept(this);
    }

  }

  /*
   * Visits GroupByClause
   */
  public void visitGroupByClause(Vector fields)
    throws PositionedError, UnpositionedError
  {
    current.append("GROUP BY ");
    for (int i = 0; i < fields.size(); i++) {
      if (i != 0) {
	current.append(", ");
      }
      ((FieldReference)fields.elementAt(i)).accept(this);
    }
    current.append(" ");
  }

  /*
   * Visits HavingClause
   */
  public void visitHavingClause(SearchCondition condition)
    throws PositionedError, UnpositionedError
  {
    current.append("HAVING");
    condition.accept(this);
    current.append(" ");
  }

  /**
   * Visits a string
   * @param	text	text
   */
  public void visitIdent(String text)
    throws PositionedError, UnpositionedError
  {
    current.append(text);
  }

  /*
   * Visits IfExpression
   */
  public void visitIfExpression(SearchCondition condition, Expression thenExpr, Expression elseExpr)
    throws PositionedError, UnpositionedError
  {
    current.append("IF ");
    condition.accept(this);

    current.append(" ");
    current.append("THEN ");
    thenExpr.accept(this);

    current.append(" ");
    current.append("ELSE ");
    elseExpr.accept(this);

    current.append(" ");
    current.append("FI");
  }

  /*
   * Visits InsertStatement
   */
  public void visitInsertStatement(String ident, FieldNameList fields, InsertSource source)
    throws PositionedError, UnpositionedError
  {
    current.append("INSERT INTO ");
    current.append(ident);
    if (fields != null) {
      current.append("(");
      fields.accept(this);
      current.append(")");
    }
    current.append(" ");
    source.accept(this);

    if (!XUtils.tableExists(ident)) {
      getContext().reportTrouble(new CWarning(source.getTokenReference(), "x-table-not-found", ident));
    } else {
      // !!! TESTER POUR LES AUTRES CAS
      //if (source instanceof TableInsertSource) {
      if ((source instanceof ValueListInsertSource) && (fields != null)) {
	if (((ValueListInsertSource)source).elemNumber() != fields.elemNumber()) {
	  getContext().reportTrouble(new CWarning(source.getTokenReference(), "x-number-colomuns-values-not-equal"));
	}
      }
      if (fields != null) {
	// check if the columns are unique
	if (fields.checkIdentsAreUnique(getContext())) {
	  // !!! ne pas oublier de verifier que le nombre d`elements dans le
	  // FieldNameList est egal au nombre de colonnes de la table
	} else {
	  // System.err.println("The columns aren't unique");
	}
      } else {
	// verifier que (ValueListInsertSource)source).elemNumber() == nombre
	// de colonnes dans la table ident
      }
    }
  }

  /*
   * Visits IntegerLiteral
   */
  public void visitIntegerLiteral(Integer value)
    throws PositionedError, UnpositionedError
  {
    current.append(value);
  }

  /*
   * Visits IntegerType
   */
  public void visitIntegerType()
    throws PositionedError, UnpositionedError
  {
    current.append("INTEGER");
  }

  /*
   * Visits IntersectSpec
   */
  public void visitIntersectSpec(CorrespondingSpec spec)
    throws PositionedError, UnpositionedError
  {
    current.append("INTERSECT");
    if (spec != null) {
      current.append(" ");
      spec.accept(this);
    }
  }

  /*
   * Visits IntersectTableReference
   */
  public void visitIntersectTableReference(IntersectSpec spec,
					   TableReference left,
					   TableReference right)
    throws PositionedError, UnpositionedError
  {
    left.accept(this);
    current.append(" ");
    spec.accept(this);
    current.append(" ");
    right.accept(this);
  }

  /*
   * Visits IntersectTableTerm
   */
  public void visitIntersectTableTerm(IntersectSpec spec,
				      SubTable left,
				      SubTable right)
    throws PositionedError, UnpositionedError
  {
    left.accept(this);
    current.append(" ");
    spec.accept(this);
    current.append(" ");
    right.accept(this);
  }

  /*
   * Visits IsPredicate
   */
  public void visitIsPredicate(Predicate predicate, boolean hasNot)
    throws PositionedError, UnpositionedError
  {
    predicate.accept(this);
    current.append(" ");
    if (hasNot) {
      current.append("IS NOT NULL");
    } else {
      current.append("IS NULL");
    }
  }

  /*
   * Visits JdbcDateLiteral
   */
  public void visitJdbcDateLiteral(Date value)
    throws PositionedError, UnpositionedError
  {
    current.append("{d ");
    current.append(value);
    current.append("}");
  }

  /*
   * Visits JdbcEscape
   */
  public void visitJdbcEscape(String name)
    throws PositionedError, UnpositionedError
  {
    current.append("{escape ");
    current.append(name);
    current.append("}");
  }

  /*
   * Visits JdbcFunction
   */
  public void visitJdbcFunction(String name, ExpressionList params)
    throws PositionedError, UnpositionedError
  {
    current.append("{fn ");
    current.append(name);
    current.append("(");
    if (params != null) {
      params.accept(this);
    }
    current.append(")}");
  }

  /*
   * Visits JdbcTimeLiteral
   */
  public void visitJdbcTimeLiteral(Time value)
    throws PositionedError, UnpositionedError
  {
    current.append("{t ");
    current.append(value);
    current.append("}");
  }

  /*
   * Visits JdbcTimeStampLiteral
   */
  public void visitJdbcTimeStampLiteral(Timestamp value)
    throws PositionedError, UnpositionedError
  {
    current.append("{ts ");
    current.append(value);
    current.append("}");
  }

  /*
   * Visits LikePredicate
   */
  public void visitLikePredicate(String sens, Expression expr, JdbcEscape escape)
    throws PositionedError, UnpositionedError
  {
    current.append("LIKE");
    if (sens != null) {
      current.append(" ");
      current.append(sens);
    }
    current.append(" ");
    expr.accept(this);
    if (escape != null) {
      current.append(" ");
      escape.accept(this);
    }
  }

  /*
   * Visits MatchesPredicate
   */
  public void visitMatchesPredicate(String sens,
				    Expression matchesExpr,
				    Expression escapeExpr)
    throws PositionedError, UnpositionedError
  {
    current.append(" MATCHES ");
    current.append(sens);
    current.append(" ");
    matchesExpr.accept(this);
    if (escapeExpr != null) {
      current.append(" ESCAPE ");
      escapeExpr.accept(this);
    }
  }

  /*
   * Visits NotCondition
   */
  public void visitNotCondition(SearchCondition condition)
    throws PositionedError, UnpositionedError
  {
    current.append("NOT ");
    condition.accept(this);
  }

  /*
   * Visits NullLiteral
   */
  public void visitNullLiteral()
    throws PositionedError, UnpositionedError
  {
    current.append("NULL");
  }

  /*
   * Visits NumericLiteral
   */
  public void visitNumericLiteral(java.math.BigDecimal value)
    throws PositionedError, UnpositionedError
  {
    current.append(value);
  }

  /*
   * Visits NumericType
   */
  public void visitNumericType(IntegerLiteral left, IntegerLiteral right)
    throws PositionedError, UnpositionedError
  {
    current.append("NUMERIC");
    if (left != null) {
      current.append("(");
      left.accept(this);
      if (right != null) {
	current.append(", ");
	right.accept(this);
      }
      current.append(")");
    }
  }

  /*
   * Visits OnJoinPred
   */
  public void visitOnJoinPred(SearchCondition condition)
    throws PositionedError, UnpositionedError
  {
    current.append("ON ");
    condition.accept(this);
  }

  /*
   * Visits OrCondition
   */
  public void visitOrCondition(SearchCondition left, SearchCondition right)
    throws PositionedError, UnpositionedError
  {
    left.accept(this);
    current.append(" OR ");
    right.accept(this);
  }

  /*
   * Visits OuterJoin
   */
  public void visitOuterJoin(TableReference leftTable,
			     String type,
			     TableReference rightTable,
			     JoinPred joinPred)
    throws PositionedError, UnpositionedError
  {
    current.append(" FROM ");
    current.append("{oj ");
    leftTable.accept(this);
    current.append(" ");
    current.append(type);
    current.append(" OUTER JOIN ");
    rightTable.accept(this);
    joinPred.accept(this);
    current.append("}");
  }

  /*
   * Visits ParenthesedTableReference
   */
  public void visitParenthesedTableReference(TableReference expr)
    throws PositionedError, UnpositionedError
  {
    expr.accept(this);
  }

  /*
   * Visits QuantifiedPredicate
   */
  public void visitQuantifiedPredicate(String type, SubTableExpression table)
    throws PositionedError, UnpositionedError
  {
    current.append(type);
    table.accept(this);
  }

  /*
   * Visits RealType
   */
  public void visitRealType()
    throws PositionedError, UnpositionedError
  {
    current.append("REAL");
  }

  /*
   * Visits SearchedCaseExpression
   */
  public void visitSearchedCaseExpression(Vector conditions, Expression elseExpr)
    throws PositionedError, UnpositionedError
  {
    for (int i = 0; i < conditions.size(); i++) {
      ((SearchedWhenClause)conditions.elementAt(i)).accept(this);
      current.append(" ");
    }
    if (elseExpr != null) {
      current.append("ELSE ");
      elseExpr.accept(this);
      //      println();
    }
  }

  /*
   * Visits SearchedWhenClause
   */
  public void visitSearchedWhenClause(SearchCondition condition, Expression thenExpr)
    throws PositionedError, UnpositionedError
  {
    current.append("WHEN ");
    condition.accept(this);
    current.append(" THEN ");
    thenExpr.accept(this);
  }

  /*
   * Visits SelectElemExpression
   */
  public void visitSelectElemExpression(Expression columnExpr, String newColumnName)
    throws PositionedError, UnpositionedError
  {
    columnExpr.accept(this);
    if (newColumnName != null) {
      current.append(" AS ");
      current.append(newColumnName);
    }
  }

  /*
   * Visits SelectElemStar
   */
  public void visitSelectElemStar(String ident)
    throws PositionedError, UnpositionedError
  {
    if (ident != null) {
      current.append(ident);
      current.append(".*");
    } else  {
      current.append("*");
    }
  }

  /*
   * Visits SelectExpression
   */
  public void visitSelectExpression(String type,
				    Vector columns,
				    FromClause fromClause,
				    WhereClause whereClause,
				    GroupByClause groupByClause,
				    HavingClause havingClause)
    throws PositionedError, UnpositionedError
  {
    current.append("SELECT ");
    if (type != null) {
      current.append(type);
      current.append(" ");
    }
    for (int i = 0; i < columns.size(); i++) {
      if (i != 0) {
	current.append(", ");
      }
      ((SelectElem)columns.elementAt(i)).accept(this);
    }

    current.append(" ");
    fromClause.accept(this);

    if (whereClause != null) {
      current.append(" ");
      whereClause.accept(this);
    }
    if (groupByClause != null) {
      current.append(" ");
      groupByClause.accept(this);
    }
    if (havingClause !=null) {
      current.append(" ");
      havingClause.accept(this);
    }

  }

  /*
   * Visits SelectStatement
   */
  public void visitSelectStatement(String type,
				   Vector columns,
				   FromClause fromClause,
				   WhereClause whereClause,
				   GroupByClause groupByClause,
				   HavingClause havingClause,
				   SortSpec sortSpec,
				   UpdSpec updateSpec)
    throws PositionedError, UnpositionedError
  {
    current.append("SELECT ");
    if (type != null) {
      current.append(type);
      current.append(" ");
    }
    for (int i = 0; i < columns.size(); i++) {
      if (i != 0) {
	current.append(", ");
      }
      ((SelectElem)columns.elementAt(i)).accept(this);
    }

    current.append(" ");
    fromClause.accept(this);

    if (whereClause != null) {
      current.append(" ");
      whereClause.accept(this);
    }
    if (groupByClause != null) {
      current.append(" ");
      groupByClause.accept(this);
    }
    if (havingClause !=null) {
      current.append(" ");
      havingClause.accept(this);
    }
    if (sortSpec != null) {
      current.append(" ");
      sortSpec.accept(this);
    }
    if (updateSpec != null) {
      current.append(" ");
      updateSpec.accept(this);
    }


    // !!!if (!fromClause.checkAliasAreUnique(this)) {
    //}

    if (sortSpec != null) {
      // check if the idents are unique
      // !!!if (sortSpec.checkIdentsAreUnique(this)) {
	if (columns.size() < sortSpec.getNumberOfIdents()) {
	  getContext().reportTrouble(new CWarning(sortSpec.getTokenReference(), "x-number-column-not-correct"));
	}
	//} else {
	//}
    }

  }

  /*
   * Visits SelectTableReference
   */
  public void visitSelectTableReference(SelectExpression expr)
    throws PositionedError, UnpositionedError
  {
    expr.accept(this);
  }

  /*
   * Visits SetFunction
   */
  public void visitSetFunction(String oper, int type, Expression expr)
    throws PositionedError, UnpositionedError
  {
    current.append(oper);
    current.append("(");
    switch(type) {
    case SetFunction.TYP_NONE:
      break;
    case SetFunction.TYP_ALL:
      current.append("ALL ");
      break;
    case SetFunction.TYP_DISTINCT:
      current.append("DISTINCT ");
      break;
    default:
      throw new RuntimeException("Unexpected type");
    }
    expr.accept(this);
    current.append(")");
  }

  /*
   * Visits SimpleCaseExpression
   */
  public void visitSimpleCaseExpression(Expression expr, Vector conditions, Expression elseExpr)
    throws PositionedError, UnpositionedError
  {
    current.append("CASE");
    expr.accept(this);
    current.append(" ");
    for (int i = 0; i < conditions.size(); i++) {
      ((SimpleWhenClause)conditions.elementAt(i)).accept(this);
      current.append(" ");
    }
    if (elseExpr != null) {
      current.append("ELSE ");
      elseExpr.accept(this);
      current.append(" ");
    }
    current.append("END");
    //   println();
  }

  /*
   * Visits SimpleIdentExpression
   */
  public void visitSimpleIdentExpression(String ident)
    throws PositionedError, UnpositionedError
  {
    current.append(ident);
  }

  /*
   * Visits SimpleSearchCondition
   */
  public void visitSimpleSearchCondition(Predicate predicate)
    throws PositionedError, UnpositionedError
  {
    predicate.accept(this);
  }

  /*
   * Visits SimpleSubTableReference
   */
  public void visitSimpleSubTableReference(TableAlias alias, TableReference table)
    throws PositionedError, UnpositionedError
  {
    current.append("(");
    table.accept(this);
    current.append(") ");
    alias.accept(this);
  }

  /**
   * Visits SimpleTableReference
   * @param	name		the table name
   * @param	alias		the alias for this reference
   */
  public void visitSimpleTableReference(Expression name, TableAlias alias)
    throws PositionedError, UnpositionedError
  {
    name.accept(this);
    if (alias != null) {
      current.append(" ");
      alias.accept(this);
    }
  }

  /*
   * Visits SimpleWhenClause
   */
  public void visitSimpleWhenClause(Expression whenExpr, Expression thenExpr)
    throws PositionedError, UnpositionedError
  {
    current.append("WHEN");
     whenExpr.accept(this);
    current.append(" THEN ");
    thenExpr.accept(this);
  }

  /*
   * Visits SmallIntType
   */
  public void visitSmallIntType()
    throws PositionedError, UnpositionedError
  {
    current.append("SMALLINT");
  }

  /*
   * Visits SortElem
   */
  public void visitSortElem(Expression elem, boolean isDesc)
    throws PositionedError, UnpositionedError
  {
    elem.accept(this);
    if (isDesc) {
      current.append(" DESC");
    }
  }

  /*
   * Visits SortSpec
   */
  public void visitSortSpec(Vector elems)
    throws PositionedError, UnpositionedError
  {
    current.append("ORDER BY ");
    for (int i = 0; i < elems.size(); i++) {
      if (i != 0) {
	current.append(", ");
      }
      ((SortElem)elems.elementAt(i)).accept(this);
    }

    //    SelectStatement newGetContext() = (SelectStatement)getContext();
    //    for (int i = 0; i < elems.size(); i++) {
    //      Expression expr = ((SortElem)elems.elementAt(i)).getExpression();
    //      if (expr == null) {
    //	// check if the IntegerLiteral is not too big
    //	Integer literal = (Integer)(((SortElem)elems.elementAt(i)).getIntegerLiteral()).getValue();
    //	if (literal != null) {
    //	  if (literal.intValue() > newGetContext().getNumberOfColumns()) {
    //	   System.err.println("The number of column " + literal.intValue() + " is too big");
    //	  }
    //	}
    //      }
    //    }
  }

  /*
   * Visits StringLiteral
   */
  public void visitStringLiteral(String value)
    throws PositionedError, UnpositionedError
  {
    current.append("'");
    current.append(value);
    current.append("'");
  }

  /*
   * Visits StringType
   */
  public void visitStringType()
    throws PositionedError, UnpositionedError
  {
    current.append("STRING");
  }

  /*
   * Visits SubTableExpression
   */
  public void visitSubTableExpression(SubTable table)
    throws PositionedError, UnpositionedError
  {
    current.append("(");
    table.accept(this);
    current.append(")");
  }

  /*
   * Visits SubTableReference
   */
  public void visitSubTableReference(SearchCondition expr)
    throws PositionedError, UnpositionedError
  {
    expr.accept(this);
  }

  /*
   * Visits TableAlias
   */
  public void visitTableAlias(String name, FieldNameList list)
    throws PositionedError, UnpositionedError
  {
    if (name != null) {
      current.append(" ");
      current.append(name);
    }
    if (list != null) {
      current.append("(");
      list.accept(this);
      current.append(")");
    }
  }

  /*
   * Visits TableExpressionInsertSource
   */
  public void visitTableExpressionInsertSource(TableReference table)
    throws PositionedError, UnpositionedError
  {
    table.accept(this);
  }

  /*
   * Visits TableInsertSource
   */
  public void visitTableInsertSource(Vector table)
    throws PositionedError, UnpositionedError
  {
    current.append("TABLE (");
    for (int i = 0; i < table.size(); i++) {
      if (i != 0) {
	current.append(", ");
      }
      ((ValueList)table.elementAt(i)).accept(this);
    }
    current.append(")");
  }

  /*
   * Visits TinyIntType
   */
  public void visitTinyIntType()
    throws PositionedError, UnpositionedError
  {
    current.append("TINYINT");
  }

  /*
   * Visits TupleInPredicate
   */
  public void visitTupleInPredicate(ExpressionList exprs, boolean hasNot, SubTableExpression table)
    throws PositionedError, UnpositionedError
  {
    current.append("[");
    exprs.accept(this);
    current.append("]");
    if (hasNot) {
      current.append(" NOT ");
    }
    current.append("IN ");
    table.accept(this);
  }

  /*
   * Visits TypeName
   */
  public void visitTypeName(String ident)
    throws PositionedError, UnpositionedError
  {
    current.append(ident);
  }

  /*
   * Visits UnaryExpression
   */
  public void visitUnaryExpression(String operator, Expression operand)
    throws PositionedError, UnpositionedError
  {
    current.append(operator);
    current.append(" ");
    operand.accept(this);
  }

  /*
   * Visits UnidiffSpec
   */
  public void visitUnidiffSpec(String oper, CorrespondingSpec spec)
    throws PositionedError, UnpositionedError
  {
    current.append(oper + " ");
    if (spec != null) {
      current.append(" ");
      spec.accept(this);
    }
  }

  /*
   * Visits UnidiffTableReference
   */
  public void visitUnidiffTableReference(UnidiffSpec spec, TableReference left, TableReference right)
    throws PositionedError, UnpositionedError
  {
    left.accept(this);
    current.append(" ");
    spec.accept(this);
    current.append(" ");
    right.accept(this);
  }

  /*
   * Visits UnidiffTableTerm
   */
  public void visitUnidiffTableTerm(UnidiffSpec spec, SubTable left, SubTable right)
    throws PositionedError, UnpositionedError
  {
    left.accept(this);
    current.append(" ");
    spec.accept(this);
    current.append(" ");
    right.accept(this);
  }

  /*
   * Visits UpdSpec
   */
  public void visitUpdSpec()
    throws PositionedError, UnpositionedError
  {
    // !!! getContext() = (SelectStatement)getContext();
    current.append("FOR UPDATE");
  }

  /*
   * Visits UpdateStatement
   */
  public void visitUpdateStatement(String schemaName,
				   String tableName,
				   Vector listIdent,
				   SearchCondition searchCondition,
				   boolean hasCondition)
    throws PositionedError, UnpositionedError
  {
    current.append("UPDATE ");
    current.append(schemaName);
    if (tableName != null) {
      current.append(" ");
      current.append(tableName);
    }
    current.append(" ");
    current.append("SET ");

    for (int i = 0; i < listIdent.size(); i++) {
      if (i != 0) {
	current.append(", ");
      }
      ((Assignment)listIdent.elementAt(i)).accept(this);
    }

    if (hasCondition) {
      current.append(" WHERE ");
      if (searchCondition != null) {
	searchCondition.accept(this);
      } else {
	current.append("CURRENT");
      }
    }
    // the table exists ?
    if (!XUtils.tableExists(schemaName)) {
      throw new UnpositionedError("x-table-not-found", schemaName);
    }

  }

  /*
   * Visits UsingJoinPred
   */
  public void visitUsingJoinPred(FieldNameList list)
    throws PositionedError, UnpositionedError
  {
    current.append("USING (");
    list.accept(this);
    current.append(")");
  }

  /*
   * Visits ValueInPredicate
   */
  public void visitValueInPredicate(ExpressionList value)
    throws PositionedError, UnpositionedError
  {
    current.append("IN (");
    value.accept(this);
    current.append(")");
  }

  /*
   * Visits ValueList
   */
  public void visitValueList(ExpressionList expr)
    throws PositionedError, UnpositionedError
  {
    current.append("(");
    expr.accept(this);
    current.append(")");
  }

  /*
   * Visits ValueListInsertSource
   */
  public void visitValueListInsertSource(ValueList value)
    throws PositionedError, UnpositionedError
  {
    current.append("VALUES ");
    value.accept(this);
  }

  /*
   * Visits WhereClause
   */
  public void visitWhereClause(SearchCondition condition)
    throws PositionedError, UnpositionedError
  {
    current.append("WHERE ");
    condition.accept(this);
  }

  // ----------------------------------------------------------------------
  // PROTECTED UTILITIES
  // ----------------------------------------------------------------------

  public SqlContext getContext() {
    return contexts.size() > 0 ? (SqlContext)contexts.peek() : null;
  }

  public void enter(SqlContext context) {
    contexts.push(context);
  }

  public void exit(SqlContext context) {
    if (contexts.pop() != context) {
      throw new RuntimeException();
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  protected StringBuffer			current;
  protected Stack				contexts = new Stack();
}
