/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Main.java,v 1.59 1999/10/16 23:25:36 vgp Exp $
 */

package at.dms.xkjc;

import java.io.File;
import java.io.IOException;
import gnu.getopt.LongOpt;
import gnu.getopt.Getopt;

import at.dms.kjc.JCompilationUnit;
import at.dms.kjc.CJavaPrettyPrinter;
import at.dms.kjc.CHandLexer;
import at.dms.util.Message;

/**
 * This class implements the entry point of the Java compiler
 */
public class Main extends at.dms.kjc.Main {

  // ----------------------------------------------------------------------
  // CONSTRUCTOR
  // ----------------------------------------------------------------------

  /**
   * Entry point
   * @param	args		the command line arguments
   */
  public static void main(String[] args) {
    if (instance != null) {
      // only one instance of Main at the same time
      throw new RuntimeException("DO NOT CALL MAIN WHILE THERE IS A COMPILE PROCESS");
    }

    boolean	success;

    instance = new Main();
    success = instance.run(args);
    instance = null;
    err.close();
    System.exit(success ? 0 : 1);
  }

  /**
   * Second entry point
   */
  public static boolean compile(String[] args) {
    if (instance != null) {
      // only one instance of Main at the same time
      throw new RuntimeException("DO NOT CALL MAIN WHILE THERE IS A COMPILE PROCESS");
    }

    boolean	success;

    instance = new Main();
    success = instance.run(args);
    instance = null;
    err.flush();
    return success;
  }

  /**
   * Initialize the compiler (read classpath, check classes.zip)
   */
  protected void initialize() {
    super.initialize();
    XStdType.init(this);
    XUtils.initialize(xkjcpath, !optionNoOperatorOverloading);
    Message.addBundle("xkjc");
    Message.addBundle("sqlc");
  }

  /**
   * shows an usage message
   * @param	msg		the error that stop the command line parser
   */
  protected void usage(String msg) {
    err.println("xkjc: " + msg);
    err.println(Message.format("usage-line", null));
  }

  /**
   * shows a help message
   */
  protected void help() {
    err.println(Message.format("usage-xkjc-path", null));
    err.println(Message.format("usage-nooo", null));
    err.println(Message.format("usage-database", null));
    super.help();
  }

  // ----------------------------------------------------------------------
  // PROTECTED METHODS
  // ----------------------------------------------------------------------

  /*
   * Parse command line arguments.
   */
  protected boolean parseArgs(String[] args) {
    Getopt	g = new Getopt("xkjc", args, SHORTOPTS, LONGOPTS, true);
  _loop_:
    for (;;) {
      switch (g.getopt()) {
      case 'b':
	optionBeautify = true;
	break;

      case 'd':
	destination = g.getOptarg();
	break;

      case 'g':
	//optionDebug = true;
	break;

      case 'h':
	usage("");
	help();
	err.flush();
	System.exit(0);

      case 'j':
	optionGenerateJava = true;
	break;

      case 'm':
	optionMultiThread = true;
	break;

      case 'n':
	optionNoWrite = true;
	break;

      case 'O':
	optionOptimize = g.getOptarg() != null ?
	  new Integer(g.getOptarg()).intValue() * 10 :
	    1;
	break;

      case 'V':
	err.println("KOPI Java Compiler Version: " + VERSION);
	err.flush();
	System.exit(0);
	break;

      case 'v':
	optionVerbose = true;
	break;

      case 'w':
	optionWarning = true;
	break;

      case 500:
	classpath = g.getOptarg();
	break;

      case 501:
	optionDeprecation = true;
	break;

	// XKJC OPTIONS
      case 600:
	XUtils.setDatabase(g.getOptarg());
	break;

      case 601:
	optionNoOperatorOverloading = true;
	break;

      case 602:
	xkjcpath = g.getOptarg();
	break;

      case -1:
	break _loop_;

      default:
	return false;
      }
    }

    for (int i = g.getOptind(); i < args.length; i++) {
      infiles.addElement(args[i]);
    }

    return true;
  }

  /**
   * parse the given file and return a compilation unit
   * side effect: increment error number
   * @param	file		the name of the file (assert exists)
   * @return	the compilation unit defined by this file
   */
  protected JCompilationUnit parseFile(File file) {
    CHandLexer		lexer = null; // to get line number if errors
    long		lastTime = System.currentTimeMillis();
    XParseManager	parsers = null;
    try {
      parsers = XParseManager.getXParseManager(this, file, file.getName());
      lexer = parsers.getXJavaLexer();
      JCompilationUnit unit = parsers.getXJavaParser().jCompilationUnit();
      if (verboseMode()) {
	console.parseFile(lastTime, file.getName());
      }
      parsers.close();
      return unit;
    } catch (antlr.ParserException e) {
      err.println(file + ":" + beautifyParseError(e, lexer.getLine()));
      errorFound = true;
      parsers.close();
    } catch (IOException e) {
      e.printStackTrace(); // !!!
      err.println("{" + file.getName() + ":" + lexer.getLine() + "} " + e.getMessage());
      errorFound = true;
    }
    if (verboseMode()) {
      console.parseFile(lastTime, file.getName()) ;
    }
    return null;
  }

  /**
   * Shows an error/warning and set error flag to true
   */
  protected CJavaPrettyPrinter getPrettyPrinterOverided(String fileName) {
    return new XJavaPrettyPrinter(fileName);
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private static final String	VERSION			= "1.3D";

  public static final String	SHORTOPTS = "bd:ghjmnO:Vvw";
  public static final
    LongOpt[]			LONGOPTS = {
      at.dms.kjc.Main.LONGOPTS[0],
      at.dms.kjc.Main.LONGOPTS[1],
      at.dms.kjc.Main.LONGOPTS[2],
      at.dms.kjc.Main.LONGOPTS[3],
      at.dms.kjc.Main.LONGOPTS[4],
      at.dms.kjc.Main.LONGOPTS[5],
      at.dms.kjc.Main.LONGOPTS[6],
      at.dms.kjc.Main.LONGOPTS[7],
      at.dms.kjc.Main.LONGOPTS[8],
      at.dms.kjc.Main.LONGOPTS[9],
      at.dms.kjc.Main.LONGOPTS[10],
      at.dms.kjc.Main.LONGOPTS[11],
      at.dms.kjc.Main.LONGOPTS[12],
      at.dms.kjc.Main.LONGOPTS[13],
      new LongOpt("database",		LongOpt.REQUIRED_ARGUMENT,	null, 600),
      new LongOpt("nooo",		LongOpt.NO_ARGUMENT,		null, 601),
      new LongOpt("xkjcpath",		LongOpt.REQUIRED_ARGUMENT,	null, 602)
    };

  private String		xkjcpath;
  private boolean		optionNoOperatorOverloading;
}
