/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: XCompoundAssignmentExpression.java,v 1.13 1999/10/08 11:19:12 vgp Exp $
 */

package at.dms.xkjc;

import at.dms.kjc.*;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * This class implements the assignment operation
 */
public class XCompoundAssignmentExpression extends JCompoundAssignmentExpression {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * This method is directly called by the parser
   * @param	where		the line of this node in the source code
   * @param	oper		the assignment operator
   * @param	left		the left operand
   * @param	right		the right operand
   */
  public XCompoundAssignmentExpression(TokenReference where,
				       int oper,
				       JExpression left,
				       JExpression right) {
    super(where, oper, left, right);
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * 15.25.2 Compound Assignment Operators
   * // x += 3;   ===>   x = (x.getType()) (x + 3)
   * @param	context		the actual context of analyse
   * @return	a pure java expression including promote node
   * @exception	PositionedError		if the check fails
   */
  public JExpression checkExpression(CExpressionContext context) throws PositionedError {
    context.setLeftSide(true);
    try {
      left = left.checkExpression(context);
    } finally {
      context.setLeftSide(false);
    }

    check(context, left instanceof CInitializable, "assignment-notlvalue");
    try {
      ((CInitializable)left).initialize(context.getBodyContext());
    } catch (UnpositionedError e) {
      throw e.getError(getTokenReference());
    }

    right = right.checkExpression(context);

    boolean convertRight = true;

    try {
      switch (oper) {
      case OPE_STAR:
	type = JMultExpression.computeType(left.getType(), right.getType());
	break;
      case OPE_SLASH:
	type = JDivideExpression.computeType(left.getType(), right.getType());
	break;
      case OPE_PERCENT:
	type = JModuloExpression.computeType(left.getType(), right.getType());
	break;
      case OPE_PLUS:
	type = JAddExpression.computeType(left.getType(), right.getType());
	break;
      case OPE_MINUS:
	type = JMinusExpression.computeType(left.getType(), right.getType());
	break;
      case OPE_SL:
      case OPE_SR:
      case OPE_BSR:
	type = JShiftExpression.computeType(left.getType(), right.getType());
	convertRight = false;
	break;
      case OPE_BAND:
      case OPE_BXOR:
      case OPE_BOR:
	type = JBitwiseExpression.computeType(left.getType(), right.getType());
	break;
      default:
	throw new RuntimeException("UNDEFINED OPERATOR:" + oper);
      }
      check(context, type.isCastableTo(left.getType()), "assignment-badtype");

      type = left.getType();

      if (convertRight) {
	right = right.convertType(type, context).checkExpression(context);
      }

      return this;
    } catch (UnpositionedError ce) {
      try {
	return tryExpanded(context).checkExpression(context);
      } catch (PositionedError e) {
        throw ce.getError(getTokenReference());
      }
    } catch (PositionedError ce) {
      try {
	return tryExpanded(context).checkExpression(context);
      } catch (PositionedError e) {
        throw ce;
      }
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE METHODS
  // ----------------------------------------------------------------------

  private JExpression tryExpanded(CExpressionContext context) throws PositionedError {
    // better check: hasSideEffect() !!!
    check(context,
	  left instanceof JLocalVariableExpression ||
	  left instanceof JClassFieldExpression ||
	  left instanceof JArrayAccessExpression, "XXX");//!!! FALSE

    TokenReference	ref = getTokenReference();
    JExpression		expr;

    switch (oper) {
    case OPE_STAR:
      expr = new XMultExpression(ref,
				 new JCheckedExpression(ref, left),
				 new JCheckedExpression(ref, right));
      break;
    case OPE_SLASH:
      expr = new XDivideExpression(ref,
				   new JCheckedExpression(ref, left),
				   new JCheckedExpression(ref, right));
      break;
    case OPE_PERCENT:
      expr = new XModuloExpression(ref,
				   new JCheckedExpression(ref, left),
				   new JCheckedExpression(ref, right));
      break;
    case OPE_PLUS:
      expr = new XAddExpression(ref,
				new JCheckedExpression(ref, left),
				new JCheckedExpression(ref, right));
      break;
    case OPE_MINUS:
      expr = new XMinusExpression(ref,
				  new JCheckedExpression(ref, left),
				  new JCheckedExpression(ref, right));
      break;
    case OPE_SL:
    case OPE_SR:
    case OPE_BSR:
      expr = new XShiftExpression(ref,
				  oper,
				  new JCheckedExpression(ref, left),
				  new JCheckedExpression(ref, right));
      break;
    case OPE_BAND:
    case OPE_BXOR:
    case OPE_BOR:
      expr = new XBitwiseExpression(ref,
				    oper,
				    new JCheckedExpression(ref, left),
				    new JCheckedExpression(ref, right));
      break;
    default:
      throw new RuntimeException("UNDEFINED OPERATOR:" + oper);
    }

    return new XAssignmentExpression(ref, left, expr);
  }
}
