/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: XExecSqlStatement.java,v 1.28 1999/10/28 14:03:13 vgp Exp $
 */

package at.dms.xkjc;

import at.dms.kjc.*;
import at.dms.sqlc.*;
import at.dms.compiler.CWarning;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;
import at.dms.compiler.JavaStyleComment;

/**
 * This class represents cursor declaration in code part
 */
public class XExecSqlStatement extends XStatement {

  // ----------------------------------------------------------------------
  // CONSTRUCTORS
  // ----------------------------------------------------------------------

  /**
   * Construct a node in the parsing tree
   * This method is directly called by the parser
   * @param	where		the line of this node in the source code
   * @param	conn		a pointer to a valid connection to a database
   * @param	body		the SQL expression to be executed
   * @param	types		a vector of type of columns in the database
   * @param	into		a list of lvalue expression to be modifier by this statement
   * @param	comment		the statement comment
   */
  public XExecSqlStatement(TokenReference where,
			   JExpression conn,
			   XSqlExpr body,
			   CType[] types,
			   JExpressionListStatement into,
			   JavaStyleComment[] comments) {
    super(where, comments);
    this.conn = conn;
    this.body = body;
    this.into = into;
    this.types = types;
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check statement and return a pure kjc abstract tree that will be used to code generation.
   *
   * @param	context		the actual context of analyse
   * @exception	PositionedError		if the check fails
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public JStatement checkAndConvert(CBodyContext context) throws PositionedError {
    for (int i = 0; i < types.length; i++) {
      try {
	this.types[i].checkType(context);
      } catch (UnpositionedError e) {
	throw e.getError(getTokenReference());
      }
    }

    if (conn == null) {
      check(context, context.getClassContext().getCClass().descendsFrom(XStdType.DBContextHandler.getCClass()),
	    "x-badDefaultContext", context.getClassContext().getCClass().getQualifiedName());
    } else {
      conn = conn.checkExpression(new CExpressionContext(context));
      if (!conn.getType().isAssignableTo(XStdType.Connection)) {
	check(context, conn.getType().isAssignableTo(XStdType.DBContextHandler),
	      "x-badConnection", conn.getType().getCClass().getQualifiedName());
	// CONSTRUCT THE CONNECTION ACCESS
	conn = XProtectedStatement.buildConnection(getTokenReference(), conn).checkExpression(new CExpressionContext(context));
      }
    }

    JExpression[] intos = into.getExpressions();
    JExpressionStatement[] intoStatements = new JExpressionStatement[intos.length];

    check(context, intos.length == types.length, "x-selectListLength");

    TokenReference ref = getTokenReference();

    if (conn == null) {
      conn = XFactory.parseConnection(ref, null, new JExpression[] {new JThisExpression(ref)});
    }

    JExpression prefix = new JNameExpression(ref, "qry$");
    for (int i = 0; i < intos.length; i++) {
      JExpression acs = XCursorField.createAccessExpression(ref, types[i], i, prefix, new CExpressionContext(context));
      intoStatements[i] = new JExpressionStatement(ref,
						   new XAssignmentExpression(ref,
									     intos[i],
									     acs), null);
    }

    JStatement stmt =  XFactory.parseSelectInto(ref,
						new JStatement[] {new JCompoundStatement(ref, intoStatements)},
						new JExpression[] {conn, body});
    stmt.checkStatement(context);

    checkSql(context);

    return stmt;
  }

  protected void checkSql(CBodyContext context) {
    SelectStatement	select = (SelectStatement)body.getSql();
    for (int i = 0; i < types.length; i++) {
      if (types[i].isPrimitive()) {
	SelectElem	elem = (SelectElem)((XTypedSqlExpr)select.getSelectElemAt(i)).getSql();

	if (elem instanceof SelectElemExpression) {
	  Expression	expr = ((SelectElemExpression)elem).getExpression();

	  if (expr instanceof SetFunction) {
	    context.reportTrouble(new CWarning(expr.getTokenReference(), "x-execute-badprimitive"));
	  }
	}
      }
    }
  }



  // ----------------------------------------------------------------------
  // CODE GENERATION
  // ----------------------------------------------------------------------

  /**
   * Generate the code in pure java form
   * It is usefull to debug and tune compilation process
   * @param	p		the printwriter into the code is generated
   */
  public void genXKJCCode(XJavaPrettyPrinter p) {
    p.printExecSqlStatement(conn, body, into);
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  protected JExpression		conn;
  protected XSqlExpr		body;
  protected JExpressionListStatement	into;
  protected CType[]		types;
}
