/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: XOverloadedMethodCallExpression.java,v 1.19 1999/10/08 11:19:12 vgp Exp $
 */

package at.dms.xkjc;

import at.dms.kjc.*;
import at.dms.compiler.PositionedError;
import at.dms.compiler.TokenReference;
import at.dms.compiler.UnpositionedError;

/**
 * This class represents methods call
 * <pre> methodname(explist) </pre>
 * the only difference with parent is that method can be checked a lot of time efficiently
 */
public class XOverloadedMethodCallExpression extends JMethodCallExpression {

  /**
   * Construct a node in the parsing tree
   * This method is directly called by the parser
   * @param	where		the line of this node in the source code
   * @param	prefix		an expression that is a field of a class representing a method
   * @param	ident		the method ident
   * @param	args		the argument of the call
   */
  public XOverloadedMethodCallExpression(TokenReference where,
					 JExpression prefix,
					 JExpression[] args) {
    super(where, prefix, args);
  }

  // ----------------------------------------------------------------------
  // CODE CHECKING
  // ----------------------------------------------------------------------

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError	an error with reference to the source file
   * @exception	UnpositionedError	this error will be positioned soon
   */
  public void preCheckExpression(CExpressionContext context) throws PositionedError {
    argsType = new CType[args.length];
    for (int i = 0; i < args.length; i++) {
      args[i] = args[i].checkExpression(context);
      argsType[i] = args[i].getType();
    }

    assert(ident != null);

    if (prefix != null) {
      prefix = prefix.checkExpression(context);
      check(context, prefix.getType().isReference(),
	    "method-badprefix", ident, prefix.getType());
    }
  }

  /**
   * Set a checked prefix for this expression (after preCheck)
   */
  public void setPrefix(JExpression prefix) {
    this.prefix = prefix;
  }

  /**
   * Check expression and evaluate and alter context
   * @param context the actual context of analyse
   * @return  a pure java expression including promote node
   * @exception	PositionedError	an error with reference to the source file
   */
  public JExpression checkExpression(CExpressionContext context) throws PositionedError {
    if (prefix != null) {
      try {
	method = prefix.getType().getCClass().lookupMethod(ident, argsType);
      } catch (UnpositionedError e) {
	throw e.getError(getTokenReference());
      }
    } else {
      try {
	method = context.lookupMethod(ident, argsType);
      } catch (UnpositionedError e) {
	throw e.getError(getTokenReference());
      }
    }

    check(context, method != null, "method-notfound", ident);

    // check access
    CClass		local = context.getClassContext().getCClass();

    check(context, method.isAccessible(local), "method-noaccess", method.getIdent());

    CClassType[]	exceptions = method.getThrowables();
    for (int i = 0; i < exceptions.length; i++) {
      if (exceptions[i].isCheckedException()) {
	if (prefix == null ||				// special case of clone
	    !prefix.getType().isArrayType() ||
	    !ident.equals("clone") ||
	    !exceptions[i].getCClass().getQualifiedName().
	    equals("java/lang/CloneNotSupportedException")) {

	  context.getBodyContext().addThrowable(new CThrowableInfo(exceptions[i],
								   this,
								   context.getBodyContext().makeSnapshot()));
	}
      }
    }

    if ((prefix == null) && !method.isStatic()) {
      prefix = new JThisExpression(getTokenReference()).checkExpression(context);
    }

    argsType = method.getParameters();
    for (int i = 0; i < args.length; i++) {
      args[i] = args[i].convertType(argsType[i], context);
    }

    return this;
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private	CType[]	argsType;
}
