/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Cursor.java,v 1.14 1999/08/25 18:04:34 graf Exp $
 */

package at.dms.xlib;

import at.dms.xtype.Date;
import at.dms.xtype.Fixed;
import at.dms.xtype.Month;
import at.dms.xtype.Time;
import at.dms.xtype.NotNullDate;
import at.dms.xtype.NotNullFixed;
import at.dms.xtype.NotNullMonth;
import at.dms.xtype.NotNullTime;

import java.sql.ResultSet;
import java.sql.Statement;
import java.sql.SQLException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.OptionalDataException;
import java.io.IOException;

/**
 * Execute a query and maintain a pointer to the retreived data
 * this data can be accessed from java with the good primitive type.
 * Thanks to kopic overloaded operator it is completly transparent.
 * Warning: Trying to access to a null value as a primitve type will
 * throw a null pointer exception
 *
 * Rmq 1: All exception are cached in order to find from the vendor specific
 * tag which error it is and if possible we rethrow the good exception
 * (DBDeadLock, ...) else we rethrow a general DBException
 *
 * Rmq 2: This class is a runtime utility for the kopi files, it is not intented
 * to use them directly in Java. You may use them from Java but there is
 * not any documentation or help to do this.
 */
public class Cursor {

  /**
   * Open the cursor with a sql statement (SELECT)
   */
  public void open(String sql, Connection conn)
    throws DBException
  {
    if (Query.traceQueries()) {
      System.out.println("CURSOR OPEN: " + sql);
    }

    try {
      stmt = conn.getJDBCConnection().createStatement();
      if (supportsCursorNames()) {
	stmt.setCursorName("K" + nextCursorId++);
      }

      stmt.executeQuery(sql);
      rs = stmt.getResultSet();
    } catch (SQLException exc) {
      throw conn.convert(exc);
    }
  }

  /**
   * Open the cursor with a sql statement (SELECT) but use default connection
   */
  public void openWOCursor(String sql) throws DBException
  {
    this.open(sql, this.conn);
  }

  // ----------------------------------------------------------------------
  // CONNECTION AND FETCH
  // ----------------------------------------------------------------------

  /**
   *
   */
  public void setDefaultConnection(Connection conn) {
    this.conn = conn;
  }

  /**
   * Move to next row of result set
   */
  public boolean next() throws DBException {
    try {
      return rs == null ? false : rs.next();
    } catch (SQLException exc) {
      throw conn.convert(exc);
    }
  }

  /**
   * Update a row (DELETE, INSERT, UPDATE)
   */
  public void executeUpdate(String sql) throws DBException {
    if (Query.traceQueries()) {
      System.out.println("CURSOR UPDATE: " + sql);
    }
    System.out.println("CURSOR UPDATE: " + sql);

    try {
      if (stmt != null) {
	stmt.executeUpdate(sql);
      }
    } catch (SQLException exc) {
      throw conn.convert(exc);
    }
  }

  /**
   * Close this cursor
   */
  public void close() throws SQLException {
    try {
      if (stmt != null) {
	stmt.close();
      }
    } catch (SQLException exc) {
      throw conn.convert(exc);
    }
  }

  /**
   * Update positioned
   */
  public int update(String sql, Object[] blobs) throws DBException {
    if (blobs != null && blobs.length != 0) {
      throw new RuntimeException("NOT YET IMPLEMENTED"); // !!!
    }
    return update(sql);
  }

  /**
   * Update positioned
   */
  public int update(String sql) throws DBException {
    try {
      if (! supportsCursorNames()) {
	throw new SQLException("operation not supported by JDBC driver");
      }

      Statement		updater;
      int		count;

      updater = conn.getJDBCConnection().createStatement();
      count = updater.executeUpdate(sql + " WHERE CURRENT OF " + rs.getCursorName());
      updater.close();

      return count;
    } catch (SQLException exc) {
      throw conn.convert(exc);
    }
  }

  /**
   * Delete positioned
   */
  public int delete(String sql) throws DBException {
    return update(sql);
  }

  // ----------------------------------------------------------------------
  // ACCESSORS (primitive type)
  // ----------------------------------------------------------------------

  public boolean getBoolean(int pos) throws SQLException {
    boolean tmp = rs.getBoolean(pos);
    if (rs.wasNull()) {
      throw new SQLException("null pointer exception" + pos);
    }
    return tmp;
  }

  public byte getByte(int pos) throws SQLException {
    byte tmp = rs.getByte(pos);
    if (rs.wasNull()) {
      throw new SQLException("null pointer exception" + pos);
    }
    return tmp;
  }

  public short getShort(int pos) throws SQLException {
    short tmp = rs.getShort(pos);
    if (rs.wasNull()) {
      throw new SQLException("null pointer exception" + pos);
    }
    return tmp;
  }

  public int getInt(int pos) throws SQLException {
    int tmp = rs.getInt(pos);
    if (rs.wasNull()) {
      throw new SQLException("null pointer exception" + pos);
    }
    return tmp;
  }

  public NotNullFixed getFixed(int pos) throws SQLException {
    java.math.BigDecimal tmp = rs.getBigDecimal(pos, 4);
    if (rs.wasNull()) {
      throw new SQLException("null pointer exception" + pos);
    }
    return new NotNullFixed(tmp);
  }

  public float getFloat(int pos) throws SQLException {
    float tmp = rs.getFloat(pos);
    if (rs.wasNull()) {
      throw new SQLException("null pointer exception" + pos);
    }
    return tmp;
  }

  public double getDouble(int pos) throws SQLException {
    double tmp = rs.getDouble(pos);
    if (rs.wasNull()) {
      throw new SQLException("null pointer exception" + pos);
    }
    return tmp;
  }

  public char getChar(int pos) throws SQLException {
    String tmp = rs.getString(pos);
    if (rs.wasNull()) {
      throw new SQLException("null pointer exception" + pos);
    }
    return tmp.charAt(0);
  }

  public String getString(int pos) throws SQLException {
    return rs.getString(pos);
  }

  public NotNullMonth getMonth(int pos) throws SQLException {
    int tmp = rs.getInt(pos);
    if (rs.wasNull()) {
      throw new SQLException("null pointer exception" + pos);
    }
    return new NotNullMonth(tmp);
  }

  public NotNullTime getTime(int pos) throws SQLException {
    java.sql.Time tmp = rs.getTime(pos);
    if (rs.wasNull()) {
      throw new SQLException("null pointer exception" + pos);
    }
    return new NotNullTime(tmp);
  }

  public NotNullDate getDate(int pos) throws SQLException {
    java.sql.Date tmp = rs.getDate(pos);
    if (rs.wasNull()) {
      throw new SQLException("null pointer exception" + pos);
    }
    return new NotNullDate(tmp);
  }

  public java.lang.Boolean getNullableBoolean(int pos) throws SQLException {
    boolean tmp = rs.getBoolean(pos);
    return rs.wasNull() ? null : new java.lang.Boolean(tmp);
  }

  public java.lang.Byte getNullableByte(int pos) throws SQLException {
    byte tmp = rs.getByte(pos);
    return rs.wasNull() ? null : new java.lang.Byte(tmp);
  }

  public java.lang.Short getNullableShort(int pos) throws SQLException {
    short tmp = rs.getShort(pos);
    return rs.wasNull() ? null : new java.lang.Short(tmp);
  }

  public java.lang.Integer getNullableInt(int pos) throws SQLException {
    int tmp = rs.getInt(pos);
    return rs.wasNull() ? null : new java.lang.Integer(tmp);
  }

  public Fixed getNullableFixed(int pos) throws SQLException {
    java.math.BigDecimal tmp = rs.getBigDecimal(pos, 4);
    if (rs.wasNull()) {
      return null;
    }
    return new NotNullFixed(tmp);
  }

  public java.lang.Float getNullableFloat(int pos) throws SQLException {
    float tmp = rs.getFloat(pos);
    return rs.wasNull() ? null : new java.lang.Float(tmp);
  }

  public java.lang.Double getNullableDouble(int pos) throws SQLException {
    double tmp = rs.getDouble(pos);
    return rs.wasNull() ? null : new java.lang.Double(tmp);
  }

  public java.lang.Character getNullableChar(int pos) throws SQLException {
    String tmp = rs.getString(pos);
    return rs.wasNull() ? null : new java.lang.Character(tmp.charAt(0));
  }

  public String getNullableString(int pos) throws SQLException {
    return getString(pos);
  }

  public Time getNullableTime(int pos) throws SQLException {
    java.sql.Time tmp = rs.getTime(pos);
    if (rs.wasNull()) {
      return null;
    }
    return new NotNullTime(tmp);
  }

  public Date getNullableDate(int pos) throws SQLException {
    java.sql.Date tmp = rs.getDate(pos);
    if (rs.wasNull()) {
      return null;
    }
    return new NotNullDate(tmp);
  }

  public Month getNullableMonth(int pos) throws SQLException {
    int tmp = rs.getInt(pos);
    if (rs.wasNull()) {
      return null;
    } else {
      return new NotNullMonth(tmp);
    }
  }

  // ----------------------------------------------------------------------
  // ACCESSORS object
  // ----------------------------------------------------------------------

  /**
   * WARNING: You should read the entire data before accessing any other column
   */
  public InputStream getBinaryStream(int pos) throws SQLException {
    return rs.getBinaryStream(pos);
  }

  /**
   * Reads the stream in a byte array
   */
  public byte[] getByteArray(int pos) throws SQLException {
    try {
      InputStream	is = rs.getBinaryStream(pos);
      if (is == null) {
	return null;
      }
      byte[]	b = new byte[is.available()];
      is.read(b);
      return b;
    } catch (java.io.IOException e) {
      throw new RuntimeException("INPUT STREAM BROKEN:" + e.getMessage());
    }
  }

  /**
   * Returns an object stored in database
   */
  public Object getNullableSerializedObject(int pos) throws SQLException {
    return getSerializedObject(pos);
  }

  /**
   * Returns an object stored in database
   */
  public Object getSerializedObject(int pos) throws SQLException {
    InputStream is = rs.getBinaryStream(pos);
    if (is == null) {
      return null;
    }

    Object	obj = null;

    try {
      obj = new ObjectInputStream(is).readObject();
      is.close();
    } catch (OptionalDataException e1) {
      throw new DBInvalidDataException(e1);
    } catch (ClassNotFoundException e2) {
      throw new DBInvalidDataException(e2);
    } catch (IOException e3) {
      throw new DBInvalidDataException(e3);
    }

    return obj;
  }

  // ----------------------------------------------------------------------
  // Kopi Serializable
  // ----------------------------------------------------------------------

  /**
   * We want to know if the object is null before create it !
   */
  public boolean isNull(int pos) throws SQLException {
    cached = rs.getBinaryStream(pos);
    return cached == null;
  }

  /**
   * Reads the stream in a byte array
   */
  public byte[] getKopiSerializable(int pos) throws SQLException {
    try {
      if (cached == null) {
	cached = rs.getBinaryStream(pos);
      }

      if (cached == null) {
	return null;
      }
      byte[]	b = new byte[cached.available()];
      cached.read(b);
      cached = null;
      return b;
    } catch (java.io.IOException e) {
      throw new RuntimeException("INPUT STREAM BROKEN:" + e.getMessage());
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE METHODS
  // ----------------------------------------------------------------------

  /*
   * Returns true if the underlying JDBC connection supports cursor names
   */
  private boolean supportsCursorNames() {
    try {
      return conn.getJDBCConnection().getMetaData().getMaxCursorNameLength() > 0;
    } catch (SQLException e) {
      return false;
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  protected Connection		conn;
  protected Statement		stmt;
  protected ResultSet		rs;

  private   InputStream cached;
  private static int		nextCursorId = 0;
}
