/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: DBContext.java,v 1.10 1999/08/25 18:04:34 graf Exp $
 */

package at.dms.xlib;

import java.util.Vector;
import java.util.Enumeration;
import java.sql.DriverManager;
import java.sql.Driver;
import java.sql.SQLException;

public class DBContext {

  /**
   * Registers a JDBC driver.
   *
   * @param	name		The class name of the JDBC driver to register.
   */
  public static void registerDriver(String name) {
    try {
      Class.forName(name);
    } catch (java.lang.ClassNotFoundException e) {
      throw new DBRuntimeException("can't access driver " + name);
    }
  }

  /**
   * Connects to database and logs on.
   *
   * @param	url		the URL of the database to connect to
   * @param	user		the name of the database user
   * @param	pass		the password of the database user
   * @return	a new kopi connection
   */
  public Connection createConnection(String url, String user, String pass)
    throws DBException {

    Driver	driver;
    try {
      driver = DriverManager.getDriver(url);
    } catch (SQLException exc) {
      throw new DBRuntimeException("can't create connection to " + url + " {" + exc.getMessage() + "}");
    }

    try {
      java.sql.Connection	con;
      at.dms.xlib.Connection	kcon;

      con = DriverManager.getConnection(url, user, pass);
      con.setAutoCommit(false);

      kcon = new Connection(this, con, driver, url, user, pass);
      connections.addElement(kcon);

      return kcon;
    } catch (SQLException exc) {
      throw getDBExceptionConverter(driver).convert(exc);
    }
  }

  /**
   * Closes the underlying JDBC connections.
   */
  public void close() throws DBException {
    Enumeration elems = getConnections().elements();
    while (elems.hasMoreElements()) {
      Connection kcon = (Connection)elems.nextElement();

      try {
	kcon.getJDBCConnection().rollback(); // !!! what's happend if closed by user ?
	kcon.getJDBCConnection().close();
      } catch (SQLException exc) {
	throw kcon.convert(exc);
      }
    }
    connections = new Vector();
  }

  /**
   * Closes the underlying JDBC connections.
   */
  public void close(Connection kcon) throws DBException {
    java.sql.Connection con = kcon.getJDBCConnection();
    connections.removeElement(kcon);

    try {
      con.rollback();
      con.close(); // !!! What to do if transaction in progress
    } catch (SQLException exc) {
      throw kcon.convert(exc);
    }
  }

  /**
   * Returns the underlying default JDBC connection.
   */
  public Connection getDefaultConnection() {
    return defaultConnection;
  }

  /**
   * sets the underlying default JDBC connection.
   */
  public void setDefaultConnection(Connection con) {
    defaultConnection = con;
  }

  /**
   * Returns all connections currenty opened
   */
  public Vector getConnections() {
    return connections;
  }

  /**
   * Starts a new transaction.
   * Note: JDBC opens transactions implicitly
   */
  public final void startWork() {
    if (inTransaction) {
       synchronized(this) {
	 try {
	   wait();
	 } catch (InterruptedException e) {
	   throw new RuntimeException();
	 }
       }
    }
    inTransaction = true;
  }

  /**
   * Commits the current transaction.
   */
  public final void commitWork() throws SQLException {
    inTransaction = true;
    Enumeration elems = getConnections().elements();
    while (elems.hasMoreElements()) {
      Connection	kcon = (Connection)elems.nextElement();
      try {
	kcon.getJDBCConnection().commit();
      } catch (SQLException exc) {
	throw kcon.convert(exc);
      }
    }
    synchronized(this) {
      notify();
    }
    inTransaction = false;
  }

  /**
   * Aborts the current transaction.
   */
  public final void abortWork() throws SQLException {
    inTransaction = true;
    Enumeration elems = getConnections().elements();
    while (elems.hasMoreElements()) {
      Connection	kcon = (Connection)elems.nextElement();
      try {
	kcon.getJDBCConnection().rollback();
      } catch (SQLException exc) {
	//throw kcon.convert(exc); (if the error has already closed the transaction)
      }
    }
    synchronized(this) {
      notify();
    }
    inTransaction = false;
  }

  /**
   * Returns if there is a transaction started
   */
  public boolean isInTransaction() {
    return inTransaction;
  }

  /**
   *
   */
  public static DBExceptionConverter getDBExceptionConverter(Driver d) {
    if (d.getClass().getName().equals("at.dms.kconnect.Driver")) {
      return transbaseConverter;
    } else {
      return oracleConverter;
    }
  }

  /**
   *
   */
  public static DBExceptionConverter getDBExceptionConverter(String driverName) {
    if (driverName.equals("Transbase-JDBC Driver")) {
      return transbaseConverter;
    } else {
      return oracleConverter;
    }
  }

  // ----------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // ----------------------------------------------------------------------

  private static final DBExceptionConverter	transbaseConverter	=
    new DBTransBaseExceptionConverter();
  private static final DBExceptionConverter	oracleConverter		=
    new DBOracleExceptionConverter();

  private Vector	connections		= new Vector();
  private Connection	defaultConnection;
  private boolean	inTransaction = false;
}
