/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: XDate.x,v 1.10 1999/09/09 10:43:45 vgp Exp $
 */

package at.dms.xoper;

import at.dms.xtype.Date;
import at.dms.xtype.NotNullDate;

/**
 * This class implements all overloaded operators for date and some utilities
 *
 * The class used to represents date is at.dms.xtype.Date
 * We define here every operators than we want to use on date
 *
 * Overloading of operator allows such expression:
 * <PRE>
 * Date	d1 = XDate.now();
 * Date d2 = d1 + 1; // tomorrow
 * int	i = d2 - d1; // one
 * </PER>
 */
public class XDate {

  // ----------------------------------------------------------------------
  // UTILITIES
  // ----------------------------------------------------------------------

  /**
   * Allows conversion (Date)"yyyy-mm-dd"
   *
   * Returns the date corresponding to the string
   */
  public static Date operator:()(String m) {
    if (m == null) {
      return null;
    } else {
      return Date.valueOf(m);
    }
  }

  /**
   * The current date
   */
  public static date now() {
    return new date(System.currentTimeMillis());
  }

  /**
   * Allows conversion "(date)Date"
   */
  public static date operator:()(Date i) {
    if (i == null) {
      throw new NullValueInCastOperator("(date)null");
    }

    return at.dms.xtype.NotNullDate.castToNotNull(i);
  }

  // ----------------------------------------------------------------------
  // SPECIALS OPERATORS (NOT USED IN DATE)
  // ----------------------------------------------------------------------

  /**
   * Allows narrowing conversion "(float)1.3d"
   */
  //public static Date operator:()(int i) {
  //  return new Integer(i);
  //}

  /**
   * Do not allow narrowing conversion "double d = new Double(12.5);"
   */
  //public static Integer operator:=(int i) {
  //  return new Integer(i);
  //}

  // ----------------------------------------------------------------------
  // UNARY OPERATORS
  // ----------------------------------------------------------------------

  /**
   * Operator "*++"
   *
   * Adds a day to this date
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the operand
   * @return	the value for l++
   */
  public static Date operator:*++(Date l) {
    if (l == null) {
      return null;
    }

    long	val = l.getTime();
    l.setTime(val + 86400000L);
    return new date(val);
  }

  /**
   * Operator "++*"
   *
   * Adds a day to this date
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the operand
   * @return	the value for ++l
   */
  public static Date operator:++*(Date l) {
    if (l == null) {
      return null;
    }

    l.setTime(l.getTime() + 86400000L);
    return new date(l.getTime());
  }

  /**
   * Operator "*--"
   *
   * Substracts a day to this date
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the operand
   * @return	the value for l--
   */
  public static Date operator:*--(Date l) {
    if (l == null) {
      return null;
    }

    long	val = l.getTime();
    l.setTime(val - 86400000L);
    return new date(val);
  }

  /**
   * Operator "--*"
   *
   * Substracts a day to this date
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the operand
   * @return	the value for --l
   */
  public static Date operator:--*(Date l) {
    if (l == null) {
      return null;
    }

    l.setTime(l.getTime() - 86400000L);
    return new date(l.getTime());
  }

  // ----------------------------------------------------------------------
  // BINARY OPERATORS
  // ----------------------------------------------------------------------

  /**
   * Operator ">"
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l > r
   */
  public static boolean operator:>(Date l, Date r) {
    if (l == null || r == null) {
      throw new NullValueInComparison("" + l + ">" + r);
    }

    return l.getTime() > r.getTime();
  }

  /**
   * Operator ">="
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l >= r
   */
  public static boolean operator:>=(Date l, Date r) {
    if (l == null || r == null) {
      throw new NullValueInComparison("" + l + ">=" + r);
    }

    return l.getTime() >= r.getTime();
  }

  /**
   * Operator "<"
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l < r
   */
  public static boolean operator:<(Date l, Date r) {
    if (l == null || r == null) {
      throw new NullValueInComparison("" + l + "<" + r);
    }
    return l.getTime() < r.getTime();
  }

  /**
   * Operator "<="
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l <= r
   */
  public static boolean operator:<=(Date l, Date r) {
    if (l == null || r == null) {
      throw new NullValueInComparison("" + l + "<=" + r);
    }
    return l.getTime() <= r.getTime();
  }

  // ----------------------------------------------------------------------
  // ADD OPERATOR
  // ----------------------------------------------------------------------

  /**
   * Operator "+"
   *
   * Adds r days to this date
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l + r
   */
  public static Date operator:+(Date l, int r) {
    if (l == null) {
      return null;
    }
    return new date(l.getTime() + r * 86400000L);
  }

  /**
   * Operator "+"
   *
   * Adds r days to this date
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l + r
   */
  public static date operator:+(date l, int r) {
    if (l == null) {
      throw new NullValueInOperation(l + " + " + r);
    }
    return new date(l.getTime() + r * 86400000L);
  }

  /**
   * Operator "+"
   *
   * Adds r days to this date
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l + r
   */
  public static Date operator:+(Date l, Integer r) {
    if (l == null || r == null) {
      return null;
    }
    return new date(l.getTime() + r.intValue() * 86400000L);
  }

  /**
   * Operator "+"
   *
   * Adds l days to this date
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l + r
   */
  public static Date operator:+(int l, Date r) {
    if (r == null) {
      return null;
    }

    return new date(r.getTime() + l * 86400000L);
  }

  /**
   * Operator "+"
   *
   * Adds l days to this date
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l + r
   */
  public static date operator:+(int l, date r) {
    if (r == null) {
      throw new NullValueInOperation(l + " + " + r);
    }

    return new date(r.getTime() + l * 86400000L);
  }

  /**
   * Operator "+"
   *
   * Adds l days to this date
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l + r
   */
  public static Date operator:+(Integer l, Date r) {
    if (l == null || r == null) {
      return null;
    }

    return new date(r.getTime() + l.intValue() * 86400000L);
  }

  // ----------------------------------------------------------------------
  // MINUS OPERATOR
  // ----------------------------------------------------------------------

  /**
   * Operator "-"
   *
   * Substracts r days to this date
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l - r
   */
  public static Date operator:-(Date l, int r) {
    if (l == null) {
      return null;
    }

    return new date(l.getTime() - r * 86400000L);
  }

  /**
   * Operator "-"
   *
   * Substracts r days to this date
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l - r
   */
  public static date operator:-(date l, int r) {
    if (l == null) {
      throw new NullValueInOperation(l + " + " + r);
    }

    return new date(l.getTime() - r * 86400000L);
  }

  /**
   * Operator "-"
   *
   * Substracts r days to this date
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l - r
   */
  public static Date operator:-(Date l, Integer r) {
    if (l == null || r == null) {
      return null;
    }

    return new date(l.getTime() - r.intValue() * 86400000L);
  }

  /**
   * Operator "-"
   *
   * Substracts two dates and returns a day
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l - r
   */
  public static Integer operator:-(Date l, Date r) {
    if (l == null || r == null) {
      return null;
    }

    return new Integer((int)((l.getTime() - r.getTime()) / 86400000L));
  }

  /**
   * Operator "-"
   *
   * Substracts two dates and returns a day
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l - r
   */
  public static int operator:-(date l, date r) {
    if (l == null || r == null) {
      throw new NullValueInOperation(l + " - " + r);
    }

    return (int)((l.getTime() - r.getTime()) / 86400000L);
  }

  // ----------------------------------------------------------------------
  // BINARY OPERATORS (COMPARISON)
  // ----------------------------------------------------------------------

  /**
   * Operator "=="
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l == r
   */
  public static boolean operator:==(Date l, Date r) {
    if (l == null && r == null) {
      return true;
    } else if (l == null || r == null) {
      return false;
    } else {
      return l.equals(r);
    }
  }

  /**
   * Operator "!="
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l != r
   */
  public static boolean operator:!=(Date l, Date r) {
    if (l == null && r == null) {
      return false;
    } else if (l == null || r == null) {
      return true;
    } else {
      return !l.equals(r);
    }
  }
}
