/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: XMonth.x,v 1.12 1999/08/08 11:29:16 vgp Exp $
 */

package at.dms.xoper;

import java.util.Calendar;
import java.util.GregorianCalendar;
import java.sql.Date;

import at.dms.xtype.Month;
import at.dms.xtype.NotNullMonth;

/**
 * This class implements all overloaded operators for month and some utilities
 *
 * The class used to represents month is java.sql.Month from the JDK libs
 * We define here every operators than we want to use on month
 *
 * Overloading of operator allows such expression:
 * <PRE>
 * Month	d1 = XMonth.now();
 * Month d2 = d1 + 1; // one month later
 * int	i = d2 - d1; // one
 * </PER>
 */
public class XMonth {

  // ----------------------------------------------------------------------
  // UTILITIES
  // ----------------------------------------------------------------------

  /**
   * The current month
   */
  public static month now() {
    return new month(XDate.now());
  }

  /**
   * Allows conversion "(month)Month"
   */
  public static month operator:()(Month i) {
    if (i == null) {
      throw new NullValueInCastOperator("(month)null");
    }

    return month.castToNotNull(i);
  }

  /**
   * Allows conversion "(month)Month"
   */
  public static Month operator:()(Date i) {
    if (i == null) {
      return null;
    }

    return new month(i);
  }

  /**
   * Allows conversion "(month)Month"
   */
  public static month operator:()(date i) {
    if (i == null) {
      throw new NullValueInCastOperator("(month)null");
    }

    return new month(i);
  }

  // ----------------------------------------------------------------------
  // SPECIALS OPERATORS
  // ----------------------------------------------------------------------

  /**
   * Allows conversion (Month)"yyyy-mm"
   *
   * Returns the first day of the month
   */
  public static Month operator:()(String m) {
    if (m == null) {
      return null;
    } else {
      return new month(m);
    }
  }

  /**
   * Allows conversion "(Date)Month"
   *
   * Returns the first day of the month
   */
  public static at.dms.xtype.Date operator:()(Month m) {
    return m == null ? null : m.getDate();
  }

  /**
   * Allows conversion "(Month)Date"
   *
   * Returns the first day of the month
   */
  public static Month operator:()(Date d) {
    return new month(d);
  }

  /**
   * Allows conversion "(Date)Month"
   *
   * Returns the first day of the month
   */
  public static date operator:()(month m) {
    return m == null ? null : m.getDate();
  }

  /**
   * Allows conversion "(Month)Date"
   *
   * Returns the first day of the month
   */
  public static month operator:()(date d) {
    return new month(d);
  }

  /**
   * Do not allow narrowing conversion "double d = new Double(12.5);"
   */
  //public static Integer operator:=(int i) {
  //  return new Integer(i);
  //}

  // ----------------------------------------------------------------------
  // UNARY OPERATORS
  // ----------------------------------------------------------------------

  /**
   * Operator "*++"
   *
   * This add a month to the month
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the operand
   * @return	the value for l++
   */
  public static Month operator:*++(Month l) {
    if (l == null) {
      return null;
    }

    int		val = l.getScalar();
    l.setScalar(val + 1);
    return new month(val);
  }

  /**
   * Operator "++*"
   *
   * This add a month to the month
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the operand
   * @return	the value for ++l
   */
  public static Month operator:++*(Month l) {
    if (l == null) {
      return null;
    }
    l.setScalar(l.getScalar() + 1);
    return new month(l.getScalar());
  }

  /**
   * Operator "*--"
   *
   * This substract a month to the month
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the operand
   * @return	the value for l--
   */
  public static Month operator:*--(Month l) {
    if (l == null) {
      return null;
    }

    int		val = l.getScalar();
    l.setScalar(val - 1);
    return new month(val);
  }

  /**
   * Operator "--*"
   *
   * This substract a month to the month
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the operand
   * @return	the value for --l
   */
  public static Month operator:--*(Month l) {
    if (l == null) {
      return null;
    }
    l.setScalar(l.getScalar() - 1);
    return new month(l.getScalar());
  }

  // ----------------------------------------------------------------------
  // BINARY OPERATORS
  // ----------------------------------------------------------------------

  /**
   * Operator ">"
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l > r
   */
  public static boolean operator:>(Month l, Month r) {
    if (l == null || r == null) {
      throw new NullValueInComparison("" + l + ">" + r);
    }
    return l.getScalar() > r.getScalar();
  }

  /**
   * Operator ">="
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l >= r
   */
  public static boolean operator:>=(Month l, Month r) {
    if (l == null || r == null) {
      throw new NullValueInComparison("" + l + ">=" + r);
    }
    return l.getScalar() >= r.getScalar();
  }

  /**
   * Operator "<"
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l < r
   */
  public static boolean operator:<(Month l, Month r) {
    if (l == null || r == null) {
      throw new NullValueInComparison("" + l + "<" + r);
    }
    return l.getScalar() < r.getScalar();
  }

  /**
   * Operator "<="
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l <= r
   */
  public static boolean operator:<=(Month l, Month r) {
    if (l == null || r == null) {
      throw new NullValueInComparison("" + l + "<=" + r);
    }
    return l.getScalar() <= r.getScalar();
  }

  // ----------------------------------------------------------------------
  // ADD OPERATOR
  // ----------------------------------------------------------------------

  /**
   * Operator "+"
   *
   * This add r months to the month
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l + r
   */
  public static Month operator:+(Month l, int r) {
    if (l == null) {
      return null;
    }
    return new month(l.getScalar() + r);
  }

  /**
   * Operator "+"
   *
   * This add r months to the month
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l + r
   */
  public static month operator:+(month l, int r) {
    if (l == null) {
      throw new NullValueInOperation(l + "+" + r);
    }
    return new month(l.getScalar() + r);
  }

  /**
   * Operator "+"
   *
   * This add r months to the month
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l + r
   */
  public static Month operator:+(Month l, Integer r) {
    if (l == null || r == null) {
      return null;
    }
    return new month(l.getScalar() + r.intValue());
  }

  /**
   * Operator "+"
   *
   * This add l months to the month
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l + r
   */
  public static Month operator:+(int l, Month r) {
    if (r == null) {
      return null;
    }
    return new month(r.getScalar() + l);
  }

  /**
   * Operator "+"
   *
   * This add l months to the month
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l + r
   */
  public static month operator:+(int l, month r) {
    if (r == null) {
      throw new NullValueInOperation(l + "+" + r);
    }
    return new month(r.getScalar() + l);
  }

  /**
   * Operator "+"
   *
   * This add l months to the month
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l + r
   */
  public static Month operator:+(Integer l, Month r) {
    if (l == null || r == null) {
      return null;
    }
    return new month(r.getScalar() + l.intValue());
  }

  // ----------------------------------------------------------------------
  // MINUS OPERATOR
  // ----------------------------------------------------------------------

  /**
   * Operator "-"
   *
   * This substract r months to the month
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l - r
   */
  public static Month operator:-(Month l, int r) {
    if (l == null) {
      return null;
    }
    return new month(l.getScalar() - r);
  }

  /**
   * Operator "-"
   *
   * This substract r months to the month
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l - r
   */
  public static month operator:-(month l, int r) {
    if (l == null) {
      throw new NullValueInOperation(l + "-" + r);
    }
    return new month(l.getScalar() - r);
  }

  /**
   * Operator "-"
   *
   * This substract r months to the month
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l - r
   */
  public static Month operator:-(Month l, Integer r) {
    if (l == null || r == null) {
      return null;
    }
    return new month(l.getScalar() - r.intValue());
  }


  /**
   * Operator "-"
   *
   * This returns the difference in month between the two months
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l - r
   */
  public static Integer operator:-(Month l, Month r) {
    if (l == null || r == null) {
      return null;
    }
    return new Integer((int)(l.getScalar() - r.getScalar()));
  }

  /**
   * Operator "-"
   *
   * This returns the difference in month between the two months
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l - r
   */
  public static int operator:-(month l, month r) {
    if (l == null || r == null) {
      throw new NullValueInOperation(l + "-" + r);
    }
    return l.getScalar() - r.getScalar();
  }

  // ----------------------------------------------------------------------
  // BINARY OPERATORS (COMPARISON)
  // ----------------------------------------------------------------------

  /**
   * Operator "=="
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l == r
   */
  public static boolean operator:==(Month l, Month r) {
    if (l == null && r == null) {
      return true;
    } else if (l == null || r == null) {
      return false;
    } else {
      return l.getScalar() == r.getScalar();
    }
  }

  /**
   * Operator "!="
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l != r
   */
  public static boolean operator:!=(Month l, Month r) {
    if (l == null && r == null) {
      return false;
    } else if (l == null || r == null) {
      return true;
    } else {
      return l.getScalar() != r.getScalar();
    }
  }
}
