/*
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: XTime.x,v 1.10 1999/08/08 11:29:16 vgp Exp $
 */

package at.dms.xoper;

import at.dms.xtype.Time;
import at.dms.xtype.NotNullTime;

/**
 * This class implements all overloaded operators for time and some utilities
 *
 * The class used to represents time is java.sql.Time from the JDK libs
 * We define here every operators than we want to use on time
 *
 * Overloading of operator allows such expression:
 * <PRE>
 * Time	d1 = XTime.now();
 * Time d2 = d1 + 1; // one second later
 * int	i = d2 - d1; // one
 * </PER>
 */
public class XTime {

  // ----------------------------------------------------------------------
  // UTILITIES
  // ----------------------------------------------------------------------

  /**
   * Allows conversion (Time)"hh:mm:ss"
   *
   * Returns the date corresponding to the string
   */
  public static Time operator:()(String m) {
    if (m == null) {
      return null;
    } else {
      return Time.valueOf(m);
    }
  }

  /**
   * The current time
   */
  public static NotNullTime now() {
    return new NotNullTime(System.currentTimeMillis());
  }

  /**
   * Allows conversion "(time)Time"
   */
  public static time operator:()(Time i) {
    if (i == null) {
      throw new NullValueInCastOperator("(time)null");
    }

    return time.castToNotNull(i);
  }

  // ----------------------------------------------------------------------
  // SPECIALS OPERATORS (NOT USED IN TIME)
  // ----------------------------------------------------------------------

  /**
   * Allows narrowing conversion "(float)1.3d"
   */
  //public static Time operator:()(int i) {
  //  return new Integer(i);
  //}

  /**
   * Do not allow narrowing conversion "double d = new Double(12.5);"
   */
  //public static Integer operator:=(int i) {
  //  return new Integer(i);
  //}

  // ----------------------------------------------------------------------
  // UNARY OPERATORS
  // ----------------------------------------------------------------------

  /**
   * Operator "*++"
   *
   * This add a second to the time
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the operand
   * @return	the value for l++
   */
  public static Time operator:*++(Time l) {
    if (l == null) {
      return null;
    }

    long	val = l.getTime();
    l.setTime(val + 1000L);
    return new NotNullTime(val);
  }

  /**
   * Operator "++*"
   *
   * This add a second to the time
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the operand
   * @return	the value for ++l
   */
  public static Time operator:++*(Time l) {
    if (l == null) {
      return null;
    }
    l.setTime(l.getTime() + 1000L);
    return new NotNullTime(l.getTime());
  }

  /**
   * Operator "*--"
   *
   * This substract a second to the time
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the operand
   * @return	the value for l--
   */
  public static Time operator:*--(Time l) {
    if (l == null) {
      return null;
    }

    long	val = l.getTime();
    l.setTime(val - 1000L);
    return new NotNullTime(val);
  }

  /**
   * Operator "--*"
   *
   * This substract a second to the time
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the operand
   * @return	the value for --l
   */
  public static Time operator:--*(Time l) {
    if (l == null) {
      return null;
    }
    l.setTime(l.getTime() - 1000L);
    return new NotNullTime(l.getTime());
  }

  // ----------------------------------------------------------------------
  // BINARY OPERATORS
  // ----------------------------------------------------------------------

  /**
   * Operator ">"
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l > r
   */
  public static boolean operator:>(Time l, Time r) {
    if (l == null || r == null) {
      throw new NullValueInComparison("" + l + ">" + r);
    }
    return l.getTime() > r.getTime();
  }

  /**
   * Operator ">="
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l >= r
   */
  public static boolean operator:>=(Time l, Time r) {
    if (l == null || r == null) {
      throw new NullValueInComparison("" + l + ">=" + r);
    }
    return l.getTime() >= r.getTime();
  }

  /**
   * Operator "<"
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l < r
   */
  public static boolean operator:<(Time l, Time r) {
    if (l == null || r == null) {
      throw new NullValueInComparison("" + l + "<" + r);
    }
    return l.getTime() < r.getTime();
  }

  /**
   * Operator "<="
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l <= r
   */
  public static boolean operator:<=(Time l, Time r) {
    if (l == null || r == null) {
      throw new NullValueInComparison("" + l + "<=" + r);
    }
    return l.getTime() <= r.getTime();
  }

  // ----------------------------------------------------------------------
  // ADD OPERATOR
  // ----------------------------------------------------------------------

  /**
   * Operator "+"
   *
   * This add r seconds to the time
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l + r
   */
  public static Time operator:+(Time l, int r) {
    if (l == null) {
      return null;
    }
    return new NotNullTime(l.getTime() + r * 1000L);
  }

  /**
   * Operator "+"
   *
   * This add r seconds to the time
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l + r
   */
  public static Time operator:+(Time l, Integer r) {
    if (l == null || r == null) {
      return null;
    }
    return new NotNullTime(l.getTime() + r.intValue() * 1000L);
  }

  /**
   * Operator "+"
   *
   * This add l seconds to the time
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l + r
   */
  public static Time operator:+(int l, Time r) {
    if (r == null) {
      return null;
    }
    return new NotNullTime(r.getTime() + l * 1000L);
  }

  /**
   * Operator "+"
   *
   * This add l seconds to the time
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l + r
   */
  public static Time operator:+(Integer l, Time r) {
    if (l == null || r == null) {
      return null;
    }
    return new NotNullTime(r.getTime() + l.intValue() * 1000L);
  }

  // ----------------------------------------------------------------------
  // MINUS OPERATOR
  // ----------------------------------------------------------------------

  /**
   * Operator "-"
   *
   * This substract r seconds to the time
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l - r
   */
  public static Time operator:-(Time l, int r) {
    if (l == null) {
      return null;
    }
    return new NotNullTime(l.getTime() - r * 1000L);
  }

  /**
   * Operator "-"
   *
   * This substract r seconds to the time
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l - r
   */
  public static Time operator:-(Time l, Integer r) {
    if (l == null || r == null) {
      return null;
    }
    return new NotNullTime(l.getTime() - r.intValue() * 1000L);
  }


  /**
   * Operator "-"
   *
   * This returns the difference in second between the two times
   *
   * If one operand is null, this will throw a null pointer exception
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l - r
   */
  public static Integer operator:-(Time l, Time r) {
    if (l == null || r == null) {
      return null;
    }
    return new Integer((int)((l.getTime() - r.getTime()) / 1000L));
  }

  // ----------------------------------------------------------------------
  // BINARY OPERATORS (COMPARISON)
  // ----------------------------------------------------------------------

  /**
   * Operator "=="
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l == r
   */
  public static boolean operator:==(Time l, Time r) {
    if (l == null && r == null) {
      return true;
    } else if (l == null || r == null) {
      return false;
    } else {
      return l.equals(r);
    }
  }

  /**
   * Operator "!="
   *
   * @param	l		the left operand
   * @param	r		the right operand
   * @return	the value for l != r
   */
  public static boolean operator:!=(Time l, Time r) {
    if (l == null && r == null) {
      return false;
    } else if (l == null || r == null) {
      return true;
    } else {
      return !l.equals(r);
    }
  }
}
